/***************
 *
 *	opt_x86cpu.hpp:	x86 CPU specific identification routines
 *	Originator:		Ben Herndon
 *
 *   to be used with seti @ home client code
 *
 */

// Hans Dorn, Oct-05-2006, Joe Segur Nov-13-2006
#if defined(__LINUX__)
#include <algorithm>
#else
#include <stdlib.h>
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <malloc.h>
#include "cpuid_x86.h"


//===========[ PROTOTYPES ]==================

void getCycleFix(void);

//---------------  class  CPU_INFO_t 


#include "memspeed.h"

const double  MILLION = 1e6;

//int CPU_INFO_t::cpu_count = 0, CPU_INFO_t::core_count = 0, CPU_INFO_t::thread_count = 0;
//uint64	CPU_INFO_t::mhz_start, CPU_INFO_t::mhz_end, CPU_INFO_t::tick_start, CPU_INFO_t::tick_end;
//uint32	CPU_INFO_t::cpuMhz = 0,	CPU_INFO_t::busMhz = 0;

CPU_INFO_t	*my_this;

// =======================================================================================
// CPU_INFO_t - CONSTRUCTOR
// =======================================================================================
void CPU_INFO_t::initialize(void) {

	my_this = this;

	has_mmx		= 
	has_sse		= 
	has_sse2	= 
	has_sse3	= 
	has_ssse3	= 
	could_sse	=
	could_sse2	=
	could_sse3	=
	could_ssse3	=
	has_sseOS	= 
	has_3Dnow	= 
	has_3DnowPlus = 
	has_mmxPlus	= false;
	cache.shared_L1_L2 = true;

	extBrand	=
	cpuMhz 		=
	busMhz 		= 0;

	cpu_count	= 
	core_count  = 
	thread_count = 1;

	memset(&cache, 0, sizeof(cache) );
	memset(&cpuid, 0, sizeof(cpuid) );

	// 1.	Establish that the processor has support for CPUID. 
	checkCpuidOpcode();
	if( !has_cpuid_opcode)	return;

//	single_cpu(16);	// Do all our gathering on one cpu (if many)
	mhz_start  = hires_edge();
	tick_start = cycleCount();

	if(!cycleCountFix)	 	getCycleFix();

	// Read in the table of CPUID values
	max_stdFunc = 1;
	max_extFunc = 1;
	for(int idx = 0; idx<= 11 && (idx<= max_stdFunc || idx<= max_extFunc); idx++)
		{
		if( idx <= max_stdFunc)	CPUID_X86(idx, &cpuid[idx][0]);
		if(idx == 0) {
			max_stdFunc =  cpuid[0][0] & 0xFF;
			xcpuid = &cpuid[max_stdFunc+1];
			}
		if( idx <= max_extFunc)
			CPUID_X86(EXT_CMD | idx, &xcpuid[idx][0]);
		if(idx == 0)
			max_extFunc = testxBit(0, eax_, 31) ? xcpuid[0][0] & 0xFF : 0;
		}

	cpuidFixNames();

	// 2.	Get vendor string and the highest standard function supported.
	if(max_stdFunc < 1)	{
		cruising_speed();
		return;
		}

		// 3. execute CPUID function 1, which returns the standard feature flags in the EDX register.
	b_family= (cpuid[1][eax_] >> 8) & 0x0F;
	b_model	= (cpuid[1][eax_] >> 4) & 0x0F;
	rev		= (cpuid[1][eax_] >> 0) & 0x0F;
	type	= (cpuid[1][eax_] >>12) & 0x03;
	extFamily=(cpuid[1][eax_] >>20) & 0xFF;
	extModel= (cpuid[1][eax_] >>16) & 0x0F;
	brandID	= (cpuid[1][ebx_] >> 0) & 0xFF;

		// 4. If bit 23 of the standard feature flags is set to 1, MMX technology is supported. 
	has_mmx		= testBit(1,edx_,23);
	could_sse	= testBit(1,edx_,25);	// Many CPUs don't set these bits correctly
	could_sse2	= testBit(1,edx_,26);
	could_sse3	= testBit(1,ecx_,0);

//	sseOSsupport();

	vendorLookup();	// Get company vendorID

	//  CPUID(1)_EBX 23...16 : CPU count The logical processor count/die is encoded in.
	switch(vendorID) {
		case _AMD:
			cache.shared_L1_L2 = false;
			core_count    = (max_extFunc < 8 ) ? 1 : (xcpuid[8][ecx_] & 0x0F)+1;
			thread_count  = 1;
			has_3Dnow	  = testxBit(1,3,31);
			has_3DnowPlus = testxBit(1,3,30);
			has_mmxPlus	  = testxBit(1,3,22);
			extBrand      = xcpuid[1][ebx_] & 0xFFF;
			break;
		case _INTEL:
			cache.shared_L1_L2 = true;
			bool	hyper	= testBit(1,edx_,28);
			if( hyper ) {
				//CPUID(4)_EAX 31...26 : cores per die - 1
				int logic_count;
				core_count   = (max_stdFunc < 4) ? 1 : ((cpuid[4][eax_] >> 26) & 0x3F)+1;
				logic_count  = (cpuid[1][ebx_] >> 16) & 0xFF;
				thread_count = logic_count / core_count;
				}
			break;
		}
	// OS # of threads
	cpu_count = os_cpu_count() / thread_count / core_count;
	cpu_count = cpu_count ? cpu_count : 1;

	getCacheSizes();
	cpuidLookup();	// Get official company names & CPU names
	cruising_speed();
}



// =======================================================================================
// getCacheSizes
// =======================================================================================
void CPU_INFO_t::getCacheSizes(void) {
	uint32	cacheBits[4] = { 0, 0, 0, 0 };

	cache.L1code =
	cache.L1.size =
	cache.L2.size =
	cache.L3.size = 0;

	// Intel "Deterministic" cache parameters
/*	eax	0:4 Cache type - 5:7 Level -  8 self init - 9 Fully assoc - 10:13 reserved
	  14:25 Threads sharing this cache
	  26:31 cores on this die
	ebx 0:11 coherency line size -1
	  12:21 Line partitions -1
	  22:31 Ways of associativity -1
    ecx number of sets - 1 (loop counter)
*/
	if ( vendorID == _INTEL && max_stdFunc >= 4) {
//		printf("Cache Method #3\nType | Lvl | Size\n");
		int level, idx = 0;
		uint16 S,P,W,L;
		uint32 size;
		while( (type = cpuid[4][eax_] & 0x1F) )
			{
			level = cpuid[4][eax_] >>  5 & 0x7;
			L = 1+ (cpuid[4][ebx_] >>  0 & 0xFFF);
			P = 1+ (cpuid[4][ebx_] >> 12 & 0x3FF);
			W = 1+ (cpuid[4][ebx_] >> 22 & 0x3FF);
			S = 1+  cpuid[4][ecx_];
			size  = L * S * P * W;
			size /= 1024;

//			static char *types[] = { "data", "code", "unif" };
//			printf("%4s | %2d  | %dK  [%d %d %d %d]\n"
//				, types[type-1], level, size, S, P, W, L);
			if ( type == 2)
				cache.L1code += size;
			else switch(level) {
				case 1: cache.L1.size += size; cache.L1.line = L;  break;
				case 2: cache.L2.size += size; cache.L2.line = L;  break;
				case 3: cache.L3.size += size; cache.L3.line = L;  break;
				}
			CPUID_X86(4, ++idx, &cpuid[4][0]);
			}
		CPUID_X86(4, 0, &cpuid[4][0]);	// Put back normal
		if( cache.L1code == 0 )         // Bug if not defined
			cache.L1code = 16;
		}
	if(cache.L1.size && cache.L2.size)	return;

	// AMD L1+L2 code/data cache info settings
	//
	if ( max_extFunc >= 5 && vendorID == _AMD) {
//		printf("Cache Method #2\n");
		cache.L1.size = (xcpuid[5][ecx_] >>24) & 0xFF;	// Data
		cache.L1.line = (xcpuid[5][ecx_] ) & 0xFF;
		cache.L1code  = (xcpuid[5][edx_] >>24) & 0xFF;	// Code
		if (max_extFunc >= 6) {
			cache.L2.size = (xcpuid[6][ecx_] >>16) & 0xFFFF;
			cache.L2.line = (xcpuid[6][ecx_] ) & 0xFF;
			if(cache.L2.size == 1) cache.L2.size = 64;	// Flaw in some Duron's (6-3-0)
			}
		}
	if(cache.L1.size && cache.L2.size)	return;


	// Earlier Intel models - cache info in tokenized values in byte table
	//
	if ( vendorID == _INTEL && max_stdFunc >= 2) {
//		printf("Cache Method #1\n");
		uint8 *bytePtr = (uint8 *) &cpuid[2][0];
		for(int x = 0 ; x < sizeof(cacheBits); x ++, bytePtr++) {
			if( x == 0) continue;
				for(int y = 0; cacheCodes[y].idByte != 0xFF; y++) {
					if(*bytePtr != cacheCodes[y].idByte)
			 			continue;
					cache.L1code  += cacheCodes[y].code;
					cache.L1.size += cacheCodes[y].data;
					cache.L2.size += cacheCodes[y].L2;
					cache.L3.size += cacheCodes[y].L3;
					if( cacheCodes[y].line > cache.L2.line)
						cache.L2.line = cacheCodes[y].line;
				}
			}
		}
	if( cache.L1.size == 0)
		{
		cache.L1.size = 16;	// TODO: write memory timing call to test 8 vs 16 vs 32K
		cache.L1code = 8;	// (a guess)
		}
	if( cache.L2.size == 0)	// Possibly the old external simm module L2
		cache.L2.size = os_L2_cache_size();
	}	



// =======================================================================================
// checkCpuidOpcode - Determine if CPUID is available
//
void CPU_INFO_t::checkCpuidOpcode(void) {
	has_cpuid_opcode = true;
	return;

	has_cpuid_opcode = can_cpuid();
}

// Hans Dorn, Oct-05-2006, Joe Segur Oct-22-2006
#if defined(__GNUC__)

                      //asm( "orps    %xmm1, %xmm1 \n");
#define test_orps()     asm(".byte 0x0F,0x56,0xC9 \n");

                      //asm( "orpd    %xmm1, %xmm1 \n" );
#define test_orpd()     asm(".byte 0x66,0x0F,0x56,0xC9 \n");

                      //asm( "haddps  %xmm1, %xmm1 \n" );
#define test_haddps()   asm(".byte 0xF2,0x0F,0x7C,0xC9 \n");

                      //asm( "phaddw  %xmm1, %xmm1 \n" );
#define test_phaddw() 	asm(".byte 0x66,0x0F,0x38,0x01,0xC9 \n");

#elif  defined( __INTEL_COMPILER )

#define test_orps() __asm{ orps	xmm0, xmm0 }
#define test_orpd() __asm{ orpd	xmm0, xmm0 }
#define test_haddps() __asm{ haddps	xmm0, xmm0 }
#define test_phaddw() __asm{ phaddw	xmm0, xmm0 }

#elif  defined(_MSC_VER)

#define test_orps() __asm{ orps	xmm0, xmm0 }
#define test_orpd() __asm{ orpd	xmm0, xmm0 }
// haddps F2 0F 7C C9
// phaddw 66 0f 38 01 C9
#define test_haddps() {__asm __emit 0F2h __asm __emit 0Fh __asm __emit 7Ch __asm __emit 0C9h }
#define test_phaddw() {__asm __emit 66 __asm __emit 0fh __asm __emit 38h __asm __emit 01h __asm __emit 0c1h }

#endif

static jmp_buf sse_jb;

static void sse_sighandler(int x) {
    longjmp(sse_jb, 10);
}

bool unsignal(void)
	{
	signal(SIGILL, SIG_DFL );	// Put sig handler back to DeFauLt
	return false;
	}

// =======================================================================================
// sseOSsupport - Check for OS support of SSE instruction
// =======================================================================================
bool CPU_INFO_t::sseOSsupport(void) {

	// For illegal instructions - go to sighandler
	signal(SIGILL, &sse_sighandler);

	if( setjmp(sse_jb) ) 
		return  unsignal();	// if it dies, remove signal handler

	test_orps();
	has_sseOS	= true;
	has_sse = true;

	test_orpd();
	has_sse2 = true;

	test_haddps();
	has_sse3 = true;

	test_phaddw();
	has_ssse3 = true;

	unsignal();
	return true;
}

// =======================================================================================
//
// vendorLookup - Find vendor token
//
void CPU_INFO_t::vendorLookup(void)
	{
	CPUID_DEFS	*idDef;
	bool	found = false;
	for( idDef = &cpuid_tbl[0]; !found ; idDef++) {

			// Find a vendor entry [0,0,0,0]
		while(idDef->famModl)
			 idDef++;

			// End of list marker?
		if( idDef->cacheL2 == -1)
			break;
			
			// Vendor code name match?
			// Table Name format:  "[" + <vendorStr> + "]"
		if( strncmp (vendor_id(), &idDef->norm_name[1], 12) != 0)
			continue;

		vendorID = idDef->cacheL2;
		found = true;
		}
	}


// =======================================================================================
//
// cpuidLookup - Find system details with info from cpuid
//
void CPU_INFO_t::cpuidLookup(void) {
	CPUID_DEFS	*idDef;
	bool	found = false;
	int		amd_nn;

	int srch_model = cpuid[1][eax_] & 0xFFF;	// Normal low 3 nibbles usually used for CPUID tables
	int srch_cache = cache.L2.size+cache.L3.size;
	int srch_brand = brandID;

	if(core_count > 1)
		srch_brand = (core_count*thread_count)<<8 | cache.L1code;
	if(cpu_count > 1)
		srch_brand |= cpu_count<<12;
	
/*	strcpy( vendor_id(), "GenuineIntel");
	srch_model = 0x686;
	srch_brand = 0x02;
	srch_cache = 256;
*/
	for( idDef = &cpuid_tbl[0]; !found ; idDef++) {

			// Find a vendor entry [0,0,0,0]
		while(idDef->famModl)
			 idDef++;

			// End of list marker?
		if( idDef->cacheL2 == -1)
			break;
			
			// Vendor code name match?
			// Table Name format:  "[" + <vendorStr> + "]"
		if( strncmp (vendor_id(), &idDef->norm_name[1], 12) != 0)
			continue;

		vendorID		= idDef->cacheL2;
		vendor.abrvName = idDef->generation;
		vendor.company	= idDef->core;

		// Athlon 64 and beyond
		//
		if(	vendorID == _AMD) {
			extBrand = xcpuid[1][1] & 0xFFF;
			if(b_family > 6 && brandID) 
				{
				amd_nn     =  srch_brand & 0x1F;
				srch_brand = (srch_brand & 0xE0) >> 3;
				}
			if(b_family >= 14) {
					//  [8:13]       [4:7]     [0:3]
					// amd_brand | ext_model | model
					amd_nn	    = extBrand & 0x2F;
					srch_brand  = (extBrand >> 6)<<8;
					srch_brand |= extModel<<4 | b_model;
				}
			}

			// Special case... _AMD & _TMT have 2 vendor.id types
			//  so if name matched first, skip the 2nd 
		if(idDef->cacheL2 == idDef[1].cacheL2)
			 idDef++;
		
			// Now find Family/Model match
		for(idDef++  ; !found && idDef->famModl ; idDef++) {
			if(idDef->famModl < srch_model)
				continue;
			if(idDef->brand > 0   && idDef->brand < srch_brand)// Brand matters?
				continue;
			if(idDef->cacheL2 > 0 && idDef->cacheL2 < srch_cache)	// Cache matters?
				continue;

			found	 = true;
			names    = idDef;
			brandID	 = srch_brand;
			strcpy(cpu_notes, idDef-> notes);
			// if(	vendorID == _AMD && b_family == 15)
			//	amd_fix_notes( cpu_notes, amd_nn )
			}
	}
}


// =======================================================================================
//
// cpuidFixNames - Fix supplied name from cpuid
//
void CPU_INFO_t::cpuidFixNames(void) {
	
//	for(int i=0; i<12; i++)
//		vendor_id()[i] = vendor_id()[i+4];
	uint32	a,b,c, *ptr;
	a = cpuid[0][1];
	b = cpuid[0][2];
	c = cpuid[0][3];
	ptr = (uint32 *)&vendor_str;
	ptr[0] = a;
	ptr[1] = c;
	ptr[2] = b;
	vendor_id()[12] = '\0';

	if (max_extFunc <4) 
		cpuid_name()[0] = '\0';			// Terminate name
	else {
		int i, j;						// trim whitespace [>_INTEL<]
		for(i =0; cpuid_name()[i] == ' '; i++)
			;
		// Only loop if i>0 (ie: space found)
		for(j=0; i > 0 && cpuid_name()[i-1]; i++, j++)
			cpuid_name()[j] = cpuid_name()[i];
		}
	}

//===============================
//
// cpu_throttle
//
float CPU_INFO_t::cpu_throttle(void)
	{
	return os_cpu_throttle();
	}


//===============================
//
// cpu_mhz - Determine CPU Mhz speed
//
int CPU_INFO_t::cpu_mhz(void)
	{

	if( cpuMhz == 0)
		{
		uint32	rez;

		mhz_end = hires_timer( rez );
		if( (mhz_end - mhz_start) * 100 / rez < 100)	// xx seconds yet?
			return 0;

//		single_cpu(16);	// Get cycle count from only 1 CPU
		mhz_end  = hires_edge();
		tick_end = cycleCount() - cycleCountFix;
		cruising_speed();

		 // Correct precision for different resolution timers - Hans Dorn, Oct-05-2006
		cpuMhz = (uint32) (double(tick_end - tick_start) / (mhz_end - mhz_start) * (rez / MILLION));
		}

	return cpuMhz;
	}



// =======================================================================================
//
// memory_testing - Determine all levels of cache speed
//
void CPU_INFO_t::memory_testing(void)
	{
	CACHE_LEVEL_ *level;

	if( cache.L1.read_speed[0] < 0 )
		return;

	bool cflush_cmd, flush;
	int	  length, tstlen, len1, len2;
	void  *buffer, *buffer_adr;

	cflush_cmd = testBit(1,edx_,19);
	length = cache.L1.size + cache.L2.size + cache.L3.size;
	length *= 3;
	buffer_adr = malloc( length * 1024 + 32);
	uint32	tmp = (uint32)buffer_adr;
	buffer = (void *)((tmp + 15u) & ~15u);

	for(int memtype = 0; memtype < 4; memtype++)
		{
		switch( memtype ) {
			case 0:
				level = &cache.L1;
				len1  = 
				len2  = level->size -1 ;
				flush = false;
				break;
			case 1:
				level = &cache.L2;
				len1  = cache.L1.size * 5 / 2; 	// * 2.5
				if(len1 < 64) len1 = 64;
				len2  = level->size * 3 / 4;
				if( cache.L2.size <= cache.L1.size ) {
					len1 = len2 = cache.L1.size + cache.L2.size - 1;
					}
				flush = false;
				break;
			case 2:
				level = &cache.L3;	break;
				len1  = cache.L2.size * 5 / 2;
				len2  = level->size * 3 / 4;
				flush = false;
			case 3:
				level = &cache.ram;
				flush = cflush_cmd;
				len1  = (cache.L2.size + cache.L3.size) * 5 / 2;
				len1  = len1>0 ? len1 : length;
				len2  = length * 3 / 4;
				level->size = length;
				break;
			}
		if( level->size == 0) continue;
#if defined(__GNUC__)
		tstlen   = std::min(len1, len2);
		int reps = std::max(30, 1024*50 / tstlen );
#else
		tstlen   = __min(len1, len2);
		int reps = __max(30, 1024*50 / tstlen );
#endif
		level->read_speed[0]  =  memspeed(buffer, tstlen, reps, flush, &read_ints);
		level->write_speed[0] =  memspeed(buffer, tstlen, reps, flush, &write_ints );

		if(has_sse) {
			level->read_speed[1]  =  memspeed(buffer, tstlen, reps, flush, &read_sse );
			level->write_speed[1] =  memspeed(buffer, tstlen, reps, flush, &write_sse );
			if( memtype == 3) {
				level->read_speed[2]  =  memspeed(buffer, length, reps, flush
//					, (cache.L2.line == 64) ? &read_cache32 : &read_cache64);
				// Read 32 seems to work best on AMD regardless of line size
					, &read_cache32 );
				level->write_speed[2] =  memspeed(buffer, length, reps, flush, &write_sse_nt);
				}
			}
		else if(has_mmx)
			{
			level->read_speed[1]  =  memspeed(buffer, tstlen, reps, flush, &read_mmx );
			level->write_speed[1] =  memspeed(buffer, tstlen, reps, flush, &write_mmx );
			if(has_3DnowPlus || has_sse)
				{
				level->write_speed[2] =  memspeed(buffer, tstlen, reps, flush, &write_mmx_nt );
				}
			}
		}
	free(buffer_adr);
	}


//===============================
//
// cache_speed - Return 'some' memory types best reading speed
//
int CPU_INFO_t::cache_speed( CACHE_LEVEL_ &level )
	{
	int	speed;

	if( level.read_speed[0] < 1)
		memory_testing();

#if defined(__GNUC__)
	speed = std::max( level.read_speed[0], std::max( level.read_speed[1], level.read_speed[2] ) );
#else
	speed = __max( level.read_speed[0], __max( level.read_speed[1], level.read_speed[2] ) );
#endif
	if(cpuMhz && speed)
		{
		double cycles = cpuMhz * MILLION;
		speed = (int)(cycles / speed  / 1024.0 * 100 );
		}

	return (cpuMhz) ? speed : -speed;
	}


//===============================
//
// L1_speed - Determine L1 cache speed
//
int CPU_INFO_t::L1_speed(void)
	{
	return cache_speed( cache.L1 );
	}
//
// L2_speed - Determine L2 cache speed
//
int CPU_INFO_t::L2_speed(void)
	{
	return (cache.L2.size) ? cache_speed( cache.L2 ) : 0;
	}
//
// L2_speed - Determine L3 cache speed
//
int CPU_INFO_t::L3_speed(void)
	{
	return (cache.L3.size) ? cache_speed( cache.L3 ) : 0;
	}
//
// RAM_speed - Determine RAM speed
//
int CPU_INFO_t::RAM_speed(void)
	{
	return cache_speed( cache.ram );
	}



//*****
//
// getCycleFix - Calculate compensation offset for overhead of cycleCount call itself
//
//  Accurate cycle counting on the pentium 4 CPU requires an issue of a CPUID 
//   instruction before getting the cycle counter.
//   This causes a long delay as CPUID is a CPU serialization function
//  (it stops all pipelining...clears out of order execution queues...etc.)
//  This function computes the delay caused by the CPUID function execution.
//   
//
uint64 cycleCountFix  = 0;

void getCycleFix(void)	{
	uint64	one, two;

	cycleCount();
	one = cycleCount();
	for(int x = 0; x < 1000; x++)
		two = cycleCount();
	cycleCountFix = (two - one)/1000;
}

