/******************
 *
 * BHSSEfold.cpp
 *
 * Description: Intel & AMD SSE optimized folding
 *  CPUs: Intel Pentium III and beyond,  AMD Athlon model 6 -  and beyond
 *
 * Original code by Ben herndon, modified by Joe Segur
 *
 */

// Copyright 2004 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

//#include "stdafx.h"

// In order to use IPP, set -DUSE_IPP and one of -DUSE_SSE3, -DUSE_SSE2,
// -DUSE_SSE or nothing(generic), IPP precedes FFTW, ooura 
// TMR
/*
Abbreviation    Meaning

px  Generic C optimized code
a6  Intel Pentium III processor code with SSE optimization
w7  Intel Pentium 4 processor code with SSE2 optimization
t7  Intel Pentium 4 processor code with SSE3 optimization
m7  Intel Pentium D processors
m7  Intel Xeon processor with Intel EM64T optimization
i7  Intel Itanium processor and Itanium 2 processor optimization
v8	Intel Core 2-based processors with SSSE3 optimization
s2  Intel IXP4XX Product Line of Network Processor optimization
*/
#if defined(USE_IPP)
#pragma message ("-----IPP-----")
#if defined(USE_SSSE3)				  
#define V8 1						  
#pragma message ("-----SSSE3/tm-----")
#include <ipp_v8.h>					  
#include <tmmintrin.h> // SSSE3       
#include <fvec.h>                     
#elif defined(USE_SSE3)				  
#define T7 1
#pragma message ("-----SSE3/pm-----") 
#include <ipp_t7.h>
#include <pmmintrin.h> // SSE3        
#include <fvec.h>                     
#elif defined(USE_SSE2)
#define W7 1
#pragma message ("-----SSE2/em-----")
#include <emmintrin.h> //SSE2        
#include <fvec.h>                    
#include <ipp_w7.h>
#elif defined(USE_SSE)
#define A6 1
#pragma message ("-----SSE/xm-----") 
#include <xmmintrin.h> //SSE         
#include <fvec.h>                    
#include <ipp_a6.h>
#else
#pragma message ("-----MMX-----")
#include <ipp_px.h>
#endif
#include <ipp.h>
#include <ippvm.h>       
#include <ipps.h>               
#endif

#include "opt_SSE.hpp"
#include "pulsefind.h"

#if defined (__GNUC__) && !defined (__forceinline)
#define __forceinline __attribute__ ((always_inline)) inline
#endif

int sse_msks1[] = {4, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
int sse_msks2[] = {4, 4, 4, 4, 4, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

ALIGN_SIMD float sse_foldLim[][4] = {
    {  0.0f,  0.0f,  0.0f,  0.0f }, // 0
    {  0.0f, 1e37f, 1e37f, 1e37f }, // 1
    {  0.0f,  0.0f, 1e37f, 1e37f }, // 2
    {  0.0f,  0.0f,  0.0f, 1e37f }, // 3
    { 1e37f, 1e37f, 1e37f, 1e37f }, // 4
    };

#define SSE_LIM(idx) (*(VEC *)&sse_foldLim[idx])


// maxp2f -
//  function: Given a 4 packed float input, return the max among them
//  concept:  Ben Herndon
//
static __forceinline float s_maxp2f( VEC max1 )
    {
    float   tMax;
    VEC     maxReg = max1;
    maxReg = s_copyH2L( maxReg, max1 );
    maxReg = s_max( maxReg, max1 );     // [0][1] vs [3][4]
    max1 = s_shuffle( maxReg, maxReg, 1 );
    maxReg = s_max1( maxReg, max1 );    // [1] vs [0]

    s_put1( &tMax, maxReg );
    return ( tMax );
    }


//
// sse_pulPoTf subroutines
//
// These functions reduce an array length by a factor of 3, 4, 5,
// or 2 by summing and return the maximum of the summed values.
// Since each adjacent float is added to the next, simd works well.
// To keep multiple FPU or SSE units in any processor busy we try to
// avoid dependancy chains  ( x += a, x += b, x += c) where each
// computation must wait til previous sum is computed.
// having two maximums running from two colums of numbers
// avoids a single max register from being the bottleneck

//
// Fold by 3 versions
//
sum_func_ptt( sse_pulPoTf3u )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;

    const int   stride = 24;
    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );

        sum1 = s_getA( &ptr1[i + 8] );          sum2 = s_getA( &ptr1[i + 12] );
        s_getU(tmp1, &ptr2[i + 8] );            s_getU(tmp2, &ptr2[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 8] );            s_getU(tmp2, &ptr3[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 8], sum1 );        s_putA( &P->dest[i + 12], sum2 );

        sum1 = s_getA( &ptr1[i + 16] );         sum2 = s_getA( &ptr1[i + 20] );
        s_getU(tmp1, &ptr2[i + 16] );           s_getU(tmp2, &ptr2[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 16] );           s_getU(tmp2, &ptr3[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 16], sum1 );       s_putA( &P->dest[i + 20], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf3 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;

    const int   stride = 8;
    for (i = 0; i < P->di - (stride - 1); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf3L8 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;

    // SSE  Pipeline #1                      SSE Pipeline #2
    //
    int mask1 = sse_msks1[ P->di ];         int mask2 = sse_msks2[ P->di ];
    sum1 = s_getA( &ptr1[0] );              sum2 = s_getA( &ptr1[4] );
    s_getU(tmp1, &ptr2[0] );                s_getU(tmp2, &ptr2[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr3[0] );                s_getU(tmp2, &ptr3[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_putA( &P->dest[0], sum1 );            s_putA( &P->dest[4], sum2 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );  sum2 = s_sub(sum2, SSE_LIM( mask2 ) );
    max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );

    return ( s_maxp2f( s_max( max1, max2 ) ) );
    }

sum_func BHSSETB3[FOLDTBLEN] =
    {
    sse_pulPoTf3L8, sse_pulPoTf3L8, sse_pulPoTf3L8, sse_pulPoTf3L8,
    sse_pulPoTf3L8, sse_pulPoTf3L8, sse_pulPoTf3L8, sse_pulPoTf3L8,
    sse_pulPoTf3L8, sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,
    sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,
    sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,
    sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,   sse_pulPoTf3,
    sse_pulPoTf3u
    };

//
// Fold by 4 versions
//
sum_func_ptt( sse_pulPoTf4u )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;

    const int   stride = 24;
    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 0] );            s_getU(tmp2, &ptr4[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );

        sum1 = s_getA( &ptr1[i + 8] );          sum2 = s_getA( &ptr1[i + 12] );
        s_getU(tmp1, &ptr2[i + 8] );            s_getU(tmp2, &ptr2[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 8] );            s_getU(tmp2, &ptr3[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 8] );            s_getU(tmp2, &ptr4[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 8], sum1 );        s_putA( &P->dest[i + 12], sum2 );

        sum1 = s_getA( &ptr1[i + 16] );         sum2 = s_getA( &ptr1[i + 20] );
        s_getU(tmp1, &ptr2[i + 16] );           s_getU(tmp2, &ptr2[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 16] );           s_getU(tmp2, &ptr3[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 16] );           s_getU(tmp2, &ptr4[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 16], sum1 );       s_putA( &P->dest[i + 20], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr4[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf4 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;

    const int   stride = 8;
    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 0] );            s_getU(tmp2, &ptr4[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr4[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }


sum_func_ptt( sse_pulPoTf4L8 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;

    // SSE  Pipeline #1                      SSE Pipeline #2
    //
    int mask1 = sse_msks1[ P->di ];         int mask2 = sse_msks2[ P->di ];
    sum1 = s_getA( &ptr1[0] );              sum2 = s_getA( &ptr1[4] );
    s_getU(tmp1, &ptr2[0] );                s_getU(tmp2, &ptr2[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr3[0] );                s_getU(tmp2, &ptr3[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr4[0] );                s_getU(tmp2, &ptr4[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_putA( &P->dest[0], sum1 );            s_putA( &P->dest[4], sum2 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );  sum2 = s_sub(sum2, SSE_LIM( mask2 ) );
    max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );

    return ( s_maxp2f( s_max( max1, max2 ) ) );
    }

sum_func BHSSETB4[FOLDTBLEN] =
    {
    sse_pulPoTf4L8, sse_pulPoTf4L8, sse_pulPoTf4L8, sse_pulPoTf4L8,
    sse_pulPoTf4L8, sse_pulPoTf4L8, sse_pulPoTf4L8, sse_pulPoTf4L8,
    sse_pulPoTf4L8, sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,
    sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,
    sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,
    sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,   sse_pulPoTf4,
    sse_pulPoTf4u
    };

//
// Fold by 5 versions
//
sum_func_ptt( sse_pulPoTf5u )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    float *ptr5 = ss[0] + P->tmp3;

    const int   stride = 24;
    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 0] );            s_getU(tmp2, &ptr4[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr5[i + 0] );            s_getU(tmp2, &ptr5[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );

        sum1 = s_getA( &ptr1[i + 8] );          sum2 = s_getA( &ptr1[i + 12] );
        s_getU(tmp1, &ptr2[i + 8] );            s_getU(tmp2, &ptr2[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 8] );            s_getU(tmp2, &ptr3[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 8] );            s_getU(tmp2, &ptr4[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr5[i + 8] );            s_getU(tmp2, &ptr5[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 8], sum1 );        s_putA( &P->dest[i + 12], sum2 );

        sum1 = s_getA( &ptr1[i + 16] );         sum2 = s_getA( &ptr1[i + 20] );
        s_getU(tmp1, &ptr2[i + 16] );           s_getU(tmp2, &ptr2[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 16] );           s_getU(tmp2, &ptr3[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 16] );           s_getU(tmp2, &ptr4[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr5[i + 16] );           s_getU(tmp2, &ptr5[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 16], sum1 );       s_putA( &P->dest[i + 20], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr4[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr5[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf5 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    float *ptr5 = ss[0] + P->tmp3;

    const int   stride = 8;
    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                      SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr3[i + 0] );            s_getU(tmp2, &ptr3[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr4[i + 0] );            s_getU(tmp2, &ptr4[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        s_getU(tmp1, &ptr5[i + 0] );            s_getU(tmp2, &ptr5[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i + 4], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr3[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr4[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_getU(tmp1, &ptr5[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf5L8 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    float *ptr5 = ss[0] + P->tmp3;

    //  SSE Pipeline #1                      SSE Pipeline #2
    //
    int mask1 = sse_msks1[ P->di ];         int mask2 = sse_msks2[ P->di ];
    sum1 = s_getA( &ptr1[0] );              sum2 = s_getA( &ptr1[4] );
    s_getU(tmp1, &ptr2[0] );                s_getU(tmp2, &ptr2[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr3[0] );                s_getU(tmp2, &ptr3[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr4[0] );                s_getU(tmp2, &ptr4[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_getU(tmp1, &ptr5[0] );                s_getU(tmp2, &ptr5[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_putA( &P->dest[0], sum1 );            s_putA( &P->dest[4], sum2 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );  sum2 = s_sub(sum2, SSE_LIM( mask2 ) );
    max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );

    return ( s_maxp2f( s_max( max1, max2 ) ) );
    }

sum_func_ptt( sse_pulPoTf5L4 )
    {
    VEC     sum1, sum2, max1, max2;
    VEC     tmp1, tmp2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    float *ptr5 = ss[0] + P->tmp3;

    int mask1 = sse_msks1[ P->di ];
    sum1 = s_getA( &ptr1[0] );
    s_getU(tmp1, &ptr2[0] );
    sum1 = s_add( sum1, tmp1 );
    s_getU(tmp1, &ptr3[0] );
    sum1 = s_add( sum1, tmp1 );
    s_getU(tmp1, &ptr4[0] );
    sum1 = s_add( sum1, tmp1 );
    s_getU(tmp1, &ptr5[0] );
    sum1 = s_add( sum1, tmp1 );
    s_putA( &P->dest[0], sum1 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
    max1 = s_max( max1, sum1 );

    return ( s_maxp2f( max1 ) );
    }

sum_func BHSSETB5[FOLDTBLEN] =
    {
    sse_pulPoTf5L4, sse_pulPoTf5L4, sse_pulPoTf5L4, sse_pulPoTf5L4,
    sse_pulPoTf5L4, sse_pulPoTf5L8, sse_pulPoTf5L8, sse_pulPoTf5L8,
    sse_pulPoTf5L8, sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,
    sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,
    sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,
    sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,   sse_pulPoTf5,
    sse_pulPoTf5u
    };

//
// Fold by 2 versions
//
sum_func_ptt( sse_pulPoTf2u )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    const int   stride = 24;

    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                          SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i +4], sum2 );

        sum1 = s_getA( &ptr1[i + 8] );          sum2 = s_getA( &ptr1[i + 12] );
        s_getU(tmp1, &ptr2[i + 8] );            s_getU(tmp2, &ptr2[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 8], sum1 );        s_putA( &P->dest[i +12], sum2 );

        sum1 = s_getA( &ptr1[i + 16] );         sum2 = s_getA( &ptr1[i + 20] );
        s_getU(tmp1, &ptr2[i + 16] );           s_getU(tmp2, &ptr2[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 16], sum1 );       s_putA( &P->dest[i +20], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf2 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    const int   stride = 8;

    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                          SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        s_getU(tmp1, &ptr2[i + 0] );            s_getU(tmp2, &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i +4], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        s_getU(tmp1, &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf2L8 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    //  SSE Pipeline #1                          SSE Pipeline #2
    //
    int mask1 = sse_msks1[ P->di ];         int mask2 = sse_msks2[ P->di ];
    sum1 = s_getA( &ptr1[0] );              sum2 = s_getA( &ptr1[4] );
    s_getU(tmp1, &ptr2[0] );                s_getU(tmp2, &ptr2[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_putA( &P->dest[0], sum1 );            s_putA( &P->dest[4], sum2 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );  sum2 = s_sub(sum2, SSE_LIM( mask2 ) );
    max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );

    return ( s_maxp2f( s_max( max1, max2 ) ) );
    }

sum_func_ptt( sse_pulPoTf2L4 )
    {
    VEC     sum1, tmp1, max1;
    int i;

    max1 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    //
    int mask1 = sse_msks1[ P->di ];
    sum1 = s_getA( &ptr1[0] );
    s_getU(tmp1, &ptr2[0] );
    sum1 = s_add( sum1, tmp1 );
    s_putA( &P->dest[0], sum1 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
    max1 = s_max( max1, sum1 );

    return ( s_maxp2f( max1 ) );
    }

// Note: Benchmarks need fully initialized 2fold table
sum_func BHSSETB2[FOLDTBLEN] =
    {
    sse_pulPoTf2L4, sse_pulPoTf2L4, sse_pulPoTf2L4, sse_pulPoTf2L4,
    sse_pulPoTf2L4, sse_pulPoTf2L8, sse_pulPoTf2L8, sse_pulPoTf2L8,
    sse_pulPoTf2L8, sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,
    sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,
    sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,
    sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,   sse_pulPoTf2,
    sse_pulPoTf2u,  sse_pulPoTf2u,  sse_pulPoTf2u,  sse_pulPoTf2u,
    sse_pulPoTf2u,  sse_pulPoTf2u,  sse_pulPoTf2u,  sse_pulPoTf2u
    };


// versions for tmp0 aligned
//
sum_func_ptt( sse_pulPoTf2ALu )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    const int   stride = 24;

    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                          SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        tmp1 = s_getA( &ptr2[i + 0] );          tmp2 = s_getA( &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i +4], sum2 );

        sum1 = s_getA( &ptr1[i + 8] );          sum2 = s_getA( &ptr1[i + 12] );
        tmp1 = s_getA( &ptr2[i + 8] );          tmp2 = s_getA( &ptr2[i + 12] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 8], sum1 );        s_putA( &P->dest[i +12], sum2 );

        sum1 = s_getA( &ptr1[i + 16] );         sum2 = s_getA( &ptr1[i + 20] );
        tmp1 = s_getA( &ptr2[i + 16] );         tmp2 = s_getA( &ptr2[i + 20] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 16], sum1 );       s_putA( &P->dest[i +20], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        tmp1 = s_getA( &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }

sum_func_ptt( sse_pulPoTf2AL )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    const int   stride = 8;

    for (i = 0; i < P->di - ( stride - 1 ); i += stride )
        {
        //  SSE Pipeline #1                          SSE Pipeline #2
        //
        sum1 = s_getA( &ptr1[i + 0] );          sum2 = s_getA( &ptr1[i + 4] );
        tmp1 = s_getA( &ptr2[i + 0] );          tmp2 = s_getA( &ptr2[i + 4] );
        sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
        max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );
        s_putA( &P->dest[i + 0], sum1 );        s_putA( &P->dest[i +4], sum2 );
        }
    max1 = s_max( max1, max2 );
    for (   ; i < P->di ; i += 4)
        {
        int mask1 = sse_msks1[ P->di-i];
        sum1 = s_getA( &ptr1[i + 0] );
        tmp1 = s_getA( &ptr2[i + 0] );
        sum1 = s_add( sum1, tmp1 );
        s_putA( &P->dest[i + 0], sum1 );
        sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
        max1 = s_max( max1, sum1 );
        }
    return ( s_maxp2f(  max1 ) );
    }


sum_func_ptt( sse_pulPoTf2AL8 )
    {
    VEC     sum1, sum2, tmp1, tmp2, max1, max2;
    int i;

    max1 = max2 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    //  SSE Pipeline #1                          SSE Pipeline #2
    //
    int mask1 = sse_msks1[ P->di ];         int mask2 = sse_msks2[ P->di ];
    sum1 = s_getA( &ptr1[0] );              sum2 = s_getA( &ptr1[4] );
    tmp1 = s_getA( &ptr2[0] );              tmp2 = s_getA( &ptr2[4] );
    sum1 = s_add( sum1, tmp1 );             sum2 = s_add( sum2, tmp2 );
    s_putA( &P->dest[0], sum1 );            s_putA( &P->dest[4], sum2 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );  sum2 = s_sub(sum2, SSE_LIM( mask2 ) );
    max1 = s_max( max1, sum1 );             max2 = s_max( max2, sum2 );

    return ( s_maxp2f( s_max( max1, max2 ) ) );
    }

sum_func_ptt( sse_pulPoTf2AL4 )
    {
    VEC     sum1, tmp1, max1;
    int i;

    max1 = s_fill( 0.0 );
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    //
    int mask1 = sse_msks1[ P->di];
    sum1 = s_getA( &ptr1[0] );
    tmp1 = s_getA( &ptr2[0] );
    sum1 = s_add( sum1, tmp1 );
    s_putA( &P->dest[0], sum1 );
    sum1 = s_sub(sum1, SSE_LIM( mask1 ) );
    max1 = s_max( max1, sum1 );

    return ( s_maxp2f( max1 ) );
    }

sum_func BHSSETB2AL[FOLDTBLEN] =
    {
    sse_pulPoTf2AL4, sse_pulPoTf2AL4, sse_pulPoTf2AL4, sse_pulPoTf2AL4,
    sse_pulPoTf2AL4, sse_pulPoTf2AL8, sse_pulPoTf2AL8, sse_pulPoTf2AL8,
    sse_pulPoTf2AL8, sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,
    sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,
    sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,
    sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,  sse_pulPoTf2AL,
    sse_pulPoTf2ALu
    };


FoldSet BHSSEfold = {BHSSETB3, BHSSETB4, BHSSETB5, BHSSETB2, BHSSETB2AL, "BH SSE"};
