#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include "config.h"
#include "math.h"

#include "util.h"
#include "pulsefind.h"
#include "cpuid_x86.h"
#include "malloc_a.h"
#include "chirpfft.h"
#include "analyzePoT.h"
#include "my_types.h"
#include "seti.h"

// Bitwise And values to identify what SIMD capability a routine needs
#define BA_ANY     0x00000001              // any CPU OK
#define BA_MMX     0x00000002
#define BA_SSE     0x00000004
#define BA_SSE2    0x00000008
#define BA_SSE3    0x00000010
#define BA_3Dnow   0x00000020
#define BA_3DnowP  0x00000040
#define BA_MMX_P   0x00000080

struct FolSub {
  FoldSet *fsp;
  int      ba;
};

FolSub FoldSubs[] = {
  &stdFold,      BA_ANY,
  &swifold,      BA_ANY,
  &sse_ben_fold, BA_SSE,
  &AKSSEfold,    BA_SSE,
  &BHSSEfold,    BA_SSE,
};

uint32_t CPUCaps;

/**********************
 */
void SetCapabilities(void) {
  CPU_INFO theCPU;

  CPUCaps = BA_ANY;
  if (theCPU.mmx()) CPUCaps |= BA_MMX;
  if (theCPU.sse()) CPUCaps |= BA_SSE;
  if (theCPU.sse2()) CPUCaps |= BA_SSE2;
  if (theCPU.sse3()) CPUCaps |= BA_SSE3;
  if (theCPU._3Dnow()) CPUCaps |= BA_3Dnow;
  if (theCPU._3DnowPlus()) CPUCaps |= BA_3DnowP;
  if (theCPU.mmxPlus()) CPUCaps |= BA_MMX_P;
}

/**********************/
// dummy folding routine
sum_func_ptt(dmyFld) {return 1.0f;}

// tables to hold function pointers of set under test
sum_func fold3test[FOLDTBLEN] = {
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
};
sum_func fold4test[FOLDTBLEN] = {
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
};
sum_func fold5test[FOLDTBLEN] = {
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
};
sum_func fold2test[FOLDTBLEN] = {
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
};
sum_func fold2ALtest[FOLDTBLEN] = {
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
  dmyFld, dmyFld, dmyFld, dmyFld,
};

FoldSet TestFoldSet = {fold3test, fold4test, fold5test, fold2test, fold2ALtest, ""};

/**********************
 *
 * Generate preplanned sequence of folds
 *
 */
int planFoldTest(PoTPlan * PSeq, float *div, int FFTtbl[][5]) {
  float period;
  int i, iL, j, di, ndivs;
  int num_adds, num_adds_2;
  int FftLength, PulsePoTLen;
  int dbinoffs[32] = {0};
  int k = 0; // plan index

  for (iL = 0; iL < 32; iL++) {
    if ((FftLength = FFTtbl[iL][0]) == 0) continue;
    PulsePoTLen = FFTtbl[iL][4];

    for (i = 32, ndivs = 1; i <= PulsePoTLen; ndivs++, i *= 2);

    period = (float)((int)((PulsePoTLen * 2) / 3)) * 0.5f;
    for (i = 1; i < ndivs; i++, period *= 0.5f) {
      dbinoffs[i] = (((int)(period + 0.5f) + 
                      (MEM_ALIGN/sizeof(float)) - 1) &
                      -(MEM_ALIGN/sizeof(float))) + 
                      dbinoffs[i - 1];
    }

    int32_t PoTL = PulsePoTLen;

    //  Periods from PulsePoTLen/3 to PulsePoTLen/4, and power of 2 fractions of.
    //   then (len/4 to len/5) and finally (len/5 to len/6)
    //
    //  For quick testing we select one from each range at FFT length 8, two at 16,
    //  four at 32, etc.
    //
    int32_t firstP, lastP, iP, iLim;
    for(num_adds = 3; num_adds <= 5; num_adds++) {
      float rnum_adds_minus1;
      float fP, fStep, fLim;
      iLim = FFTtbl[iL][num_adds - 2];
      switch(num_adds) {
        case 3: lastP = (PoTL * 2) / 3;  firstP = (PoTL * 1) / 2; rnum_adds_minus1 = 0.5f; break;
        case 4: lastP = (PoTL * 3) / 4;  firstP = (PoTL * 3) / 5; rnum_adds_minus1 = 1.0f / 3.0f; break;
        case 5: lastP = (PoTL * 4) / 5;  firstP = (PoTL * 4) / 6; rnum_adds_minus1 = 0.25f; break;
      }
      fStep = (float)(lastP - firstP) / (float)(iLim);
      fP = (float)(lastP) - (fStep / 3.0f);

      for (iP = 0; iP < iLim; iP++, fP -= fStep) {
        PSeq[k].dest = div + dbinoffs[0]; // Output storage
        PSeq[k].di = di = (int)(fP * rnum_adds_minus1);
        PSeq[k].tmp0 = (int)(fP * rnum_adds_minus1 + 0.5f);
        PSeq[k].tmp1 = (int)(fP * 2.0f * rnum_adds_minus1 + 0.5f);

        switch(num_adds) {
          case 3:
            PSeq[k].fun_ptr = (di < FOLDTBLEN) ? fold3test[di] : fold3test[FOLDTBLEN-1];
            break;
          case 4:
            PSeq[k].tmp2 = (int)(fP * 3.0f * rnum_adds_minus1 + 0.5f);
            PSeq[k].fun_ptr = (di < FOLDTBLEN) ? fold4test[di] : fold4test[FOLDTBLEN-1];
            break;
          case 5:
            PSeq[k].tmp2 = (int)(fP * 3.0f * rnum_adds_minus1 + 0.5f);
            PSeq[k].tmp3 = (int)(fP * 4.0f * rnum_adds_minus1 + 0.5f);
            PSeq[k].fun_ptr = (di < FOLDTBLEN) ? fold5test[di] : fold5test[FOLDTBLEN-1];
            break;
        }

        k++;                    // next plan
        num_adds_2 = 2 * num_adds;

        for (j = 1; j < ndivs ; j++) {
          PSeq[k].offset = dbinoffs[j - 1];
          PSeq[k].dest = div + dbinoffs[j];
          PSeq[k].tmp0 = di & 1;
          di /= 2;
          PSeq[k].tmp0 += di + PSeq[k].offset;
          PSeq[k].di = di;
          if (PSeq[k].tmp0 & 3) PSeq[k].fun_ptr  = (di < FOLDTBLEN) ? fold2test[di] : fold2test[FOLDTBLEN-1];
          else PSeq[k].fun_ptr  = (di < FOLDTBLEN) ? fold2ALtest[di] : fold2ALtest[FOLDTBLEN-1];

          k++;                    // next plan
          num_adds_2 *= 2;
        }  // for (j =
      } // for (iP = 
    } // for(num_adds =
  } // for(iL =
  PSeq[k].di = 0; // stop
/*
  for (int iDmp = 0; iDmp < 25; iDmp++) {
    printf("di= %d dest= %d ptr= %d tmp0= %d tmp1= %d\n",
           PSeq[iDmp].di, PSeq[iDmp].dest, PSeq[iDmp].fun_ptr, PSeq[iDmp].tmp0, PSeq[iDmp].tmp1);
  }
*/
  return (k);
}


/**********************
 *
 * Test folding subroutine sets and choose best
 *
 */
int FoldTest(ChirpFftPair_t * ChirpFftPairs, int num_cfft, int nsamples, bool do_print) {
  int i, iL, j, k, ndivs, NumPlans = 0, MaxPulsePoT = PoTInfo.PulseMax;
  uint64 timer, onetime, timing, best_timing, reftime, speed = 0x55555555 << 30;
  double accuracy,  best_accuracy, errmax, dTmp = 1e+30;
  int best;
  int PoTLen, PulsePoTLen, Overlap, FFTtbl[32][5] = {0, 0, 0, 0, 0};
  double NumSamples = nsamples;
  float *SrcSel[2];

  // Get the count of chirp/fft pairs for each FFT length where Pulse finding
  // will be done.
  for (i = 0; i < num_cfft; i++) {
    if (ChirpFftPairs[i].PulseFind) {
      for (iL = 0, j = ChirpFftPairs[i].FftLen; j ; iL++, j >>= 1);
      FFTtbl[iL][0] = ChirpFftPairs[i].FftLen;
      FFTtbl[iL][2] += 1;
    }
  }

  // For testing, the chirp/fft table may just be a short list. If so, add the FFT
  // length to the counts to simulate distribution for a full table. (This also affects
  // very high angle range WUs, but Pulse folding is a small factor there so looser
  // testing will have very little effect.)
  if (num_cfft < 14) {
    for (iL = 0; iL < 32; iL++) FFTtbl[iL][2] += FFTtbl[iL][0];
  }

  // Scale the counts so the minimum is 1. Use the scaled value directly for fold
  // sequences starting with a fold by 4. Adjust the values for fold by 3 and
  // fold by 5 to get close to the 10, 9, 8 ratios. Add the PoT length values to
  // the table, and tot up the needed number of PoTPlans.
  for (iL = 0; iL < 32; iL++) {
    if ((FFTtbl[iL][0]) && ((double)FFTtbl[iL][2] < dTmp)) dTmp = (double)FFTtbl[iL][2];
  }
  for (iL = 0; iL < 32; iL++) {
    if (FFTtbl[iL][0]) {
      FFTtbl[iL][2] = (int)((double)FFTtbl[iL][2] / dTmp + 0.5);
      FFTtbl[iL][1] = (int)((double)FFTtbl[iL][2] * 10.0 / 9.0 + 0.5);
      FFTtbl[iL][3] = FFTtbl[iL][2] + FFTtbl[iL][2] - FFTtbl[iL][1];
      PoTLen = (int)(NumSamples / FFTtbl[iL][0] + 0.5);
      GetPulsePoTLen(PoTLen, &PulsePoTLen, &Overlap);
      FFTtbl[iL][4] = PulsePoTLen;
      for (i = 32, ndivs = 1; i <= PulsePoTLen; ndivs++, i *= 2);
      NumPlans += 3 * FFTtbl[iL][2] * ndivs;
    }
  }
  if (NumPlans == 0) return 0; // No pulse finding in this WU, abort test.

  // At some angle ranges the total number of test folds may be too few to get good
  // timing measurements. Scale up so there are at least 8K test folds.
  while (NumPlans < 8192) {
    for (iL = 0; iL < 32; iL++) {
      FFTtbl[iL][1] *= 2;
      FFTtbl[iL][2] *= 2;
      FFTtbl[iL][3] *= 2;
    }
    NumPlans *= 2;
  }

//fprintf(stderr, "Calculated Preplans = %d\n", NumPlans);
//NumPlans *= 2; // temporary safety measure


  // Now we know what we're doing - allocate memory in which to do it.
  PoTPlan *PlanBuf = (PoTPlan *)malloc_a((NumPlans + 1) * sizeof(PoTPlan), MEM_ALIGN);
  float *indata = (float *)malloc_a(MaxPulsePoT * sizeof(float), MEM_ALIGN);
  float *outdata = (float *)malloc_a(MaxPulsePoT * sizeof(float), MEM_ALIGN);
  float *maxdata = (float *)malloc_a(NumPlans * sizeof(float), MEM_ALIGN);
  float *save = (float *)malloc_a(NumPlans * sizeof(float), MEM_ALIGN);

  if (!PlanBuf || !indata || !outdata || !maxdata || !save) {
    if (PlanBuf)
      free_a(PlanBuf);
    if (indata)
      free_a(indata);
    if (outdata)
      free_a(outdata);
    if (maxdata)
      free_a(maxdata);
    if (save)
      free_a(save);
    return 0;   // Can't test, make no change
  }

  double  test_start = dtime();

  SrcSel[0] = indata;
  SrcSel[1] = outdata;
  SetCapabilities();

  // Generate PowerSpectrum random data
  for (i = 0; i < MaxPulsePoT; i++) {
    float fr1 = (float)(rand()) / RAND_MAX;
    float fr2 = (float)(rand()) / RAND_MAX;
    indata[i] = fr1 * fr1 + fr2 * fr2;
  }
  if (do_print) fprintf(stderr, "\n- [ pulse fold select ] -\n");

  for (i = 0; (i * sizeof(FolSub)) < sizeof(FoldSubs); i++) {
  if (!(CPUCaps & FoldSubs[i].ba)) continue;
      j = 0;
      timing = 0;
      CopyFoldSet(&TestFoldSet, FoldSubs[i].fsp);
      int n = planFoldTest(PlanBuf, outdata, FFTtbl);
//    if (!i) fprintf(stderr, "Actual Preplans = %d\n", n);
      while (j < 10) {
        memset(maxdata, 0, NumPlans * sizeof(float));
        ramming_speed();
        timer = cycleCount();
        for ( k = 0; PlanBuf[k].di; k++) {
          maxdata[k] = PlanBuf[k].fun_ptr(SrcSel, &PlanBuf[k]);
        }
        onetime = cycleCount() - timer;
        cruising_speed();
        if (j) timing = std::min(onetime, timing);
        else timing = onetime;
        j++;
      }
      accuracy = 0;
      errmax = 0;
      if (i == 0) {
        memcpy(save, maxdata, NumPlans * sizeof(float));
        reftime = timing;
      } else {
        for (j = 0; j < NumPlans; j++) {
          double diffsq = (save[j] - maxdata[j]) * (save[j] - maxdata[j]);
          accuracy += diffsq;
          if (diffsq > errmax) errmax = diffsq;
        }
      }
      errmax = sqrt(errmax);
      if (do_print) {
        fprintf(stderr, "%30s: %9llu x%2.2f %8.2g\n",
                FoldSubs[i].fsp->name, timing, double(reftime)/double(timing), accuracy);
        if (errmax >= 1e-3)
          fprintf(stderr, "%30s errmax = %7.5f (reject)\n", FoldSubs[i].fsp->name, errmax);
      }
      if ((timing < speed) && (accuracy < 1e-6) && (errmax < 1e-3)) {
        speed = timing;
        best = i;
        best_timing = timing;
        best_accuracy = accuracy;
      }
  }
  free_a(PlanBuf);
  free_a(indata);
  free_a(outdata);
  free_a(maxdata);
  free_a(save);
  if (do_print) {
    fprintf( stderr, "%30s: %9llu (chosen)\n", FoldSubs[best].fsp->name, best_timing );
    fprintf(stderr, "Test Time: %1.2f seconds\n\n", dtime() - test_start );
  }
  CopyFoldSet(&Foldmain, FoldSubs[best].fsp);
  extern int  verbose;
  if(do_print && !verbose)
    exit(0);
  return 0;
}


// std_tableSum2
sum_func_ptt( std_tableSum2 )
    {
    float tmp_max = 0;
    float *ptr1 = ss[1]+P->offset;
    float *ptr2 = ss[1]+P->tmp0;

    for ( int i = 0; i < P->di; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ptr1[i] + ptr2[i];
        P->dest[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }
    return ( tmp_max );
    }

// std_tableSum3
sum_func_ptt( std_tableSum3 )
    {
    float   tmp_max;
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    tmp_max = 0;

    for ( int i = 0; i < P->di; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ptr1[i] + ptr2[i] + ptr3[i];
        P->dest[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }
    return ( tmp_max );
    }

// std_tableSum4
sum_func_ptt( std_tableSum4 )
    {
    float   tmp_max;
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    tmp_max = 0;

    for ( int i = 0; i < P->di; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ptr1[i] + ptr2[i] + ptr3[i] + ptr4[i];
        P->dest[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }

    return ( tmp_max );
    }

// std_tableSum5
sum_func_ptt( std_tableSum5 )
    {
    float   tmp_max;
    float *ptr1 = ss[0];
    float *ptr2 = ss[0] + P->tmp0;
    float *ptr3 = ss[0] + P->tmp1;
    float *ptr4 = ss[0] + P->tmp2;
    float *ptr5 = ss[0] + P->tmp3;
    tmp_max = 0;

    for ( int i = 0; i < P->di; i++ )
        {
        register float  tmpfloat;
        tmpfloat = ptr1[i] + ptr2[i] + ptr3[i] + ptr4[i] + ptr5[i];
        P->dest[i] = tmpfloat;
        if ( tmpfloat > tmp_max )
            {
            tmp_max = tmpfloat;
            }
        }

    return ( tmp_max );
    }

sum_func stdFTB3[FOLDTBLEN]   = {
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3,
  std_tableSum3, std_tableSum3, std_tableSum3, std_tableSum3
};
sum_func stdFTB4[FOLDTBLEN]   = {
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4,
  std_tableSum4, std_tableSum4, std_tableSum4, std_tableSum4
};
sum_func stdFTB5[FOLDTBLEN]   = {
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5,
  std_tableSum5, std_tableSum5, std_tableSum5, std_tableSum5
};
sum_func stdFTB2[FOLDTBLEN]   = {
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2
};
sum_func stdFTB2AL[FOLDTBLEN] = {
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2,
  std_tableSum2, std_tableSum2, std_tableSum2, std_tableSum2
};
FoldSet stdFold = {stdFTB3, stdFTB4, stdFTB5, stdFTB2, stdFTB2AL, "Standard"};
