/******************
 *
 * opt_3DNow.cpp
 * 
 * Description: AMD 3DNow and 3DNow+ optimzized functions
 * CPUs:  3DNow = (begin) AMD K6-3D, All Athlon+Duron+Athlon 64...
 * 
 */

// Copyright 2004 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <limits.h>
#include <xutility>
#include "opt_3Dnow.hpp"
// #include "std_fixes.h"

// [ CODE ] You must have a local VEC variable negReg visible
int negR3d[4] = { INT_MIN, 0, 0, INT_MIN };
int mask_all[2]	= { _b1111, _b1111 };
int mask_hi[2]	= { _b1111, _b0000 };
int mask_lo[2]	= { _b0000, _b1111 };

float zeros_m[2] = { 0, 0};
float recip_2[2] = { .5000000, .5000000};

#define mmx_f_zeros	((MMX *)zeros_m)
#define mmx_f_twos	((MMX *)recip_2)
#define mmx_mask	((MMX *)mask_all)
#define mmx_mask_h	((MMX *)mask_hi)
#define mmx_mask_l	((MMX *)mask_lo)



// getMaxFromPacked - Given a SIMD packed float input, return the max among them
//
// author: Concept of subroutine rather than inline code Ben Herndon
//         3DNow version - Ben Herndon
//
//  WARNING - returns to FPU state, leaves XMM state
//
__forceinline void s_maxp2f( float &f_max, MMX max1) {
	float	tmp_max[2];
	MMX		max2, xmask;

	xmask = s_get(mmx_mask_l);
	max2  = s_and(max1, xmask);		
	max2  = s_accum(max2, max2);  // max2_l now has max1_h
	max1  = s_max( max1,  max2 ); // [0] vs [1]
	s_put(tmp_max, max1);
	s_end();
 	f_max = tmp_max[0];	
}

#define ahead 10
//
// amd_tableSum2 -Used by find_pulse to average blocks of values
//
// author: Concept of subroutine rather than inline code Ben Herndon
//         3Dnow version - Ben Herndon
//
__forceinline void amd_sum2
	(float *sums, c_fp ptr1, c_fp ptr2, MMX &max1, MMX &divisor)
	{
	MMX	sum1, tmp1;
	sum1 = s_get( &ptr1[0] );
	tmp1 = s_get( &ptr2[0] );
	sum1 = s_add( sum1,	tmp1 );
	sum1 = s_mult( sum1, divisor );
	max1 = s_max( max1,	sum1 );
	s_put(	&sums[0], sum1 );
	}

//
//
//
tableSum2_ptt( amd_tableSum2 )
	{
	MMX		max1, max2,	sum1, tmp1;
	int	i;
	const int	stride = 8;

	s_begin();
	MMX	divisor	= s_get(mmx_f_twos);
	max1 = max2	= s_get(mmx_f_zeros);

	i = length;
	if( i & 1) 
		{
		i--;
		sum1 = s_get1( ptr1[i] );
		tmp1 = s_get1( ptr2[i] );
		sum1 = s_add( sum1,	tmp1 );
		sum1 = s_mult( sum1, divisor );
		max1 = s_max( max1,	sum1 );
		s_put(	&sums[i], sum1 );
		}
	for	(    ; i & (stride-1) ; i -= 2 )
		amd_sum2( sums+i-2, ptr1+i-2, ptr2+i-2, max2, divisor);
	for	(    ; i ; i -= stride )
		{
		//	3DNow Pipelines 1 & 2
		//
		amd_sum2( sums+i-2, ptr1+i-2, ptr2+i-2, max1, divisor);
		amd_sum2( sums+i-4, ptr1+i-4, ptr2+i-4, max2, divisor);
		amd_sum2( sums+i-6, ptr1+i-6, ptr2+i-6, max1, divisor);
		amd_sum2( sums+i-8, ptr1+i-8, ptr2+i-8, max2, divisor);
		}

	s_maxp2f(f_max, s_max(max1, max2) );
	}


//
//
//
f_sum_ptt( amd_f_sum )
	{
	MMX		sum1, sum2,	sum3, sum4, tmp1, tmp2;
	int		i;
	float	f_sum[2];
	const int	stride = 8;
	
	s_begin();
	sum1 = sum2 = sum3 = sum4 = s_get(mmx_f_zeros);
	
	// if mask = 1 - get high 3 and subtract 3
	// if mask = 2 -- get high 2 and subtract 2
	// if mask = 3 -- get high 1 and subtract 1
	// if mask = 0 -- get all or and subtract 4

	i = length;
	if( i & 1) 
		{
		tmp2 = s_get1( buf[i-1] );
		sum2 = s_add( sum2,	tmp2 );
		i--;
		}
	for	(  ; i & (stride-1) ; i -= 2 )
		{
		tmp2 = s_get( &buf[i-2] );
		sum2 = s_add( sum2,	tmp2 );
		}
	if( i )
		{
		for	(   ; i  ; i -= stride ) 
			{
			s_fetch( (void *)(buf + i - 6*stride) );

			tmp1 = s_get(&buf[i-2] );
			sum1 = s_add( sum1,	tmp1 );
			tmp2 = s_get(&buf[i-4] );
			sum2 = s_add( sum2,	tmp2 );
			tmp1 = s_get(&buf[i-6] );
			sum3 = s_add( sum3,	tmp1 );
			tmp2 = s_get(&buf[i-8] );
			sum4 = s_add( sum4,	tmp2 );
			}
		
		sum2 = s_add(sum2, s_add(sum3, sum4) );
		}
	sum1 = s_add( sum1, sum2 );
	sum1 = s_accum(sum1, sum1);
	s_put( f_sum, sum1 );
	
	s_end();
	f_total = f_sum[0];
	}

// =============================================================================
//	  v_GetPowerSpectrum
// =============================================================================


GetPowerSpectrum_ptt( amd_v_GetPowerSpectrum )
	{
	int	i, bin;
	float *workBuf = (float *)FreqData;

	s_begin();
	for	( i	= 0, bin = 0; i < this_fft_len; i++, bin += bin_len)
		{
		MMX sum1;
		MMX *p = (MMX *)(&PowerSpectrum[bin_off + bin]);

		s_fetchW( &FreqData[i+16][0] );
		sum1 = s_get(&FreqData[i][0]);
		sum1 = s_mult(sum1, sum1);
		sum1 = s_accum(sum1, sum1);
			// WARNING: !! this store overwrites FreqData[i][0] and [i][0],
			//		so loop must go bottom to top !!
			//  reusing buffer - not needed after our psNum compute.
		s_put(&workBuf[i], sum1);	// 	workBuf[i] = psNum;

			// Try to get:
			// movsd  [reg32],[xmmreg]
			// movnti [p],[reg32]
		s_putNC(p,  sum1 );
		}
	s_end();

	ex_count_flops( 3 * this_fft_len );
	}
