/* benchmark.cpp:

 * Function: test bench code for seti@home client
 * Originator: Ben Herndon
 *
 * Test bench for optimization and benchmarking routines
 * to be used with seti @ home client code
 *
 * Thanks to Eric Korpela and analyzeFuncs_vector for ideas for buffer
 *  verify techinque (difference^2), output format
 *  multiple retries of function, tables of version function pointers
 *
 * Test bench for optimization and benchmarking routines
 * to be used with seti @ home client code
 *
 * - tests repeatedly the timing of user specified functions
 * - finds minimum execution time of multiple runs of the function
 * - gives code maximum priority (current process) during timing and
 *   function execution...to try to avoid system overhead & context switches
 *   adding their time to the function being tested
 * - verifies output of tested function vs. original "known good" output
 * - determines the version of the function with the lowest time at runtime
 *
 * To add a function to be tested:
 * 1. add an enum to f_token to represent your function
 * 2. include whatever header file you need for 'prototypes' and 'extern's
 *    of your function and various versions of it
 * 3. duplicate an existing line of the  bench_list[] table
 *    modifying it with your function's particulars
 *    (description, simd used, your new token from step 1)
 * 4. inside of the functions  BENCH_C::prep_it(), post_it(), time_it()
 *    and verify_it() there are switch statements using the tokens
 *    from step 1.
 *    Use the example case statement to make equivalent
 *      a. prep_it() preparations for running your function
 *         (buffers, buffer pointers, intiialized random data, trig tables, etc)
 *      b. post_it() to save away the "known good" results from the base/original
 *         function's output, for later comparison to other version's output
 *      c. time_it() - the actual calling and timing of the function
 *      d. verify_it() - compare the output to original "known good" results
 *
 * 5. A function call to this code must be placed within whatever program source
 *    you are planning to test.   In seti@home you could place a call either in
 *    seti_analyze()  or main() (after certain setup functions are called)
 * 6. To begin your benchmarking make a call to optimize_init() in this module.
 *    Currently there is an exit() call in optimize_init so program will stop after
 *    benchmarking all versions...for development purposes.
 *    A released version would have remove this exit() call, of course.
 *
 * 7. When called, the program outputs statistics for each function version
 *    to 'stderr' . (for seti, this is the stderr.txt file)
 *    Time used,
 *    ratio of time vs original function,
 *    accuracy rate of output buffer (if any)
 *    returned value (if any)
 *
 *  There are default verify_it and post_it code which should cover most situations.
 *  The case statements for the demo function don't do anything for those 2 _it()s
 *
 */
#include "stdafx.h"
#include "cpuid_x86.h"
#include "math.h"
#define MAX_CYCLES (2000ul*1024ul*1024ul)

// Prototype references for optimized routines
//

//#include "opt_SSE.hpp" - now in optimize.hpp
//#include "opt_SSE2.hpp" - now in optimize.hpp
//#include "opt_3Dnow.hpp"
#include "pulsefind.h"
#include "chirpfft.h"
#include "util.h"

#if defined(USE_IPP)
#pragma message ("-----IPP-----")
#if defined(USE_SSSE3)				  
#define V8 1						  
#pragma message ("-----SSSE3-----")
#include <ipp_v8.h>					  
#include <tmmintrin.h> // SSSE3       
#include <fvec.h>                     
#elif defined(USE_SSE3)				  
#define T7 1
#pragma message ("-----SSE3-----") 
#include <ipp_t7.h>
#include <pmmintrin.h> // SSE3        
#include <fvec.h>                     
#elif defined(USE_SSE2)
#define W7 1
#pragma message ("-----SSE2-----")
#include <emmintrin.h> //SSE2        
#include <fvec.h>                    
#include <ipp_w7.h>
#elif defined(USE_SSE)
#define A6 1
#pragma message ("-----SSE-----") 
#include <xmmintrin.h> //SSE         
#include <fvec.h>                    
#include <ipp_a6.h>
#else
#pragma message ("-----MMX-----")
#include <ipp_px.h>
#endif
#include <ipp.h>
#include <ippvm.h>       
#include <ipps.h>
#elif defined( USE_FFTWF )
    #include "fftw3.h"
#endif

// Original functions - unmodified from release seti worker code
//
enum f_token { CHIRP, PWRSPEC, PSONLY, TRNSTRP, PEAK, TMEAN, PULSE, SUM2_TBL
	, F_SUM, CHI_SQ, _FFT, _LOOP_TIME, _END_ };

extern ChirpData_func			orig_ChirpData;

extern int  find_pulse( const float *fp_PulsePot, int PulsePotLen, float pulse_thresh,
                        int TOffset, int FOffset );

// Hand reorganized c++ versions of functions
//

extern f_GetPeak_func			sse_f_GetPeak;
extern f_sum_func				sse_f_sum;
extern f_sum_func				sse_f_sum2;

extern f_sum_func				amd_f_sum;

extern ChirpData_func			sse2_ChirpData;
extern ChirpData_func			sse1_ChirpData_ak;
extern ChirpData_func			sse2_ChirpData_ak;
extern ChirpData_func			sse3_ChirpData_ak;

extern bool	enough_ram( double needed );
static bool canTrigArray, wantTrigArray;

// Function pointers, used inside existing functions, for sections of those functions
//   that lend themselves to optimization
//  Original code is removed and put into opt_unopt
//  Function call is put in place of original section of code
//  Function call is actually to a function pointer
//  Pointer is replaced at runtime with best variety of function for whatver
//   host it is running on
//
//  (these are what everyone calls through)

ChirpData_func          *ChirpData = &std_v_ChirpData;
TrigArrayInit_func      *TrigArrayInit = &InitTrigArray;
TrigArrayFree_func      *TrigArrayFree = &FreeTrigArray;
f_GetPeak_func          *f_GetPeak = &std_f_GetPeak;
GetPowerSpectrum_func   *GetPowerSpectrum = &std_v_GetPowerSpectrum;
PwrSpectrumOnly_func    *PwrSpectrumOnly = &orig_v_GetPowerSpectrum;
TransposeStrip_func     *TransposeStrip = &orig_v_Transpose4;
f_GetChiSq_func			*f_GetChiSq = &std_f_GetChiSq;

lower_rebin_func		*lower_rebin;
rms_dev_func			*rms_deviation;

f_sum_func				*f_sum;
tableSum2_func			*tableSum2;
tableSum3_func			*tableSum3;
tableSum4_func			*tableSum4;
tableSum5_func			*tableSum5;

extern FoldSet sse_ben_fold;

//  Benchmark table list
//
//  For each main function with versions..there is a token in the f_token enum table
//  Each line in the table is a different optimized version of a main function
//  Flags identify what SIMD must be present for version to execute
//
enum _simd_type { _fpu, _3DNow, _3DNowPP, _sse, _sse2, _sse3, _altivec };

struct bench_lst
    {
    f_token     token;
    _simd_type  simd_used;
    int         subtype;
	bool		tested;
    char        *name;
    void        *theFunc;
    }	bench_list[] =
    {
	PWRSPEC,	_fpu,	0,	false,	"PowerSpectrum--",	(void *)&std_v_GetPowerSpectrum,
	PSONLY, 	_fpu,	0,	false,	"PwrSpectOnly--",	(void *)&orig_v_GetPowerSpectrum,
    TRNSTRP,	_fpu,	0,	false,	"Transpose--",	(void *)&orig_v_Transpose,
    TRNSTRP,	_fpu,	0,	false,	"Transpose2",	(void *)&orig_v_Transpose2,
    TRNSTRP,	_fpu,	0,	false,	"Transpose4",	(void *)&orig_v_Transpose4,
	CHIRP,		_fpu,	0,	false,	"ChirpData--",	(void *)&orig_ChirpData,
    PEAK,		_fpu,	0,	false,	"GetPeak--",	(void *)&std_f_GetPeak,
    PULSE,		_fpu,	0,	true,	"find_pulse--",	(void *)&find_pulse,
//	SUM2_TBL,	_fpu,	0,	false,	"sumTables2--",	(void *)&std_tableSum2,
    F_SUM,		_fpu,	0,	false,	"f_sum--",		(void *)&std_f_sum,
	CHI_SQ,		_fpu,	0,	false,	"GetChiSq--",	(void *)&std_f_GetChiSq,
#if defined( USE_FFTWF )
	_FFT,		_fpu,	0,	false,	"FFT FFTW3(64K)",		NULL,
#elif defined ( USE_IPP )
	#pragma message( "-----IPP-----" )
    #if defined( USE_SSE3 )
		#pragma message( "-----SSE3-----" )
		_FFT,		_sse3,	0,	false,	"IPP FFT SSE3(64K)",	NULL,
    #elif defined( USE_SSE2 )
		#pragma message( "-----SSE2-----" )
		_FFT,		_sse2,	0,	false,	"IPP FFT SSE2(64K)",	NULL,
    #elif defined( USE_SSE )
        #pragma message( "-----SSE-----" )
        _FFT,       _sse,   0,	false,  "IPP FFT SSE1(64K)",    NULL,
    #else
        #pragma message( "-----mmx-----" )
        _FFT,       _fpu,   0,	false,  "IPP FFT MMX(64K)",    NULL,
    #endif
#endif

#ifndef NO_TRIGARRAY
	CHIRP,		_fpu,	1,	false,	"TrigArray",	(void *)&std_v_ChirpData,
#endif
    PWRSPEC,    _fpu,	0,	true,	"hand_opt",		(void *)&opt_v_GetPowerSpectrum,
	CHIRP,		_fpu,	1,	true,	"unrolled",		(void *)&opt_ChirpData,
    PEAK,		_fpu,	0,	false,	"hand_opt",		(void *)&opt_f_GetPeak,
//	SUM2_TBL,	_fpu,	0,	false,	"hand_opt",		(void *)&opt_tableSum2,
    F_SUM,		_fpu,	0,	false,	"unroll4",		(void *)&opt_f_sum,
	CHI_SQ,		_fpu,	0,	false,	"hoisted+abs(",	(void *)&opt_f_GetChiSq,

//	SUM2_TBL,	_sse,	0,	false,	"hand_sse",		(void *)&sse_tableSum2,
    PEAK,		_sse,	0,	false,	"sse_vector",	(void *)&sse_f_GetPeak,
    F_SUM,		_sse,	0,	false,	"hand_sse",		(void *)&sse_f_sum,
    F_SUM,		_sse,	0,	false,	"sse_vector",	(void *)&sse_f_sum2,
	PWRSPEC,	_sse,	0,	false,	"sse_GetTPS",	(void *)&sse_GetTPS_sc8,
	PSONLY, 	_sse,	0,	false,	"sse_GetPSO_npr",	(void *)&sse_GetPSO_sc16_npr,
	PSONLY, 	_sse,	0,	false,	"sse_GetPSO_p32",	(void *)&sse_GetPSO_sc16_p32,
	PSONLY, 	_sse,	0,	false,	"sse_GetPSO_p64",	(void *)&sse_GetPSO_sc16_p64,
	PSONLY, 	_sse,	0,	false,	"sse_GetPSO_p128",	(void *)&sse_GetPSO_sc16_p128,
    TRNSTRP,	_sse,	0,	false,	"sse_Trans4ntw",	(void *)&sse_vTranspose4ntw,
    TRNSTRP,	_sse,	0,	false,	"sse_pfTrans8x4ntw",	(void *)&sse_vpfTranspose8x4ntw,

//	SUM2_TBL,	_3DNow,	0,	true,	"hand_3Dnow",	(void *)&amd_tableSum2,
//  F_SUM,		_3DNow,	0,	true,	"hand_3Dno",	(void *)&amd_f_sum,

	CHIRP,		_sse,	0,	false,	"sse1_akChirp",		(void *)&sse1_ChirpData_ak,

//	CHIRP,		_sse2,	1,	false,	"sse2_chirp",		(void *)&sse2_ChirpData,
	CHIRP,		_sse2,	0,	false,	"sse2_akChirp",		(void *)&sse2_ChirpData_ak,
	PWRSPEC,	_sse2,	0,	false,	"hand_sse2",		(void *)&sse2_v_GetPowerSpectrum,
	PWRSPEC,	_sse2,	0,	false,	"sse2_GetTPS",		(void *)&sse2_GetTPS_sc8,

	CHIRP,		_sse3,	0,	false,	"sse3_akChirp",		(void *)&sse3_ChirpData_ak,

	_END_,		_fpu,	0,	false,	"end",			NULL
    };


// =============================================================================
//    BENCH_C
//
//   A class to perform the timing comparisons of optimized versions of functions
// =============================================================================
class BENCH_C
    {
	public:
				BENCH_C()  : workBuf(0) {};
				~BENCH_C( void ) { if(workBuf) free_a(workBuf);};
		int		bestFuncs( CPU_INFO &our_cpu );
		bool	last_bench( int idx ) { return bench_list[idx].token == _END_; };
		bool	same_group( int idx ) { return bench_list[idx].token == bench_list[base_func].token; };
		bool	tested( int idx ) { return bench_list[idx].tested; };
		void	prep_it( void );
		void	post_it( void  );
		uint64	time_it(  int idx );
		double	verify_it( int idx );
		void	report_it ( int idx );
		void	save_best( int idx );
		bool	simd_match( CPU_INFO &our_cpu, int simd_used );

		// commands passed to testing functions
		enum test_cmds { PREP, POST, TIME, VERIFY, FASTEST, CLEANUP };

		void test_chirp( int command );
		void test_chi_sq( int command );
	private:
		int         NumDataPoints;
		float		*workBuf;
		double		bench_start;

		bool		normal_verify, normal_post;

		int			base_func, variety;
		int			test_points, test_size;
		float		*orig_buf, *in_buf[5], *out_buf;
		double		orig_value, test_value;
		double      accuracy, accu_sum;
		int			loop_count;
		uint64		timing, base_time, best_time;
		uint32		time_overhead;
		enum		{no_pre, zero_out, fill_in} pre_test;

		double		user_d1, user_d2, user_d3;
		int			user_i1, user_i2, user_i3;

        #if defined( USE_IPP )
			IppsFFTSpec_C_32fc  *FftSpec;
        #elif defined( USE_FFTWF )
			fftwf_plan  da_fft_plan;
		#endif

	};

typedef BENCH_C BENCH;

// =============================================================================
//
// BENCH_C::bestFuncs
//
// Loop through each main_function to be optimized
//  1. run the unoptimized original so that we can
//    a) get valid output data to compare the output of sub-versions
//    b) get a timing on how fast original vs. improved varieties
//  2. continue through the list of functions only executing versions
//     of current test function
//  3. time them, verify their output, record fastest
//  4. when we reach end of table - varieties testing --
//  5. we then go back and test next main_function
//
// =============================================================================
int BENCH_C::bestFuncs( CPU_INFO &our_cpu)
	{
	bench_start = dtime();

	if(show_benchmark)
		fprintf(stderr, "------- [ benchmark ] --------\n");
	NumDataPoints = 1024*1024;
    workBuf = (float *)malloc_a( 10 * NumDataPoints * sizeof( float ), MEM_ALIGN );

    // FORCE_FRAME_POINTER;
    if ( !workBuf )
        {
        fprintf( stderr, "Memory allocation failed in benchmark\n" );
        return -1;
        }

    // Make some random data
	const float ourmin = -2.0;
	const float ourmax =  2.0;
	const float ourfact = (ourmax-ourmin)/RAND_MAX;
	for (int i = 0; i < NumDataPoints*2; i++ )
		workBuf[i] = ourmax - rand()*ourfact;	// + or - max rand #
/*
    for (int i = 0; i < NumDataPoints*2 - 1; i += 2 )
		{
		workBuf[i + 0] = 1.0f;  //real
		workBuf[i + 1] = 0.0f;  //imag
		}
*/
//	time_overhead = time_it( _LOOP_TIME );

    // install_sighandler();

	// Loop through all classes of function to test
    for ( base_func = 0; !last_bench( base_func ); base_func++ )
        {
        int     best_variety;
		double	best_value;

		if( tested(base_func))	continue;
		if ( ! simd_match( our_cpu, bench_list[base_func].simd_used ) )
			continue;

		accuracy = 0;
		base_time = 0;
		orig_value = 0;

		variety = base_func;
			// Prepare data, time calls, and prepare to test remaining varieties
		prep_it();
        base_time  = time_it( base_func );
        post_it( );
        report_it( base_func );

		best_time = base_time;
		best_variety = base_func;

        // loop through & test all varieties of this function
        //
		for ( variety = base_func + 1; !last_bench( variety ); variety++ )
            {
//			fprintf(stderr, "%s\n", bench_list[variety].name);
            if ( !same_group( variety) || tested(variety)  )
				continue;	// Not a sub-variety of base

			if ( ! simd_match( our_cpu, bench_list[variety].simd_used ) )
				continue;

			prep_it();
			time_it( variety );
            verify_it( variety );
            report_it( variety  );

			bench_list[variety].tested = true;
			if ( timing < best_time && (accuracy < 1e-5
				||  (accuracy < 5e-3 && bench_list[variety].token == CHIRP) ) )
                {
                best_time    = timing;
                best_variety = variety;
				best_value   = test_value;
                }
            } // end of varieties
		save_best(best_variety);

        } // end of optimizable functions

	if(show_benchmark)
		fprintf(stderr, "Bench Time: %1.2f seconds\n", dtime()-bench_start );

	// uninstall_sighandler();
	return 0;
    }

// =============================================================================
//    simd_match
// =============================================================================
bool BENCH_C::simd_match( CPU_INFO &our_cpu, int simd_used )
    {
	bool	have_simd = true;

	switch( simd_used )
		{
		case _sse:		have_simd = our_cpu.sse(); break;
		case _sse2:		have_simd = our_cpu.sse2(); break;
		case _sse3:		have_simd = our_cpu.sse3(); break;
		case _3DNow:	have_simd = our_cpu._3Dnow(); break;
		case _3DNowPP:	have_simd = our_cpu._3DnowPlus(); break;
		case _altivec:
		default:	// _fpu
			break;
		}
	return have_simd;
	}


// =============================================================================
//    prep_it
// =============================================================================
void BENCH_C::prep_it( void )
    {
	// Default prep_it values
	//  |  in_buf[0]  |  in_buf[1]  | out_buf  |  orig_buf  |
	//  |-- generated random data---|
	//
	test_points = NumDataPoints;
	test_size	= test_points * sizeof(float);
	in_buf[0]	= &workBuf[0];
	in_buf[1]	= &workBuf[test_points];
 	out_buf		= &workBuf[NumDataPoints*2];
	orig_buf	= &out_buf[test_points ];
	pre_test	= zero_out;

	int	fft_len = 64*1024;
	int order = 0;
	switch ( bench_list[base_func].token )
        {
		case _FFT:
			#if defined( USE_IPP )
				ippStaticInit();                // initialization of IPP library
				for ( int tmp = fft_len; !(tmp & 1); order++ )
					tmp >>= 1;
				ippsFFTInitAlloc_C_32fc( &FftSpec, order, IPP_FFT_NODIV_BY_ANY, ippAlgHintFast );
				pre_test = fill_in;
			#elif defined( USE_FFTWF )
				extern void load_wisdom(void);
				load_wisdom();
				// TODO: Deallocate these at the end of the function
				da_fft_plan = fftwf_plan_dft_1d(
						fft_len,
						( sah_complex * ) out_buf,
						( sah_complex * ) in_buf[0],
						FFTW_BACKWARD,
						FFTW_PATIENT  | FFTW_PRESERVE_INPUT );
			#endif
			break;
		case PEAK:
		case CHI_SQ:
				// Setup f_weight[] negative offsets to match positive
				// f_weight will be
			test_points = 1000;
			for(int i=0; i< test_points; i++) {
				out_buf[test_points-i] = in_buf[1][i];
				out_buf[test_points+i] = in_buf[1][i];
				}
			in_buf[1] = &out_buf[test_points];
			pre_test = no_pre;
			break;
        case PWRSPEC:
            // Testing with simulated output of 64K FFT
            test_points = NumDataPoints / 16;
            test_size = test_points * sizeof(sah_complex);
            orig_buf = &out_buf[NumDataPoints];
            pre_test = fill_in;
            test_value = 0;
			break;
        case PSONLY:
            // Testing with simulated output of 64K FFT
            test_points = NumDataPoints / 16;
            test_size = test_points * sizeof(sah_complex);
            orig_buf = &out_buf[NumDataPoints];
            pre_test = fill_in;
            test_value = 0;
			break;
        case TRNSTRP:
            // Testing with simulated output of 16K FFT
            test_points = NumDataPoints / 16;            // but 4 rows = 64K
            test_size = NumDataPoints * sizeof(float);   // for PowerSpectrum fill_in
            orig_buf = &out_buf[NumDataPoints];
            pre_test = fill_in;
            test_value = 0;
            break;
        case SUM2_TBL:
			//memcpy( in_buf[2], in_buf[0], test_points*sizeof(float));
			break;
        case CHIRP:
            CacheChirpCalc = true;
			test_points = NumDataPoints / 4;
            in_buf[1]   = &workBuf[NumDataPoints * 2];
            out_buf     = &workBuf[NumDataPoints * 4];
            orig_buf    = &workBuf[NumDataPoints * 6];
            test_size   = test_points * 2 * sizeof(float);
            user_i1 = (int) (49.7 / TEST_MIN_CHIRP_STEP);
            user_d1 = TEST_MIN_CHIRP_STEP;
            user_d2 = 9765.625000;  // Subband Sample Rate
/*          <subband_desc>
            ...
            <sample_rate>9765.625</sample_rate>
*/
            if ( bench_list[variety].subtype == 1 )
	            InitTrigArray( test_points, user_d1, user_i1, user_d2);
			user_i1 = -(abs(user_i1) + 0);
			pre_test = no_pre;
            break;
		case F_SUM:
				test_points = NumDataPoints-6;
				pre_test = no_pre; break;
		case PULSE:     break;
        case TMEAN:     break;
		case _LOOP_TIME: break;
        }
    }
//
// post_it - base function post processing
//
void BENCH_C::post_it(  void )
    {
	normal_post = true;
	switch ( bench_list[variety].token )
        {
		case _FFT:	break;
        case SUM2_TBL:	break;
        case CHIRP:     break;
        case PWRSPEC:   break;
        case PSONLY:    break;
        case TRNSTRP:   break;
        case F_SUM:
        case CHI_SQ:
        case PEAK:      normal_post = false; break;
        case PULSE:     break;
        case TMEAN:     break;
		case _LOOP_TIME: break;
        }
	if(normal_post) {
		memcpy( orig_buf, out_buf, test_size);
		int points = test_size/sizeof(float);
		accu_sum = 0;
        for (int j=0; j < points; j++)
			accu_sum += fabs( orig_buf[j] );
		accu_sum /= points;
		}
	orig_value = test_value;
    }


// =============================================================================
//   BENCH_C::time_it
// =============================================================================
#define call_it( func_ ) (( func_ *) func_p)

uint64 BENCH_C::time_it(  int idx )
    {
    int     token, loops;
    void    *func_p;
	uint64	cycles, end_cyc, total_run;

	token = bench_list[variety].token;
    func_p = bench_list[idx].theFunc;

	// Clear output buffer, to avoid failed variety func but leftover good data

	timing = uint64(1)<<62;
	total_run	=
	end_cyc		= cycleCount();
	for(loops = 0; loops < 50 && (end_cyc-total_run)< MAX_CYCLES; loops++)
		{
		if(pre_test == zero_out)	memset( out_buf, 0, test_size );
		if(pre_test == fill_in)		memcpy( out_buf, workBuf, test_size );
		ramming_speed();
		cycles = cycleCount();
		switch ( bench_list[idx].token )
			{
			case _FFT:
				#if defined( USE_IPP )
					ippsFFTInv_CToC_32fc(
						( Ipp32fc * ) out_buf,
						( Ipp32fc * ) out_buf,
						FftSpec,
						NULL );
				#elif defined( USE_FFTWF )
					fftwf_execute_dft( da_fft_plan, in_buf[0],	out_buf );
				#endif
				break;
			case SUM2_TBL:
/*				float tmp_max;
				for(int runit=0; runit< 10; runit++) {
					call_it(tableSum2_func )( tmp_max,  5, out_buf+50, in_buf[0], in_buf[1] );
					call_it(tableSum2_func )( tmp_max, 11, out_buf+50, in_buf[0], in_buf[1] );
					call_it(tableSum2_func )( tmp_max, 22, out_buf+50, in_buf[0], in_buf[1] );
					call_it(tableSum2_func )( tmp_max,  7, out_buf+50, in_buf[0], in_buf[1] );
					call_it(tableSum2_func )( tmp_max, 15, out_buf+50, in_buf[0], in_buf[1] );
					}
				call_it(tableSum2_func )( tmp_max, 31, out_buf,    in_buf[0], in_buf[1] );
				test_points = 31;
				test_value = tmp_max;
*/				break;
			case CHIRP:
					// Put a value in the buffer to guarantee it is being overwritten
					// by the chirping function.
				in_buf[1][400] = 123456789.0;

					// TEST_MIN_CHIRP_STEP 0.002776
					// chirps .5 bins with 128K point FFT
					// buf[0]=cx_DataArray, buf[1]=cx_ChirpDataArray
					// expects pairs of floats, so halve test_points
                if (user_i1 < 0) user_i1 -= 2; // (more negative)
                user_i1 = -user_i1;
                call_it(ChirpData_func)
                    ( (const sah_complex *)in_buf[0], (sah_complex *)in_buf[1]
                        , user_i1, user_d1*user_i1, test_points,   user_d2 );
				if (user_i1 < 0) user_i1 -= 2; // (more negative)
                user_i1 = -user_i1;
                call_it(ChirpData_func)
                    ( (const sah_complex *)in_buf[0], (sah_complex *)in_buf[1]
                        , user_i1, user_d1*user_i1, test_points,   user_d2 );

				test_value = 0;
				if( loops == 0 )
					memcpy( out_buf, in_buf[1], test_size);
				break;
			case PWRSPEC:
				call_it( GetPowerSpectrum_func )
				    ( (sah_complex *)out_buf, &orig_buf[NumDataPoints], test_points, 0, NumDataPoints / test_points);
				break;
            case PSONLY:
                call_it( PwrSpectrumOnly_func )
                    ( (sah_complex *)out_buf, out_buf, test_points);
                break;
            case TRNSTRP:
                call_it( TransposeStrip_func ) ( 16384, 64, 4, in_buf[0], out_buf);
                break;
			case PEAK:
				// buf[0]=fp_PoT, buf[1]=f_weight
				for(int i=0; i < 50; i++) {
					test_value = call_it( f_GetPeak_func)
						(in_buf[0], test_points/2, test_points/2-3, 1.2384, 897.654, in_buf[1]);
					}
				break;

			case CHI_SQ:
				// buf[0]=fp_PoT, buf[1]=f_weight
				float	xsq_null;
				test_value = call_it( f_GetChiSq_func)
					(in_buf[0], test_points/2, test_points/2-3, 1.2384, 897.654, in_buf[1]
					 ,128 ,&xsq_null);
				break;

			case TMEAN:
				//all we need to test here is f_sum
				break;

			case F_SUM:
//				for(int j=0; j< 10; j++)
//                   in_buf[0][j] = 1;
				float total;
				call_it( f_sum_func) (total, in_buf[0], 32768);	test_value = total;
				for(int i=0; i < 100; i++) {
					call_it( f_sum_func) (total, in_buf[0], 512);	test_value = total;
					}
				for(int i=0; i < 500; i++) {
					call_it( f_sum_func) (total, in_buf[0], 15);	test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 10);	test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 7);		test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 5);		test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 4);		test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 3);		test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 2);		test_value += total;
					call_it( f_sum_func) (total, in_buf[0], 1);		test_value += total;
					}
				break;
			case PULSE:
				// special case - switch in variety of summing/folding routine
				// pointers
				break;
			case _LOOP_TIME: break;
			}
		end_cyc = cycleCount();
		cruising_speed();
		if( end_cyc > cycles) {
			cycles = end_cyc-cycles;
			timing = std::min(cycles, timing);
			}
		}
	loop_count = loops;
	return timing;
    }

// =============================================================================
//    BENCH_C::verify_it
// =============================================================================
double BENCH_C::verify_it( int idx )
    {
	normal_verify = true;
	accuracy = 0;
    switch ( bench_list[idx].token )
        {
		case _FFT:	break;
        case SUM2_TBL:	break;
			break;
        case CHIRP:
			if (bench_list[idx].subtype == 1) FreeTrigArray( );
			break;
        case PWRSPEC:   break;
        case PSONLY:    break;
        case TRNSTRP:   break;
		case F_SUM:
        case CHI_SQ:
		case PEAK:
            if (test_value == 0.0) accuracy = fabs(orig_value);
			else accuracy = fabs(1-orig_value/test_value);
			normal_verify = false;
			break;
        case PULSE:     break;
        case TMEAN:     break;
		case _LOOP_TIME: break;
        }

	if(normal_verify) {
		double diff;
		int show_cnt = 0;
		int points = test_size/sizeof(float);
        for (int j=0; j < points; j++) {
			if (orig_buf[j] == out_buf[j]) continue;
			diff = orig_buf[j]-out_buf[j];
/*
			if( show_cnt < 20 && fabs(diff) > 1e-3) {
				show_cnt++;
				fprintf(stderr, " diff: [%d] o=%8.8g n=%8.8g\n", j,orig_buf[j], out_buf[j]);
				}
*/
//			if (orig_buf[j] == 0) accuracy += fabs(diff);
//			else accuracy += fabs(diff) / fabs(orig_buf[j]);

			accuracy += diff*diff;
//			if( diff > accuracy) accuracy = diff;	// pow(diff,2)
			}
		// multiply by return value diff - is VERY important - in summing,min, max
//		diff = orig_value-test_value;
//		accuracy *= (diff*diff)+1;
		}

	return accuracy;
    }

// =============================================================================
//    BENCH_C::report_it
// =============================================================================
void BENCH_C::report_it ( int idx )
	{
	static int last_token = -1;

	if( !show_benchmark )	return;
//	fprintf( stderr, "%30s: %9I64u x%2.2f %8.2g"
	fprintf( stderr, "%30s: %9llu x%2.2f %8.2g"	
		, bench_list[idx].name, timing
		, double(base_time)/double(timing) , accuracy);
	if( last_token != bench_list[idx].token )
		fprintf(stderr, " -- [ avg magnitude = %8.4f (%d)]", accu_sum, loop_count );
	else if(orig_value) fprintf(stderr, " t=%8.4f o=%8.4f", test_value, orig_value);
	last_token = bench_list[idx].token;
	fprintf(stderr, "\n");
	fflush(stderr);
	}

// =============================================================================
//    BENCH_C::save_best
// =============================================================================
void BENCH_C::save_best ( int best )
	{
    void    *func_p;
	if( show_benchmark ) {
		char buffer[60];
		sprintf(buffer, "%s[%s]", bench_list[base_func].name
			, (best != base_func ) ? bench_list[best].name : "original" );
		fprintf( stderr, "%30s: %9llu (chosen)\n------------------------------\n", buffer, best_time );
		}
    func_p = bench_list[best].theFunc;
	switch ( bench_list[best].token )
        {
//      case SUM2_TBL:	tableSum2 = (( tableSum2_func *) func_p); break;
        case CHIRP:
			canTrigArray = enough_ram( 67 );      // S@H usage with Tetsuji TrigArray
			if (bench_list[best].subtype == 1)
				{
				wantTrigArray = true;
				TrigArrayInit = ((TrigArrayInit_func *) &InitTrigArray);
				TrigArrayFree = ((TrigArrayFree_func *) &FreeTrigArray);
				}
			else wantTrigArray = false;
			ChirpData = (( ChirpData_func *) func_p);
			break;
        case PWRSPEC:   GetPowerSpectrum = (( GetPowerSpectrum_func *) func_p); break;
        case PSONLY:    PwrSpectrumOnly  = (( PwrSpectrumOnly_func *) func_p); break;
        case TRNSTRP:   TransposeStrip   = (( TransposeStrip_func *) func_p); break;
        case F_SUM:		f_sum            = (( f_sum_func *) func_p); break;
        case CHI_SQ:	f_GetChiSq       = (( f_GetChiSq_func *) func_p); break;
        case PEAK:		f_GetPeak        = (( f_GetPeak_func *) func_p); break;
		case _FFT:	break;
        }
	}


//=================================================
//
// optimize_init - Prepare the rest of the seti worker for optimized functions
//
// Return: whether we need chirp cache buffers
//=================================================

bool optimize_init( void )
    {
    CPU_INFO    our_cpu;
	BENCH       seti_bench;

    L1_cache	= our_cpu.cacheL1();	// Let rest of code know cache sizes
    L2_cache	= our_cpu.cacheL2();
	L1_line		= our_cpu.cacheLine();
    worker_cnt  = our_cpu.thread_cnt();	// How many cores/cpus share the Cache

	ChirpData   = &std_v_ChirpData;		// chirpfft's chirpdata
    f_GetPeak   = &std_f_GetPeak;		// gaussfit cpu slowdowns

    GetPowerSpectrum = &std_v_GetPowerSpectrum;
    PwrSpectrumOnly = &orig_v_GetPowerSpectrum;
    TransposeStrip   = &orig_v_Transpose4;
    f_sum       = &std_f_sum;
	tableSum2	= &std_tableSum2;
	tableSum3	= &std_tableSum3;
	tableSum4	= &std_tableSum4;
	tableSum5	= &std_tableSum5;
//
// FPU Optimized versions
//
	lower_rebin     = &std_rebin;       // find_pulse internal loops as function calls
    rms_deviation   = &std_rms_deviation;

	// TEST TIMING OF FUNCTIONS
//	if(show_benchmark) {
	seti_bench.bestFuncs(our_cpu);

//	extern int  verbose;
//	if(show_benchmark && !verbose)
//		exit(0);
//		}

//
// Original un-Optimized versions
//

/*
	f_sum       = &std_f_sum;
	tableSum2	= &opt_tableSum2;
	tableSum3	= &opt_tableSum3;
	tableSum4	= &opt_tableSum4;
	tableSum5	= &opt_tableSum5;

    ChirpData	= &opt_ChirpData;
    f_GetPeak	= &opt_f_GetPeak;
    f_GetChiSq	= &opt_f_GetChiSq;
*/
    if(our_cpu._3Dnow())
        {
    //  GetPowerSpectrum = &amd_v_GetPowerSpectrum;
    //  f_sum     = &amd_f_sum;
    //  tableSum2 = &amd_tableSum2;
        }

    if ( our_cpu.sse() )
        {
		CopyFoldSet( &Foldmain, &sse_ben_fold );
		if( wantTrigArray && !canTrigArray)
			ChirpData = &sse1_ChirpData_ak;
        }

    if ( our_cpu.sse2() )
        {
//      GetPowerSpectrum = &sse2_v_GetPowerSpectrum;
		if( wantTrigArray && !canTrigArray)
			ChirpData = &sse2_ChirpData_ak;
        }
    if ( our_cpu.sse3() )
        {
		if( wantTrigArray && !canTrigArray)
			ChirpData = &sse3_ChirpData_ak;
        }
	if( show_benchmark && wantTrigArray && !canTrigArray)
		fprintf( stderr, "Chosen Chirp replaced, insufficient memory.\n");

	return wantTrigArray && canTrigArray;	// Flag - do we need cached chirp buffers.
    }

// Tell caller if more than (needed) ram is available for current task
//  (factored by # of cpus and tasks being used)
bool	enough_ram( double needed ) {
	int	ncpus;
	double sys_memory = app_init_data.host_info.m_nbytes;

	ncpus = app_init_data.global_prefs.max_cpus;
	ncpus = (ncpus == 0) ? os_cpu_count() : std::min(ncpus , os_cpu_count() );
	needed *=  1024.0 * 1024.0 * ncpus;   // for apps
	needed +=  1024.0 * 1024.0 * 30.0;    // for OS, BOINC, etc.

	return ( sys_memory > 0) ? (needed <= sys_memory) : true ;
	}


//----------------------------------------------
//  global values used by other parts of the seti code
//

int L1_cache;       // Let rest of code know cache sizes
int L2_cache;
int L1_line;
int worker_cnt;  // How many cores/cpus share the Cache
double force_use;
bool show_benchmark = false;
