// cpuid_x86.h
//
//

#include "my_types.h"
#include "opt_os_interface.h"

//---------- Constants ---------//


/*	__SSE__		
	UNPCKHPS	(CAca) = (AaBb, CcDd)
	UNPCKLPS	(Dbdb) = (AaBb, CcDd)

	MOVHPS		(MmBb) = (AaBb, Mm)
	MOVLPS		(AaMm) = (AaBb, Mm)
	MOVHLPS		(AaDd) = (AaBb, CcDd)
	MOVLHPS		(CcBb) = (AaBb, CcDd)

	__3DNow__
	+PFNACC		([b-B][a-A])	= (Aa, Bb)
	+PFPNACC 	([b+B][a-A])	= (Aa, Bb)
	+PSWAPD	 	(bB)			= (Aa, Bb)
	
	PFACC 	 	([A+a][B+b])	= (Aa, Bb)
	PFADD 	 	([A+B][a+b])	= (Aa, Bb)
	PFMUL	 	([A*B][a*b])	= (Aa, Bb)
	PFSUB	 	([A-B][a-b])	= (Aa, Bb)
	PFSUBR	 	([B-A][B-A])	= (Aa, Bb)
	PFMAX	 	([max(A,B)][max(a,b)])	= (Aa, Bb)
	PFSUB	 	([A-B][a-b])	= (Aa, Bb)
	OPCODE	 	(xx)	= (Aa, Bb)
*/

//---------- SSE macros ---------//

//  s_get(p)  - GET 4 floats from unaligned address
//  s_put(p)	- PUT 4 floats at unaligned address
//  FILLsim(a,b,c,d) - GET 4 floats directly into register
//  s_add(a,b)	- Add 4 floats to another 4 floats, return result - Alignment or registers required
//  s_div(a,b)	- Divide 4 floats in a by those in b
//  s_max(a,b)	- Return the 4 highest floats between a and b

// && (defined(__i386__) || defined(__x86_64__))

//---------- Prototypes ---------//

//*****
//
// cycleCount - Machine specific function to get cycle counter
//
extern uint64 cycleCount( void);

extern uint64 cycleCountFix;


//---------- Classes ---------//


typedef struct {
	int16	size, line;
	int		read_speed[3];
	int		write_speed[3];
	int		copy_speed[3];
	} CACHE_LEVEL_;

typedef struct {
	CACHE_LEVEL_	L1, L2, L3, ram;
	int16	L1code;
	bool	shared_L1_L2, exclusive;
	} CACHE_INFO;


// Lookup structure for all processors
//  identify first entry in table that is equal or beyond current
//  CPU stats
//
typedef struct {
	int		famModl:15;
	int		brand  :17;
	int		cacheL2:16;
	char *norm_name;
	char *generation;
	char *core;
	char *notes;
	} CPUID_DEFS;

extern CPUID_DEFS cpuid_tbl[];



// =============================================================================
//
// class CPU_INFO_t
// 
// =============================================================================
class CPU_INFO_t {
	public:
		CPU_INFO_t() { initialize(); };
		~CPU_INFO_t() {};

		CACHE_INFO & getCache(void) { return cache; };
		int		cacheL1()		{ return cache.L1.size * 1024; };
		int		codeL1()		{ return cache.L1code * 1024; };
		int		cacheL2()		{ return cache.L2.size * 1024; };
		int		cacheL3()		{ return cache.L3.size * 1024; };
		int		cacheLine()		{ return cache.L2.line; };
		bool	mmx(void)		{ return has_mmx; };
		bool	mmxPlus(void)	{ return has_mmxPlus; };
		bool	_3Dnow(void)	{ return has_3Dnow; };
		bool	_3DnowPlus(void){ return has_3DnowPlus; };
		bool	sse(void)		{ return has_sse; };
		bool	sse2(void)		{ return has_sse2; };
		bool	sse3(void)		{ return has_sse3; };
		bool	ssse3(void)		{ return has_ssse3; };
		bool	could64(void)	{ return testxBit(1, edx_, 29); };

		bool	testBit(uint32 flags, int shift)
			{return ( (flags>>shift)& 1) != 0;			};
		bool	testBit(int x, int y, int shift)
			{return ((cpuid[x][y]>>shift)& 1) != 0;	};
		bool	testxBit(int x, int y, int shift)
			{return ((xcpuid[x][y]>>shift)& 1) != 0;	};

		char	*vendor_name(void)	 { return vendor.abrvName; };
		char	*company(void)	 { return vendor.company; };
		char	*vendor_id(void) { return vendor_str; };
		char	*cpuid_name(void){ return (char *) &xcpuid[2][0]; };

		char	*generation(void){ return names->generation; };
		char	*model(void)	{ return names->norm_name; };
		char	*core(void)		{ return names->core; };
		char	*notes(void)	{ return cpu_notes; };

		int		cpu_cnt(void)	{ return cpu_count; };
		int		core_cnt(void)	{ return core_count; };
		int		thread_cnt(void){ return thread_count; };

		int		s_family(void)	{ return b_family; };
		int		s_model(void)	{ return b_model; };
		int		s_rev(void)		{ return rev; };
		int		brand(void)		{ return brandID; };

		int		x_family(void)	{ return extFamily; };
		int		x_model(void)	{ return extModel; };
		int		x_brand(void)	{ return extBrand; };

		float	cpu_throttle(void);
		int		cpu_mhz(void);
		int		L1_speed(void);
		int		L2_speed(void);
		int		L3_speed(void);
		int		RAM_speed(void);

		uint32	cpuid_(int x, int y){ return cpuid[x][y]; }
		uint32	xcpuid_(int x, int y){ return xcpuid[x][y]; }

	/* static  const */ enum { eax_=0,ebx_,ecx_,edx_ }; // Hans Dorn, Oct-05-2006

private:
		int		brandID;
		int		b_family;	//	3 - 386 family, 4 - i486 family, 5 - Pentium family, 6 - Pentium Pro family
		int		b_model;	//
		int		rev;
		int		type;

		int		extBrand;
		int		extFamily;
		int		extModel;

		int		vendorID;

		bool	has_cpuid_opcode;
		bool	has_mmx;
		bool	has_mmxPlus;
		bool	has_3Dnow;
		bool	has_3DnowPlus;
				// These are true only if could_sse and has_sseOS
		bool	has_sseOS;
		bool	has_sse, has_sse2, has_sse3, has_ssse3;
		bool	could_sse, could_sse2, could_sse3, could_ssse3;
	
		int		max_stdFunc;
		int		max_extFunc;

		uint64	mhz_start, mhz_end, tick_start, tick_end;
		uint32	cpuMhz;
		uint32	busMhz;
		
		int		cpu_count, core_count, thread_count;

		CACHE_INFO	cache;
		CPUID_DEFS	*names;
		struct	{
				char	*abrvName;
				char	*company;
			} vendor;
		char	vendor_str[14];
		char	cpu_notes[48];

		uint32	cpuid[20][4];
		uint32	(*xcpuid)[4];

	private:
		void	initialize(void);
		void	checkCpuidOpcode(void);	// Use generic code to detect w/o causing bad opcode
		void	getCacheSizes(void);
		bool	sseOSsupport(void);
		void	cpuidLookup(void);
		void	vendorLookup(void);
		void	cpuidFixNames(void);
		int		cache_speed( CACHE_LEVEL_ &level );
		void	memory_testing(void);
	};				
			
typedef class CPU_INFO_t CPU_INFO;



//
// x86 CPU SPECIFIC functions to identify capabilities
//

#define EXT_CMD 0x80000000

enum	{
	_INTEL=1, _AMD, _CYRIX, _UMC, _NEXGEN, _RISE, _SIS, _TMT, _CENTAUR, _NSC
	};


#if defined(_MSC_VER) && !defined(_ICC)


__forceinline uint32 CPUID_X86
	(uint32 command, uint32 &_eax, uint32 &_ebx, uint32 &_ecx, uint32 &_edx)
	{
	uint32	a,b,c,d;
	__asm {
		push	ecx
	    mov eax, [command]
		xor	ecx,ecx
		cpuid			; <<CPUID>>
		mov	[a],eax
		mov	[b],ebx
		mov	[c],ecx
		mov	[d],edx
	}
	_eax = a, _ebx = b, _ecx = c, _edx = d; 
	return _eax;
	}

__forceinline void CPUID_X86
	(uint32 command, uint32 *buffer)
	{
	__asm {
	    mov eax, [command]
		xor	ecx,ecx
		cpuid			; <<CPUID>>
		push	edx
		mov		edx,[buffer]
		mov		[edx+ 0],eax
		mov		[edx+ 4],ebx
		mov		[edx+ 8],ecx
		pop		eax
		mov		[edx+12],eax
		}
	}

__forceinline void CPUID_X86
	(uint32 command, uint32 cx_val, uint32 *buffer)
	{
	__asm {
	    mov eax, [command]
	    mov ecx, [cx_val]
		cpuid			; <<CPUID>>
		push	edx
		mov		edx,[buffer]
		mov		[edx+ 0],eax
		mov		[edx+ 4],ebx
		mov		[edx+ 8],ecx
		pop		eax
		mov		[edx+12],eax
		}
	}

/*
*/
__forceinline bool  can_cpuid(void)
	{
	register uint32 result;
	__asm {
		pushfd						//
		pushfd						//
		pop     eax                 // store EFLAGS in eax.
		xor     eax, 0x200000       // switch bit 21.
		push    eax                 // copy "changed" value to stack.
		popfd                       // save "changed" eax to EFLAGS.
		pushfd						// restore original flags.
		pop     edx					// 
		popfd                       // restore original flags
		xor     eax, edx            //   See if bit changeable.
		mov     result, eax			//   Save the value in eax to a variable.
		}
	return result != 0;
	}

// Use CPUID to stop the instruction pipeline and get starting cycle
//
//  Accurate cycle counting on the pentium 4 CPU requires an issue of a CPUID 
//   instruction before getting the cycle counter.
//   This causes a long delay as CPUID is a CPU serialization function
//  (it stops all pipelining...clears out of order execution queues...etc.)
//
__forceinline uint64 cycleCount( void)
	{
	uint64 cycles;
	__asm {
		push ebx
		push ecx
		xor	eax,eax
		cpuid
		rdtsc
		mov	DWORD PTR [cycles+0],eax
		mov	DWORD PTR [cycles+4],edx
		pop	ecx
		pop	ebx
	}	
	return (cycles);
}

// Now the GNU C++ compiler versions
//
#elif defined(__GNUC__) || defined(_ICC)

static inline bool can_cpuid(void)
{
	register uint32 one, two;
    __asm ( \
		"pushfl \n" \
		"pushfl \n" \
		"popl	%0\n" \
		"xorl	%0, $0x0200000\n" \
		"pushl	%0\n" \
		"popfl\n" \
		"pushfl\n" \
		"popl	%1\n" \
		"popfl\n" \
		: "=a" (one), "=b" (two) /* output */ );
    return (one != two);
}


inline uint32 CPUID_X86 
	(uint32 cmd, uint32 *buffer)
	{
    __asm ("cpuid" \
	     : "=a" (buffer[0]), "=b" (buffer[1]), "=c" (buffer[2]), "=d" (buffer[3]) \
	     : "a" (cmd) );
	return buffer[0];
	}

inline uint32 CPUID_X86 
	(uint32 cmd, uint32 cx_val, uint32 *buffer)
	{
    __asm ("cpuid" \
	     : "=a" (buffer[0]), "=b" (buffer[1]), "=c" (buffer[2]), "=d" (buffer[3]) \
	     : "a" (cmd), "c" (cx_val) );
	return buffer[0];
	}


inline uint32 CPUID_X86 
	(uint32 cmd, uint32 &eax, uint32 &ebx, uint32 &ecx, uint32 &edx)
	{
    __asm ("cpuid" \
	     : "=a" (eax), "=b" (ebx), "=c" (ecx), "=d" (edx) \
	     : "a" (cmd) );
	return eax;
	}

inline uint64 cycleCount( void) {
    uint64 cycles;
		// Use CPUID to stop the instruction pipeline and get starting cycle
    __asm volatile ("cpuid \n rdtsc" : "=A" (cycles) : "a" (0) : "%ebx", "%ecx" );
    return (cycles);
}

#endif



// Lookup structure for AMD processor "performance" equivalent number
//   given Mhz, and L2 cache size
//
typedef struct {
	short	L2Cache, mhz, perfNum;
	}
	AMD_PERFTBL;
extern AMD_PERFTBL perfTbl[];

// Lookup structure for cache type on earlier Pentium models
//
typedef struct {
		uint8	idByte, line;
		int16	code, data, L2, L3;
	} _cacheCodes;
extern _cacheCodes cacheCodes[];


