/***************
 *
 *	memspeed.cpp:	x86 CPU specific speed testing functions
 *	Originator:		Ben Herndon
 *
 *   to be used with seti @ home client code
 *
 *   Compile this module for SSE2 or better
 */
#if defined(__GNUC__)
#include <algorithm>
#else
#include <stdlib.h>
#endif

#include "my_types.h"
#include "opcodes_SSE2.hpp"
#include "memspeed.h"
#include "opt_os_interface.h"
#include "cpuid_x86.h"

//-------------------------------------------------------------
//
//  Block read & write speed testing functions
//
//
// memspeed - The "testing tester" - times the calling function times
//
int memspeed(void *ptr, int len, int repetitions, bool cflush, mem_func *test_mem)
	{
	uint64	overhead, ticks, min_ticks;
	int		loops,	timing;
	int		bytes = len * 1024;

	// Maximum CPU time for this process, don't worry we're quick
	ramming_speed();

	// Time loop overhead
	overhead = (1<<30);
	loops = 10;
	do	{
		ticks = cycleCount();
		test_mem(ptr, 0 );
		ticks = cycleCount()-ticks;
		overhead = std::min(overhead, ticks);
		} while( --loops );

	// Prime cache(s) of whatever memory type
	if(!cflush)
		test_mem(ptr, bytes );

	// Compute total time + adjust for overhead
	min_ticks = (1<<30);
	loops = repetitions;
	do	{
		// Clear cache (if instructed to do so)
		// - only usefull for RAM speed tests 
		// - Also flush size might be 64 but we don't want to pass that info so
		//  assume 32
		if(cflush) {
			char *chptr = (char *) ptr;
			for(int i=0; i < bytes; i += 32)
				s_flush( &chptr[i] );
			s_fence_all();	// Wait Make all the flushes "real"
			}

		// Time the access speed
		ticks = cycleCount();
		test_mem(ptr, bytes );
		if(cflush)	s_fence_all();	// Wait Make all reads/writes "real"
		ticks = cycleCount() - ticks - overhead;
		min_ticks = std::min(min_ticks, ticks);
		} while( --loops );

	// Normal CPU usage
	cruising_speed();

	// convert to Cycles/ 100 kilobytes (increase resolution)
	timing = (int32) min_ticks * 100 / len;

	return timing;
	}


//-------------
// read_ints
//
void read_ints(void *ptr, int length)
	{
	int32	sum1, sum2;
	int32	*ptr32 = (int32 *)ptr;
	volatile int32	*vptr = ptr32;
	int int_cnt = length / sizeof( int32 );

	const int stride = 8;
	for(int j = 0; j < int_cnt ; j += stride )
		{
		sum1 += ptr32[j+0];
		sum2 += ptr32[j+1];
		sum1 += ptr32[j+2];
		sum2 += ptr32[j+3];
		sum1 += ptr32[j+4];
		sum2 += ptr32[j+5];
		sum1 += ptr32[j+6];
		sum2 += ptr32[j+7];
		}
	// since vptr is volatile, forces compiler to do the adds above
	*vptr = sum1 + sum2;
	}

//-------------
// read_mmx
//
void read_mmx(void *ptr, int length)
	{
	__m64	sum1, sum2;
	__m64	*x_ptr = (__m64 *)ptr;
	volatile int *vol_ptr = (int *)x_ptr;
	union { __m64 i64; int i32[2]; } sum3;
	int mmx_cnt = length / sizeof( __m64 );

	_mm_empty();
	const int stride = 8;
	for(int j = 0; j < mmx_cnt ; j += stride )
		{
		sum1 = _mm_add_pi32(sum1, x_ptr[j+0]);
		sum2 = _mm_add_pi32(sum2, x_ptr[j+1]);
		sum1 = _mm_add_pi32(sum1, x_ptr[j+2]);
		sum2 = _mm_add_pi32(sum2, x_ptr[j+3]);
		sum1 = _mm_add_pi32(sum1, x_ptr[j+4]);
		sum2 = _mm_add_pi32(sum2, x_ptr[j+5]);
		sum1 = _mm_add_pi32(sum1, x_ptr[j+6]);
		sum2 = _mm_add_pi32(sum2, x_ptr[j+7]);
		}
	// since vol_ptr is volatile, forces compiler to do the adds above
	sum3.i64 = _mm_add_pi32(sum1, sum2);
	*vol_ptr = sum3.i32[0];
	_mm_empty();
	}

//-------------
// read_sse
//
void read_sse(void *ptr, int length)
	{
	XMM	sum1, sum2;
	XMM	*x_ptr = (XMM *)ptr;
	volatile float	*vol_ptr = (float *)x_ptr;
	float	sum3;
	int sse_cnt = length / sizeof( XMM );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		sum1 = s_add(sum1, x_ptr[j+0]);
		sum2 = s_add(sum2, x_ptr[j+1]);
		sum1 = s_add(sum1, x_ptr[j+2]);
		sum2 = s_add(sum2, x_ptr[j+3]);
		sum1 = s_add(sum1, x_ptr[j+4]);
		sum2 = s_add(sum2, x_ptr[j+5]);
		sum1 = s_add(sum1, x_ptr[j+6]);
		sum2 = s_add(sum2, x_ptr[j+7]);
		}
	// since vol_ptr is volatile, forces compiler to do the adds above
	s_put1(&sum3, s_add(sum1, sum2));
	*vol_ptr = sum3;
	}

//-------------
// read_sse_U - Unaligned
//
void read_sse_U(void *ptr, int length)
	{
	XMM	sum1, sum2, tmp1, tmp2;
	XMM	*x_ptr = (XMM *)(((char *)ptr)+1);
	volatile float	*vol_ptr = (float *)x_ptr;
	float	sum3;
	int sse_cnt = length / sizeof( XMM );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		s_getU(tmp1, &x_ptr[j+0] );
		s_getU(tmp2, &x_ptr[j+1] );
		sum1 = s_add(sum1, tmp1);
		sum2 = s_add(sum2, tmp2);
		s_getU(tmp1, &x_ptr[j+2] );
		s_getU(tmp2, &x_ptr[j+3] );
		sum1 = s_add(sum1, tmp1);
		sum2 = s_add(sum2, tmp2);
		s_getU(tmp1, &x_ptr[j+4] );
		s_getU(tmp2, &x_ptr[j+5] );
		sum1 = s_add(sum1, tmp1);
		sum2 = s_add(sum2, tmp2);
		s_getU(tmp1, &x_ptr[j+6] );
		s_getU(tmp2, &x_ptr[j+7] );
		sum1 = s_add(sum1, tmp1);
		sum2 = s_add(sum2, tmp2);
		}
	// since vol_ptr is volatile, forces compiler to do the adds above
	s_put1(&sum3, s_add(sum1, sum2));
	*vol_ptr = sum3;
	}

//-------------
// read_cache32 - Precache w/line size of 32
//
void read_cache32(void *ptr, int length)
	{
	XMM	sum1, sum2;
	XMM	*x_ptr = (XMM *)ptr;
	volatile float	*vol_ptr = (float *)x_ptr;
	float	sum3;
	int sse_cnt = length / sizeof( XMM );

	s_fetch( &x_ptr[0] );
	s_fetch( &x_ptr[2] );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		s_fetch( &x_ptr[j+16+0] );
		sum1 = s_add(sum1, x_ptr[j+0]);
		sum2 = s_add(sum2, x_ptr[j+1]);
		s_fetch( &x_ptr[j+16+2] );	// Prefetch +32 bytes (for P3s w 32byte line size)
		sum1 = s_add(sum1, x_ptr[j+2]);
		sum2 = s_add(sum2, x_ptr[j+3]);
		s_fetch( &x_ptr[j+16+4] );	// Prefetch +32 bytes (for P3s w 32byte line size)
		sum1 = s_add(sum1, x_ptr[j+4]);
		sum2 = s_add(sum2, x_ptr[j+5]);
		s_fetch( &x_ptr[j+16+6] );	// Prefetch +32 bytes (for P3s w 32byte line size)
		sum1 = s_add(sum1, x_ptr[j+6]);
		sum2 = s_add(sum2, x_ptr[j+7]);
		}
	// since vol_ptr is volatile, forces compiler to do the adds above
	s_put1(&sum3, s_add(sum1, sum2));
	*vol_ptr = sum3;
	}


//-------------
// read_cache64 - Precache w/line size of 64
//
void read_cache64(void *ptr, int length)
	{
	XMM	sum1, sum2;
	XMM	*x_ptr = (XMM *)ptr;
	volatile float	*vol_ptr = (float *)x_ptr;
	float	sum3;
	int sse_cnt = length / sizeof( XMM );

	s_fetch( &x_ptr[0] );
	s_fetch( &x_ptr[4] );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		s_fetch( &x_ptr[j+16+0] );
		sum1 = s_add(sum1, x_ptr[j+0]);
		sum2 = s_add(sum2, x_ptr[j+1]);
		sum1 = s_add(sum1, x_ptr[j+2]);
		sum2 = s_add(sum2, x_ptr[j+3]);
		s_fetch( &x_ptr[j+16+4] );
		sum1 = s_add(sum1, x_ptr[j+4]);
		sum2 = s_add(sum2, x_ptr[j+5]);
		sum1 = s_add(sum1, x_ptr[j+6]);
		sum2 = s_add(sum2, x_ptr[j+7]);
		}
	// since vol_ptr is volatile, forces compiler to do the adds above
	s_put1(&sum3, s_add(sum1, sum2));
	*vol_ptr = sum3;
	}

//-------------
// write_ints
//
void write_ints(void *ptr, int length)
	{
	int32	sum1;
	int32	*ptr32 = (int32 *)ptr;
	int int_cnt = length / sizeof( int32 );

	sum1 = 0;
	const int stride = 8;
	for(int j = 0; j < int_cnt ; j += stride )
		{
		ptr32[j+0] = sum1;
		ptr32[j+1] = sum1;
		ptr32[j+2] = sum1;
		ptr32[j+3] = sum1;
		ptr32[j+4] = sum1;
		ptr32[j+5] = sum1;
		ptr32[j+6] = sum1;
		ptr32[j+7] = sum1;
		}
	}

//-------------
// write_mmx
//
void write_mmx(void *ptr, int length)
	{
	__m64	sum1;
	__m64	*x_ptr = (__m64 *)ptr;
	int mmx_cnt = length / sizeof( __m64 );

	_mm_empty();
	const int stride = 8;
	for(int j = 0; j < mmx_cnt ; j += stride )
		{
		x_ptr[j+0] = sum1;
		x_ptr[j+1] = sum1;
		x_ptr[j+2] = sum1;
		x_ptr[j+3] = sum1;
		x_ptr[j+4] = sum1;
		x_ptr[j+5] = sum1;
		x_ptr[j+6] = sum1;
		x_ptr[j+7] = sum1;
		}
	_mm_empty();
	}


//-------------
// write_mmx_nt
//
void write_mmx_nt(void *ptr, int length)
	{
	__m64	sum1;
	__m64	*x_ptr = (__m64 *)ptr;
	int mmx_cnt = length / sizeof( __m64 );

	_mm_empty();
	const int stride = 8;
	for(int j = 0; j < mmx_cnt ; j += stride )
		{
		_mm_stream_pi(&x_ptr[j+0], sum1);
		_mm_stream_pi(&x_ptr[j+1], sum1);
		_mm_stream_pi(&x_ptr[j+2], sum1);
		_mm_stream_pi(&x_ptr[j+3], sum1);
		_mm_stream_pi(&x_ptr[j+4], sum1);
		_mm_stream_pi(&x_ptr[j+5], sum1);
		_mm_stream_pi(&x_ptr[j+6], sum1);
		_mm_stream_pi(&x_ptr[j+7], sum1);
		}
	_mm_empty();
	s_fence_writes();
	}

//-------------
// write_sse
//
void write_sse(void *ptr, int length)
	{
	XMM	sum1;
	XMM	*x_ptr = (XMM *)ptr;
	int sse_cnt = length / sizeof( XMM );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		x_ptr[j+0] = sum1;
		x_ptr[j+1] = sum1;
		x_ptr[j+2] = sum1;
		x_ptr[j+3] = sum1;
		x_ptr[j+4] = sum1;
		x_ptr[j+5] = sum1;
		x_ptr[j+6] = sum1;
		x_ptr[j+7] = sum1;
		}
	s_fence_writes();
	}

//-------------
// write_sse_nt - Non-temporal write test
//
void write_sse_nt(void *ptr, int length)
	{
	XMM	sum1;
	XMM	*x_ptr = (XMM *)ptr;
	int sse_cnt = length / sizeof( XMM );

	const int stride = 8;
	for(int j = 0; j < sse_cnt ; j += stride )
		{
		s_putA_nc((float*)&x_ptr[j+0], sum1);
		s_putA_nc((float*)&x_ptr[j+1], sum1);
		s_putA_nc((float*)&x_ptr[j+2], sum1);
		s_putA_nc((float*)&x_ptr[j+3], sum1);
		s_putA_nc((float*)&x_ptr[j+4], sum1);
		s_putA_nc((float*)&x_ptr[j+5], sum1);
		s_putA_nc((float*)&x_ptr[j+6], sum1);
		s_putA_nc((float*)&x_ptr[j+7], sum1);
		}
	s_fence_writes();
	}
