/* name: opt_FPU.cpp
 *
 * Purpose:  test bench code for seti@home client 
 * Originator: Ben Herndon 

 * Optimized functions for any C++ or FPU (Floating point Processor Unit)
 *
 * This file should not be compiled with any compiler SIMD enabling options
 *  as these functions serve as a baseline for optimzation improvement 
 *  timing tests
 *
 */
#include "stdafx.h"
#include "mathimf.h"
#include "chirpfft.h"
#include "analyze.h"

#undef max
#define max(a,b) ((a>b)?a:b)




// opt_v_GetPowerSpectrum 
//  8-10-06 BENH	- Unrolled loop 3 times to allow for FPU latency & reduce loop overhead 
// xx-xx-03 ERICK	- Created function
//
GetPowerSpectrum_ptt( opt_v_GetPowerSpectrum )
    {
	int	i, bin;
	float *workBuf = (float *)FreqData;

	ALIGNED_YES( FreqData );
	ALIGNED_YES( PowerSpectrum );
	for	( i	= 0, bin = 0; i < this_fft_len; i++, bin += bin_len)
		{
		float psNum = FreqData[i][0] * FreqData[i][0] + FreqData[i][1] * FreqData[i][1];
		PowerSpectrum[bin_off + bin] =	// Large cache miss here...can it be fixed?
		workBuf[i] = psNum;
		}
    }



// opt_v_ChirpData 
// 8-10-06 BENH - Unrolled loop 3 times to allow for FPU latency & reduce loop overhead 
//				- Removed conditionals from loop 
// xx-xx-03 ERICK -  Created function
//
ChirpData_ptt( opt_ChirpData )
    {
    float   chirp_sign; // BENH - Pulled conditional out of loop
    if ( chirp_rate_ind == 0 )
        {
        memcpy( cx_ChirpDataArray, cx_DataArray,  int( ul_NumDataPoints * sizeof(sah_complex)) );  
        return ( 0 );
        }

    // calculate trigonometric array this function returns w/o doing
    // anything when sign of chirp_rate_ind reverses. so we have to take care
    // of it.
    //
    chirp_sign = ( chirp_rate_ind >= 0 ) ? 1 : -1;

	const int stride = 2;
    int i = 0;
    int last = ul_NumDataPoints - ( stride - 1 );
    if ( CacheChirpCalc )
        {
#define BLOCKING 0

#if BLOCKING
		double old_rate = CurrentChirpRateInd;
		for(int outer = i; outer < last; outer += 1000)
			{
			int	end = i + 1000;
			if(end > last) end = last;

				// Each time we must fool 'Calc' that we are doing loop 1st time
			CurrentChirpRateInd = old_rate;
			CalcTrigArray(i, end-i, chirp_rate_ind );
			for (		; i < end ; i += stride )
#else

		CalcTrigArray(0, ul_NumDataPoints, chirp_rate_ind );
		for (		; i < last ; i += stride )
#endif
			{
			register double c1, c2, d1, d2;
			register float  R1, R2, I1, I2, t1, t2;

			c1 = CurrentTrig[i + 0].Cos;
			d1 = CurrentTrig[i + 0].Sin * chirp_sign;   // BENH - opt - avoid branch by multiply
			c2 = CurrentTrig[i + 1].Cos;
			d2 = CurrentTrig[i + 1].Sin * chirp_sign;

			// d = (chirp_rate_ind >0)? CurrentTrig[i].Sin :
			// -CurrentTrig[i].Sin; Sometimes chirping is done in place. We
			// don't want to overwrite data prematurely.
			//
			R1 = cx_DataArray[i + 0][0] * c1;
			t1 = cx_DataArray[i + 0][1] * d1;
			I1 = cx_DataArray[i + 0][0] * d1;
			t2 = cx_DataArray[i + 0][1] * c1;
			cx_ChirpDataArray[i + 0][0] = R1 - t1;
			cx_ChirpDataArray[i + 0][1] = I1 + t2;
			R2 = cx_DataArray[i + 1][0] * c2;
			t1 = cx_DataArray[i + 1][1] * d2;
			I2 = cx_DataArray[i + 1][0] * d2;
			t2 = cx_DataArray[i + 1][1] * c2;
			cx_ChirpDataArray[i + 1][0] = R2 - t1;
			cx_ChirpDataArray[i + 1][1] = I2 + t2;
			}
		}
#if BLOCKING
	}
#endif

//	cx_ChirpDataArray[64][1] = 3.012345;

    // Too little memory to cache sin/cos or just falling out of
    // 'CacheChirpCalc' loop
	if( i < ul_NumDataPoints)
		{  // use original routine to finish up any tailings (max stride-1 elements)
		orig_ChirpData(cx_DataArray+i, cx_ChirpDataArray+i
			, chirp_rate_ind, chirp_rate, ul_NumDataPoints-i, sample_rate);
		}

    count_flops( 12 * ul_NumDataPoints );
    return ( 0 );
    }

/* Loop invariance calculations: 
 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 {									// Original equation
	time  = j/sample_rate;
	ang   = M_PI*2*chirp_rate*time*time;
 ...
 invar = M_PI*2*chirp_rate;			// 1st invariance
 {
	time  = j/sample_rate;
	ang   = invar * time * time;
 ...
 invar = M_PI*2*chirp_rate;			// 2nd 
 {
	time  = j * (1/sample_rate);	//(same as j/sample_rate)
	ang   = invar *  time * time;
 ...
 invar = M_PI*2*chirp_rate * (1/sample_rate) * (1/sample_rate);
 {
	time = j;
	 ang = invar * time* time;		// 3rd & final
 */

// =============================================================================
// sum2tables
//
//  Sum then average the values from two tables, store results in third
//  This is the most called routine of the summing, and should have most
//  optimization
//
tableSum2_ptt( opt_tableSum2 )
	{
	int i;
	double	sum1, sum2, divNum=1/2.0, max2, max1;
	max2 = max1 = 0.0;
   	for (i = 0 ; i < length-1; i += 2) {
		//  FPU pipeline #1					FPU pipeline #2
		sum1 = ptr1[i+0] + ptr2[i+0];	sum2 = ptr1[i+1] + ptr2[i+1];
		sum1 *= divNum;					sum2 *= divNum;
		sums[i+0] = sum1;				sums[i+1] = sum2;
		max1 = max(max1, sum1);			max2 = max(max2, sum2);
		}
   	for (   ; i < length; i++) {
		sum1 = (ptr1[i] + ptr2[i]) * divNum;
		sums[i] = sum1;
		max1 = max(max1, sum1);
		}
	f_max = max(max1, max2);
    }

//
// sum3tables
//
tableSum3_ptt( opt_tableSum3 )
	{
	double	sum1, sum2, divNum=1/3.0, max2, max1;
	int i;
	max2 = max1 = 0;
   	for (i = 0 ; i < length-1; i += 2) {
		sum1 = ptr1[i+0] + ptr2[i+0];	sum2 = ptr1[i+1] + ptr2[i+1];
		sum1 += ptr3[i+0];				sum2 += ptr3[i+1];
   		sum1 *= divNum;			   		sum2 *= divNum;
		sums[i+0] = sum1;				sums[i+1] = sum2;
		max1 = max(max1, sum1);			max2 = max(max2, sum2);
		}
   	for (  ; i < length; i++) {
		sum1 = (ptr1[i] + ptr2[i] + ptr3[i]  ) * divNum;
		sums[i] = sum1;
		max1 = max(max1, sum1);
		}
	return max(max1, max2);
    }

//
// sum4tables
//
tableSum4_ptt( opt_tableSum4 )
	{
	double	sum1, sum2, divNum=1/4.0, max2, max1;
	int i;
	max2 = max1 = 0;
   	for (i = 0 ; i < length-1; i += 2) {
		sum1 = ptr1[i+0] + ptr2[i+0];	sum2 = ptr1[i+1] + ptr2[i+1];
		sum1 += ptr3[i+0];				sum2 += ptr3[i+1];
		sum1 += ptr4[i+0];				sum2 += ptr4[i+1];
   		sum1 *= divNum;			   		sum2 *= divNum;
		sums[i+0] = sum1;				sums[i+1] = sum2;
		max1 = max(max1, sum1);			max2 = max(max2, sum2);
		}
   	for (  ; i < length; i++) {
		sum1 = (ptr1[i] + ptr2[i] + ptr3[i]  ) * divNum;
		sums[i] = sum1;
		max1 = max(max1, sum1);
		}
	return max(max1, max2);
    }

//
// sum5tables
//
tableSum5_ptt( opt_tableSum5 )
	{
	double	sum1, sum2, divNum=1/5.0, max2, max1;
	int i;
	max2 = max1 = 0;
   	for (i = 0 ; i < length-1; i += 2) {
		sum1 = ptr1[i+0] + ptr2[i+0];	sum2 = ptr1[i+1] + ptr2[i+1];
		sum1 += ptr3[i+0];				sum2 += ptr3[i+1];
		sum1 += ptr4[i+0];				sum2 += ptr4[i+1];
		sum1 += ptr5[i+0];				sum2 += ptr5[i+1];
   		sum1 *= divNum;			   		sum2 *= divNum;
		sums[i+0] = sum1;				sums[i+1] = sum2;
		max1 = max(max1, sum1);			max2 = max(max2, sum2);
		}
   	for (  ; i < length; i++) {
		sum1 = (ptr1[i] + ptr2[i] + ptr3[i]  ) * divNum;
		sums[i] = sum1;
		max1 = max(max1, sum1);
		}
	return max(max1, max2);
    }

//
// opt_f_sum - Sum a table of numbers
//
f_sum_ptt( opt_f_sum )
	{
	float	sum1, sum2, sum3, sum4;
	const int stride = 4;
	sum1 = sum2 = sum3 = sum4 = 0;
	int i;
	for	(i = length; i > stride-1; i -= stride )
		{
		sum1 += buf[i-1];
		sum2 += buf[i-2];
		sum3 += buf[i-3];
		sum4 += buf[i-4];
		}

	for	(  ; i  ; i-- )
		sum1 += buf[i-1];

	f_total = sum1 +sum2 + sum3 + sum4;
	}



// =============================================================================
//  f_GetChiSq  - see gausfit.cpp for full comments
//
f_GetChiSq_ptt( opt_f_GetChiSq )
	{
	int     i;
    float   f_ChiSq, f_null_hyp, f_PredictedPower, f_tot_weight;

    f_ChiSq    = 0;
    f_null_hyp = 0;
    f_tot_weight = 0;

// BENH: hoisted some variables out of the loop, compiler might have..but be sure
//
// Original lines
//      f_PredictedPower = f_MeanPower + f_PeakPower * f_weight[abs( i - ul_TOffset )];
//      ...
//      f_PredictedPower /= f_MeanPower;
// Hoisting:
//		f_PredictedPower = (f_MeanPower + f_PeakPower * f_weight[idx])/f_MeanPower;
//		xx = (f_MeanPower/f_MeanPower) + (f_PeakPower * f_weight[idx])/f_MeanPower;
//      xx = 1 + (f_PeakPower * f_weight[idx]))/f_MeanPower
//
//      xx = 1 + (f_PeakPower*f_weight[idx])/f_MeanPower
//      xx = 1 + f_PeakPower * f_weight[idx] * (1/f_MeanPower)
//	    xx = 1 + f_weight[idx] * ((1/f_MeanPower) * f_PeakPower)
//	    xx = 1 + f_weight[idx] * (f_PeakPower/f_MeanPower)
//		xx = 1 + f_weight[idx] * hoist
	double	meanPeakHoist= f_PeakPower/f_MeanPower;
	float	f_rPowerLen  = 1.0/float(ul_PowerLen);
	float	f_rMeanPower = 1.0/f_MeanPower;

    for ( i = 0; i < ul_PowerLen; i++ )
        {
		// 	used to be f_weight[abs(i - ul_TOffset)]
		f_PredictedPower = 1.0 + f_weight[i - ul_TOffset] * meanPeakHoist;
//		f_tot_weight	+= f_weight[idx];	// Computed but never used

		double  signal, r_noise, PoT_i;
		
		signal	= f_PredictedPower;	// prior pfc = 2
		r_noise	= rebin / ( 2.0 * (signal < 1.0f ? 0 : sqrt(signal-1)) + 1 );	// 6fpc
		PoT_i	= fp_PoT[i] * f_rMeanPower; 

		f_ChiSq		+= SQUARE( PoT_i -signal ) * r_noise;	// fpc=4
		f_null_hyp	+= SQUARE( PoT_i -1	     ) * r_noise;	// fpc=4
		}

    count_flops( 20 * ul_PowerLen + 5);	// 13 fpc vs original 22
    f_ChiSq		*= f_rPowerLen;	// Float recip powerlen
    f_null_hyp	*= f_rPowerLen;

    if ( xsq_null ) *xsq_null = f_null_hyp;
    return ( f_ChiSq );
    }



// opt_f_GetPeak  
// Purpose:Peak power is calculated as the weighted sum of all powers
// within ul_HalfSumLength of the assumed gaussian peak. 
// The weights are given by the gaussian function itself. 
// BUG WATCH : for the f_PeakScaleFactor to work,
//  ul_HalfSumLength *must* be set to sigma.
//
// 8-10-06 BENH - Unrolled loop 3 times to allow for FPU latency & reduce loop overhead
//				- Removed conditionals from loop, removed buff[ abs(equation) ] indexing
//				  created 2 loops to compensate
// xx-xx-03 ERICK - Created function 
//
f_GetPeak_ptt( opt_f_GetPeak )
    {
    float   sum1, sum2, sum3, sum4;
    int     first, last, i;
    int     weight_i;	// f_weight[] is actually middle of a table

	first	= ul_TOffset - ul_HalfSumLength;
    last	= ul_TOffset + ul_HalfSumLength + 1;
    weight_i= first - ul_TOffset;	// negative indexes are ok

    const int   stride = 4;

	sum1 = sum2 = sum3 = sum4 = 0.0;
    i = first;

 /*
    if ( weight_i < 0 )
        {

        // First the going down direction (in the f_weight table)
        weight_i = abs( weight_i );
        for (	; i < last && weight_i > stride - 1; i += stride, weight_i -= stride )
            {
            sum1 += ( fp_PoT[i + 0] - f_MeanPower ) * f_weight[weight_i - 0];
            sum2 += ( fp_PoT[i + 1] - f_MeanPower ) * f_weight[weight_i - 1];
            sum3 += ( fp_PoT[i + 2] - f_MeanPower ) * f_weight[weight_i - 2];
            sum4 += ( fp_PoT[i + 3] - f_MeanPower ) * f_weight[weight_i - 3];
            }

        for (	; weight_i; i++ ) sum1 += ( fp_PoT[i] - f_MeanPower ) * f_weight[weight_i--];
        }
*/
	 // Then the going up direction (in the f_weight table)
	for (	; i < last-(stride-1); i += stride, weight_i += stride )
        {
        sum1 += ( fp_PoT[i + 0] - f_MeanPower ) * f_weight[weight_i + 0];
        sum2 += ( fp_PoT[i + 1] - f_MeanPower ) * f_weight[weight_i + 1];
        sum3 += ( fp_PoT[i + 2] - f_MeanPower ) * f_weight[weight_i + 2];
        sum4 += ( fp_PoT[i + 3] - f_MeanPower ) * f_weight[weight_i + 3];
        }

    for (	; i < last; i++ ) 
		sum1 += ( fp_PoT[i] - f_MeanPower ) * f_weight[weight_i++];

    sum1 += sum2 + sum3 + sum4;
    count_flops( 6 * ul_HalfSumLength );
    return ( sum1 * f_PeakScaleFactor );
    }

/* Not used yet, needs GCC version of asm 
//
//
//
inline void  q_memcpy(void *dest, const void *src, int len)
	{
	__asm {
		mov edi, dest
		mov esi, src
		mov ecx, len
		shr ecx, 2
		rep movsd
		mov ecx, len
		and ecx, 3
		rep movsb
		}
	}



//
//
//
void  *ultra_memcpy(void *in_dest, const void *in_src, size_t length)
	{
	const char	*src = (char *)in_src;
	char		*dest =  (char *)in_dest;

	if (length > 64-1 ) 
		{
		const int SIMD_LEN = 16; // sizeof(VEC);
		int delta = ((size_t)dest) & ( SIMD_LEN-1 );
		if( delta ) // Make destination address pointer SIMD aligned (16 byte)
			{
				delta = SIMD_LEN - delta;
				q_memcpy(dest, src, delta);
				dest  += delta;
				src   += delta;
				length -= delta;
			}

		// How many blocks of 64 bytes?
		int blocks;
		blocks  = length >> 6;
		length &= 63;

		if(our_cpu.has_sse2())
			copy_sse2( src, dest, blocks);
		else if(our_cpu.has_SSE())
			copy_SSE( src, dest, blocks);
		else if(our_cpu.has_3DNowPlus())
			copy_MMXnt( src, dest, blocks);
		else if(our_cpu.has_mmx())
			copy_MMX( src, dest, blocks);
		VEC temp0, temp1, temp2, temp3;

		dest = (char *)f_dst;
		src  = (char *)f_src;

		}
	if(length)
		q_memcpy(dest, src, length);
	return in_dest;
	}
*/
