/******************
 *
 * opt_sse2.cpp
 * 
 * Description: Intel & AMD SSE2 optimzized functions
 * CPUs: Intel Pentium IV, AMD Athlon Model 6+, AMD Athlon 64
 */

// Copyright 2004 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "stdafx.h"

#include <limits.h>
#include "opt_SSE2.hpp"
#include "opt_SSE.hpp"
#include "chirpfft.h"

ALIGN_SIMD long long sse2_dnums[][2] = {
	{LLONG_MIN,		0		},	
	{	0,		LLONG_MIN	},
	{	0,			0		},
	};

#define DNEG_R	  (*(VEC_D *)&sse2_dnums[0])
#define DNEG_I	  (*(VEC_D *)&sse2_dnums[1])
#define DNEG_NONE (*(VEC_D *)&sse2_dnums[2])


__inline vFloat vec_recip2(vFloat v) 
	{
	// obtain estimate
	vFloat estimate = _mm_rcp_ps( v );
	// one round of Newton-Raphson
	return _mm_add_ps(_mm_mul_ps(_mm_sub_ps(ONE, _mm_mul_ps(estimate, v)), estimate), estimate);
	}  


// =============================================================================
//	  v_GetPowerSpectrum
// =============================================================================


GetPowerSpectrum_ptt( sse2_v_GetPowerSpectrum )
	{
	int	i, bin;
	float *workBuf = (float *)FreqData;


	ALIGNED_YES( FreqData );
	ALIGNED_YES( PowerSpectrum );
	for	( i	= 0, bin = 0; i < this_fft_len; i++, bin += bin_len)
		{
		VEC sum1, sum2;
		int *p = (int *)(&PowerSpectrum[bin_off + bin]);

		s_fetch( &FreqData[i+16][0] );
		sum1 = s_get1(&FreqData[i][0]);
		sum1 = s_mult(sum1, sum1);
		sum2 = s_get1(&FreqData[i][1]);
		sum2 = s_mult(sum2, sum2);
		sum1 = s_add(sum1, sum2);
			// WARNING: !! this store overwrites FreqData[i][0], so loop must go bottom to top !!
			//  reusing buffer - not needed after our psNum compute.
		s_put1(&workBuf[i], sum1);	// 	workBuf[i] = psNum;

			// Try to get:
			// movd  [reg32],[xmmreg]
			// movnti [p],[reg32]
		s_put1_NC(p,  sum1 );
		}

		// When using non caching writes (non-temporal), you should allways force
		// the writes to be "globaly visible" to possible other CPUs
	s_fence_writes();
	}

//
// Vectorized version
//
// Original author: Ben Herndon
// Adapted by Joe Segur
//
// Identical to SSE version except using non-cache writes to transposed array
//
GetPowerSpectrum_ptt( sse2_GetTPS_sc8 )
    {
    VEC     sum1, sum2, sum3, sum4, tmp1, tmp2;
    int i, j, bin;
    float *workBuf = (float *)FreqData;
    const int stride = 8;               // in Data Points

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, j = 0, bin = 0;
          j < this_fft_len - (stride - 1) ;
          i += stride * 2, j += stride, bin += bin_len * stride)
        {

        //      SSE Pipeline #1
                //      SSE Pipeline #2
        sum1  = s_getA( &workBuf[i+0*4] );     
        sum1  = s_mult( sum1, sum1);
                sum3  = s_getA( &workBuf[i+2*4] );
        tmp1  = s_swapRI(sum1);
                sum3  = s_mult( sum3, sum3);
        sum1  = s_add(sum1, tmp1);
                tmp2  = s_swapRI(sum3);
        s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 0 * bin_len], sum1);
                sum3  = s_add(sum3, tmp2);
        sum2  = s_getA( &workBuf[i+1*4] );
                s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 4 * bin_len], sum3);
        sum2  = s_mult( sum2, sum2);
                sum4  = s_getA( &workBuf[i+3*4] );
        tmp1  = s_swapRI(sum2);
                sum4  = s_mult( sum4, sum4);
        sum2  = s_add(sum2, tmp1);
                tmp2  = s_swapRI(sum4);
        s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 2 * bin_len], sum2);
                sum4  = s_add(sum4, tmp2);
        sum1  = s_shuffle(sum1, sum2, 0x88);
                s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 6 * bin_len], sum4);
        s_putA(&workBuf[j+0*4], sum1);
                sum3  = s_shuffle(sum3, sum4, 0x88);
        sum1  = s_swapRI(sum1); 
                s_putA(&workBuf[j+1*4], sum3);
        s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 1 * bin_len], sum1);
                sum3  = s_swapRI(sum3);
        sum1  = s_shuffle(sum1, sum1, 0xC6);
                s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 5 * bin_len], sum3);
        s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 3 * bin_len], sum1);
                sum3  = s_shuffle(sum3, sum3, 0xC6);
                s_put1_NC( (int *)&PowerSpectrum[bin_off + bin + 7 * bin_len], sum3);
        }
    // Should never be used, FFT length < 8 or not power of 2
    for (  ; j < this_fft_len; j++, bin += bin_len)
        {
        workBuf[j] = FreqData[j][0] * FreqData[j][0] + FreqData[j][1] * FreqData[j][1];
        PowerSpectrum[bin_off + bin] = workBuf[j];
        }
    s_fence_writes();
    }


// =============================================================================
//
// sse2_ChirpData
// 9-19-06 BENH - Attempted block loading of the 3 input tables into L1 cache before
//               computing (as suggested in AMD's optimizing .pdf)
//			     loop for cacheLen/4/sizeof(element) 
//					- force load begin element of each cache line
//               do a separate loop for each table (serial memory accesses)
//               Slowed computing speed by 18%
// 9-15-06 BENH  Attempt to use non temporal cacheing of the 3 input tables
//		         look ahead by 256 bytes
//               Reduced speed by 50%
// 9-5-06 BENH - Unrolled loop 3 times to allow for FPU latency & reduce loop overhead 
//				- Removed conditionals from loop 
// xx-xx-03 ERICK -  Created function
//
// =============================================================================
ChirpData_ptt( sse2_ChirpData )
    {
    float   chirp_sign; // BENH - Pulled conditional out of loop
    if ( chirp_rate_ind == 0 )
        {
        memcpy( cx_ChirpDataArray, cx_DataArray,  int( ul_NumDataPoints * sizeof(sah_complex)) );  
        return 0;
        }

    // Calculate trigonometric array. This function returns w/o doing
    // anything when sign of chirp_rate_ind reverses. so we have to take care
    // of it.
    //
    chirp_sign = ( chirp_rate_ind >= 0 ) ? 1 : -1;

	const int stride = 2;
    int i = 0;
    int last = ul_NumDataPoints - ( stride - 1 );
    if ( CacheChirpCalc )
        {
		VEC	chirp_signS = (chirp_sign<0) ? NEG_I : NEG_NO;

			// We do Tetsuji's sin/cos calc inside our loop
			// but only if abs(index offset) = ab
		int rate_index = abs( chirp_rate_ind );
		if( rate_index - CurrentChirpRateInd > 1 )
			CalcTrigArray(0, ul_NumDataPoints, rate_index - 1 );
		CurrentChirpRateInd = rate_index;

		for (		; i < last ; i += stride )
			{
			VEC		cdcd, dcdc;
			VEC_D	cd1, cd2;
            double  SinX, CosX, SinS, CosS;

//			s_fetch_temp( &CurrentTrig[i+8] );		// Non-temporal pre-caching seems to be slower by 50%
			s_fetch( &TrigStep[i+8] );				//  could be due to several addresses can't share
			s_fetch_temp( &cx_DataArray[i+16] );	// cache at same time due to address overlap

			if (chirp_sign > 0)
				{
				SinX = CurrentTrig[i+0].Sin;
                CosX = CurrentTrig[i+0].Cos;
                SinS = TrigStep[i+0].Sin;
                CosS = TrigStep[i+0].Cos;
                CurrentTrig[i+0].Sin = SinX * CosS + CosX * SinS;
                CurrentTrig[i+0].Cos = CosX * CosS - SinX * SinS;
    
                SinX = CurrentTrig[i+1].Sin;
                CosX = CurrentTrig[i+1].Cos;
                SinS = TrigStep[i+1].Sin;
                CosS = TrigStep[i+1].Cos;
                CurrentTrig[i+1].Sin = SinX * CosS + CosX * SinS;
                CurrentTrig[i+1].Cos = CosX * CosS - SinX * SinS;
				}

			cd1 = d_getA( &CurrentTrig[i+0].Sin );
			cd2 = d_getA( &CurrentTrig[i+1].Sin );

			cdcd = s_copyL2H( d_cvt2s(cd1), d_cvt2s(cd2));
			cdcd = s_xor( cdcd, chirp_signS);
			dcdc = s_swapRI( cdcd );

			cdcd = s_mult(cdcd, *(VEC *)cx_DataArray[i]);
			dcdc = s_mult(dcdc, *(VEC *)cx_DataArray[i]);

			cdcd = s_xor(cdcd, NEG_R);	// negate real value(s)

			cdcd = s_add(cdcd, dcdc );	// real-imag,  imag+real
			s_putA_nc(cx_ChirpDataArray[i], cdcd);

/*			if( i & 7 ) continue;
			if( i <= 15 ) continue;

			s_flush( &CurrentTrig[i-8] );	// Can't flush addresses we don't own
			s_flush( &TrigStep[i-8] );		// Conditionals or flush statements make loop
			s_flush( &cx_DataArray[i-16] );	// too slow
*/			}
		}

    // Too little memory to cache sin/cos or just falling out of 'CacheChirpCalc' loop
    // ( never happens, length is allways multiple of 128 )
	if( i < ul_NumDataPoints)
		{  // use original routine to finish up any tailings (max stride-1 elements)
		orig_ChirpData(cx_DataArray+i, cx_ChirpDataArray+i
			, chirp_rate_ind, chirp_rate, ul_NumDataPoints-i, sample_rate);
		}

    count_flops( 12 * ul_NumDataPoints );
    return 0 ;
    }


//
// copy_SSE2 - Use Non temporal writes, and non-temporal pre-fetch 
//   nt prefetch only available with SSE2+ - Athlon and Pentium 4+
//
void  copy_SSE2(char *dest, const char *src, int blocks)
	{
	float *f_src = (float*)src;
	float *f_dst = (float*)dest;
	VEC temp0, temp1, temp2, temp3;

	s_fetch_temp( f_src );
	s_fetch_temp( f_src+16 );
	if( ((size_t)src) & 15u ) // misaligned source
		while( blocks-- )
			{
			s_fetch_temp( f_src + 32 );
			s_getU(temp0, &f_src[ 0] );
			s_getU(temp1, &f_src[ 4] );
			s_getU(temp2, &f_src[ 8] );
			s_getU(temp3, &f_src[12] );
			s_putA_nc( f_dst +  0, temp0 );
			s_putA_nc( f_dst +  4, temp1 );
			s_putA_nc( f_dst +  8, temp2 );
			s_putA_nc( f_dst + 12, temp3 );
			f_src += 16, f_dst += 16;
			}
	else // aligned source, use movaps instead of movups
		{
		while( blocks-- )
			{
			s_fetch_temp( f_src + 32 );
			temp0 = s_getA( f_src + 0 );
			temp1 = s_getA( f_src + 4 );
			temp2 = s_getA( f_src + 8 );
			temp3 = s_getA( f_src +12 );
			s_putA_nc( f_dst + 0 , temp0 );
			s_putA_nc( f_dst + 4 , temp1 );
			s_putA_nc( f_dst + 8 , temp2 );
			s_putA_nc( f_dst +12, temp3 );
			f_src += 16, f_dst += 16;
			}
		}
	s_fence_writes();
	}

typedef __m128d vDouble;

// =============================================================================
//
// sse2_vChirpData
// version by: Alex Kan - SSE2 mods (haddsum removal) BH
//   http://tbp.berkeley.edu/~alexkan/seti/
//
ChirpData_ptt( sse2_ChirpData_ak )
	{
	// int offset, int numPoints, double chirpRate
	int i;

	if (chirp_rate_ind == 0) {
		memcpy(cx_ChirpDataArray, cx_DataArray,	(int)ul_NumDataPoints * sizeof(sah_complex)	);
		return 0;
		}

	int vEnd;  
	double srate = chirp_rate * 0.5 / (sample_rate * sample_rate);
	vDouble rate = _mm_set1_pd(chirp_rate * 0.5 / (sample_rate * sample_rate));
	vDouble roundVal = _mm_set1_pd(srate >= 0.0 ? TWO_TO_52 : -TWO_TO_52);

	// main vectorised loop
	vEnd = ul_NumDataPoints - (ul_NumDataPoints & 3);
	for (i = 0; i < vEnd; i += 4) {
		const float *data = (const float *) (cx_DataArray + i);
		float *chirped = (float *) (cx_ChirpDataArray + i);
		vDouble di = _mm_set1_pd(i);
		vDouble a1 = _mm_add_pd(_mm_set_pd(1.0, 0.0), di);
		vDouble a2 = _mm_add_pd(_mm_set_pd(3.0, 2.0), di);
		vDouble x1, y1;

		__m128 d1, d2;
		__m128 cd1, cd2;
		__m128 td1, td2;
		__m128 x;
		__m128 y;
		__m128 s;
		__m128 c;
		__m128 m;

		// load the signal to be chirped
		s_fetch_temp( data+32 );
		d1 = _mm_load_ps(data);
		d2 = _mm_load_ps(data+4);

		// calculate the input angle
		a1 = _mm_mul_pd(a1, a1);
		a2 = _mm_mul_pd(a2, a2);
		a1 = _mm_mul_pd(a1, rate);
		a2 = _mm_mul_pd(a2, rate);

		// reduce the angle to the range (-0.5, 0.5)
		x1 = _mm_add_pd(a1, roundVal);
		y1 = _mm_add_pd(a2, roundVal);
		x1 = _mm_sub_pd(x1, roundVal);
		y1 = _mm_sub_pd(y1, roundVal);
		a1 = _mm_sub_pd(a1, x1);
		a2 = _mm_sub_pd(a2, y1);

		// convert pair of packed double into packed single
		x = _mm_movelh_ps(_mm_cvtpd_ps(a1), _mm_cvtpd_ps(a2));

		// square to the range [0, 0.25)
		y = _mm_mul_ps(x, x);

		// perform the initial polynomial approximations
		s = _mm_mul_ps(y, SS4);
		c = _mm_mul_ps(y, CC3);						
		s = _mm_add_ps(s, SS3);
		c = _mm_add_ps(c, CC2);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS2);
		c = _mm_add_ps(c, CC1);
		s = _mm_mul_ps(s, y);
		c = _mm_mul_ps(c, y);
		s = _mm_add_ps(s, SS1);
		s = _mm_mul_ps(s, x);
		c = _mm_add_ps(c, ONE);

		// perform first angle doubling
		x = _mm_sub_ps(_mm_mul_ps(c, c), _mm_mul_ps(s, s));
		y = _mm_mul_ps(_mm_mul_ps(s, c), TWO);

		// calculate scaling factor to correct the magnitude
		//      m1 = vec_nmsub(y1, y1, vec_nmsub(x1, x1, TWO));
		//      m2 = vec_nmsub(y2, y2, vec_nmsub(x2, x2, TWO));
		m = vec_recip2(_mm_add_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y)));

		// perform second angle doubling
		c = _mm_sub_ps(_mm_mul_ps(x, x), _mm_mul_ps(y, y));
		s = _mm_mul_ps(_mm_mul_ps(y, x), TWO);

		// correct the magnitude (final sine / cosine approximations)
		c = _mm_mul_ps(c, m);
		s = _mm_mul_ps(s, m);

/*		c1 c2 c3 c4
		s1 s2 s3 s4

		R1 i1 R2 I2		R3 i3 R4 i4

		R1 * c1  +  (i1 * s1 * -1)
		i1 * c1  +   R1 * s1  
		R2 * c2  +  (i2 * s2 * -1)
		i2 * c2  +   R2 * s2
*/

		x = s_xor( s_swapRI(d1) , NEG_R);
		y = s_xor( s_swapRI(d2) , NEG_R);
		cd1 = _mm_shuffle_ps(c, c, 0x50);	// 01 01 00 00  AaBb => BBbb => c3c3c4c4
		cd2 = _mm_shuffle_ps(c, c, 0xfa);	// 11 11 10 10  AaBb => AAaa => c1c1c2c2
		td1 = _mm_shuffle_ps(s, s, 0x50);
		td2 = _mm_shuffle_ps(s, s, 0xfa);

		cd1 = _mm_mul_ps(cd1, d1);
		cd2 = _mm_mul_ps(cd2, d2);
		td1 = _mm_mul_ps(td1, x);
		td2 = _mm_mul_ps(td2, y);

		cd1 = _mm_add_ps(cd1, td1);
		cd2 = _mm_add_ps(cd2, td2);

		// store chirped values
		_mm_stream_ps(chirped+0, cd1);
		_mm_stream_ps(chirped+4, cd2);
		}
    s_fence_all();

    // Too little memory to cache sin/cos or just falling out of
    // 'CacheChirpCalc' loop - never happens, length is allways multiple of 128
	if( i < ul_NumDataPoints)
		{  // use original routine to finish up any tailings (max stride-1 elements)
		orig_ChirpData(cx_DataArray+i, cx_ChirpDataArray+i
			, chirp_rate_ind, chirp_rate, ul_NumDataPoints-i, sample_rate);
		}

	ex_count_flops( 12* ul_NumDataPoints );

	return 0;
	}
