// Copyright 2003 Regents of the University of California
// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.
// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING. If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions and
// distribute a linked executable. You must obey the GNU General Public
// License in all respects for all of the code used other than the FFT library
// itself. Any modification required to support these libraries must be
// distributed in source code form. If you modify this file, you may extend
// this exception to your version of the file, but you are not obligated to
// do so. If you do not wish to do so, delete this exception statement from
// your version.
// analyzeFuncs.C

#define DO_SMOOTH

#include "config.h"

#include <cstdio>
#include <iostream>
#include <cstdlib>
#include "mathimf.h"
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#include <time.h>

//#include "sincos.h" 
#include "util.h"
#include "s_util.h"
#include "boinc_api.h"
#ifdef BOINC_APP_GRAPHICS
    #include "reduce.h"
    #include "graphics_data.h"
    #include "graphics_api.h"
    #ifdef DYNAMIC_GRAPHICS
        #include "graphics_lib.h"
    #endif
#endif
#include "diagnostics.h"
#include "pulsefind.h"

// In order to use IPP, set -DUSE_IPP and one of -DUSE_SSE3, -DUSE_SSE2,
// -DUSE_SSE or nothing(generic), IPP precedes FFTW, ooura
// TMR
/*
Abbreviation    Meaning

px  Generic C optimized code
a6  Intel Pentium III processor code with SSE optimization
w7  Intel Pentium 4 processor code with SSE2 optimization
t7  Intel Pentium 4 processor code with SSE3 optimization
m7  Intel Pentium D processors
m7  Intel Xeon processor with Intel EM64T optimization
i7  Intel Itanium processor and Itanium 2 processor optimization
s2  Intel IXP4XX Product Line of Network Processor optimization
*/
// In order to use IPP, set -DUSE_IPP and one of -DUSE_SSE3, -DUSE_SSE2,
// -DUSE_SSE or nothing(generic), IPP precedes FFTW, ooura 
// TMR
/*
Abbreviation    Meaning

px  Generic C optimized code
a6  Intel Pentium III processor code with SSE optimization
w7  Intel Pentium 4 processor code with SSE2 optimization
t7  Intel Pentium 4 processor code with SSE3 optimization
m7  Intel Pentium D processors
m7  Intel Xeon processor with Intel EM64T optimization
i7  Intel Itanium processor and Itanium 2 processor optimization
v8	Intel Core 2-based processors with SSSE3 optimization
s2  Intel IXP4XX Product Line of Network Processor optimization
*/
#if defined(USE_IPP)
#pragma message ("-----IPP-----")
#if defined(USE_SSSE3)				  
#define V8 1						  
#pragma message ("-----SSSE3-----")
#include <ipp_v8.h>					  
#include <tmmintrin.h> // SSSE3       
#include <fvec.h>                     
#elif defined(USE_SSE3)				  
#define T7 1
#pragma message ("-----SSE3-----") 
#include <ipp_t7.h>
#include <pmmintrin.h> // SSE3        
#include <fvec.h>                     
#elif defined(USE_SSE2)
#define W7 1
#pragma message ("-----SSE2-----")
#include <emmintrin.h> //SSE2        
#include <fvec.h>                    
#include <ipp_w7.h>
#elif defined(USE_SSE)
#define A6 1
#pragma message ("-----SSE-----") 
#include <xmmintrin.h> //SSE         
#include <fvec.h>                    
#include <ipp_a6.h>
#else
#pragma message ("-----MMX-----")
#include <ipp_px.h>
#endif
#include <ipp.h>
#include <ippvm.h>       
#include <ipps.h>
#elif defined( USE_FFTWF )
    #pragma message( "----FFTW----" )
    #include "fftw3.h"
bool    had_wisdom;
double  wis_start;

#else
    #pragma message( "----ooura----" )
    #include "fft8g.h"
#endif // USE_IPP
#include "mathimf.h"
#include "seti.h"
#include "analyze.h"
#include "analyzeReport.h"
#include "gaussfit.h"
#include "spike.h"
#include "malloc_a.h"
#include "analyzeFuncs.h"
#include "analyzePoT.h"
#include "chirpfft.h"
#include "worker.h"
#include "filesys.h"
#ifdef BOINC_APP_GRAPHICS
    #include "gdata.h"
    #include "sah_gfx.h"
#endif
#include "progress.h"

#include "Optimizer/optimize.hpp"
#include "Optimizer/chicken.h"
#include "Optimizer/cpuid_x86.h"


#if defined( USE_IPP )
    static int      MaxBufSize = 0;
    static Ipp8u    *FftBuf = NULL;
#endif // USE_IPP

#ifndef M_PI
    #define M_PI    3.14159265358979323846
#endif

#define INVALID_CHIRP   2e+20

double          ProgressUnitSize;
double          progress = 0, remaining = 1;

// Global work buffer - used by many functions
//  size is at least 512K
/*
    (256.0K) smooth DataOutChunk
    (512.0K) PulsePoT
      (512.0K) triplet.aboveBins
      (280.8K) find_pulse.div
        (53.3K) FoldedPOT

    smooth DataOutChunk = WorkData
    x PulsePoT          = WorkData
    x   triplet.aboveBins   = WorkData[ PoTInfo.MaxPoTLen + 16 ];
    x    find_pulse.div     = WorkData[ PoTInfo.MaxPoTLen + 16 ];
    x       FoldedPOT       = WorkData[ PoTInfo.MaxPoTLen + 16 +
                 (PoTInfo.PulseMax*7/4) * sizeof(float) + (maxdivs*MEM_ALIGN)];

*/
sah_complex         *WorkData;
float               *WorkDataSub1, *WorkDataSub2;

// Find maximum FFT length for which transpose of PowerSpectrum is needed.
int getMTFL(int maxFFTLen)
    {
    double      WUDuration, SlewRate, BeamsPerWU;
    int         MaxGaussFFTLen, MaxPulseFFTLen, MaxTripletFFTLen, t = maxFFTLen;

    WUDuration  = ( double ) ( swi.nsamples / swi.subband_sample_rate );
    SlewRate    = static_cast< double >( swi.angle_range ) / WUDuration;
    MaxGaussFFTLen = swi.nsamples / swi.analysis_cfg.gauss_pot_length;
    if ((SlewRate < swi.analysis_cfg.pot_min_slew)
        || (SlewRate > swi.analysis_cfg.pot_max_slew)) MaxGaussFFTLen = 0;
    BeamsPerWU  = std::max( swi.true_angle_range / swi.beam_width, 1.0 );
    MaxPulseFFTLen = std::min( ( int ) swi.analysis_cfg.pulse_fft_max, ( int ) (
        swi.nsamples / (swi.analysis_cfg.pulse_min * BeamsPerWU  * swi.analysis_cfg.pulse_beams) ) );
    MaxTripletFFTLen = ( int ) (
        swi.nsamples / ((swi.analysis_cfg.triplet_min - 0.500001) * BeamsPerWU ) );

    while (t)
        {
        if ( t <= MaxGaussFFTLen ) break;
        if ( t <= MaxPulseFFTLen ) break;
        if ( t <= MaxTripletFFTLen ) break;
        t /= 2;
        }
    return t;
    }

// Tell caller if more than (x) ram is in current host
bool    enough_ram( unsigned int needed ) {
    return ((app_init_data.host_info.m_nbytes != 0)
          &&  (app_init_data.host_info.m_nbytes >= needed));
    }

void load_wisdom(void)
    {
    #ifdef USE_FFTWF
    double  sz;
    FILE    *wisdom;
    had_wisdom = false;
    wis_start = dtime();
    fftwf_set_timelimit(120);
    if ( wisdom = boinc_fopen( "wisdom.sah", "r" ) )
        {
        had_wisdom = true;
        char    *wiz = ( char * ) MEM.calloc( "wisdom", 1024, 64 );
        int     n = 0;
        while ( wiz && n < 64 * 1024 && !feof( wisdom ) )
            {
            n += fread( wiz + n, 1, 80, wisdom );
            }

        fftwf_import_wisdom_from_string( wiz );
        MEM.free( wiz );
        fclose( wisdom );
        }
    #endif
    }

void save_wisdom(void)
    {
    #ifdef USE_FFTWF
    if (!had_wisdom )
        fprintf(stderr, "FFTW Wisdom Planning: %1.2f seconds\n", dtime()-wis_start );
    FILE    *wisdom = boinc_fopen( "wisdom.sah", "w" );
    if ( wisdom )
        {
        char    *wiz = fftwf_export_wisdom_to_string();
        if ( wiz )
            fwrite( wiz, strlen( wiz ), 1, wisdom );
        fclose( wisdom );
        }
    #endif
    }

void notify_user( char *what_to_say )
    {
    #ifdef BOINC_APP_GRAPHICS
        if (!nographics()) strcpy(sah_graphics->status, what_to_say);
    #endif
    }

// =============================================================================
// seti_analyze
//      The main analysis function. Args: state pointer to data, # of
//      points, starting chirp/fftlen Must be called with unchirped data;
//      this function modifies (chirps) the data in place swi parsed WU header
// =============================================================================

int seti_analyze( ANALYSIS_STATE &state )
    {
    sah_complex         *DataIn;        // Buffers (in the order they are used)
    sah_complex         *ChirpedData;
    float               *PowerSpectrum;
    int                 NumDataPoints;
    int                 num_cfft;
    int                 icfft;
    int                 cfft_left;
    int                 NumFfts, ifft, fftlen;
    int                 CurrentSub;
    int                 FftNum;

    double              MinChirpStep;   // Chirping variables
    float               chirprate;
    int                 last_chirp_ind, chirprateind;

    double              progress_diff, progress_in_cfft, cputime0;
    int                 retval;
    unsigned long       bitfield;
    unsigned long       FftLen;
    #if defined( USE_IPP )
        IppsFFTSpec_C_32fc  *FftSpec[MAX_NUM_FFTS];
        int                 BufSize;
    #endif
    bitfield    = swi.analysis_cfg.analysis_fft_lengths;
    DataIn      = state.savedWUData;
    NumDataPoints = state.npoints;
    ChirpedData   = NULL;
    WorkData      = NULL;
    PowerSpectrum = NULL;
    num_cfft      = 0;
    MinChirpStep  = 0.0;
    last_chirp_ind = -1 << 20;
    cputime0 = 0;
    retval = 0;

    #if defined( USE_IPP )
        ippStaticInit();                // initialization of IPP library
    #elif defined( USE_FFTWF )
        // plan space for fftw
        fftwf_plan  analysis_plans[MAX_NUM_FFTS];
    #else
        // fields need by the ooura fft logic
        int         *BitRevTab[MAX_NUM_FFTS];
        float       *CoeffTab[MAX_NUM_FFTS];
    #endif



    // Allocate data array and work area arrays.
    ChirpedData = state.data;
// Crunch3r    
    PowerSpectrum = (float*) calloc_a(NumDataPoints, sizeof(float), MEM_ALIGN);
    if(PowerSpectrum == NULL) SETIERROR(MALLOC_FAILED, "PowerSpectrum == NULL");
// Crunch3r    
    notify_user( "Choosing optimal functions" );

    // The optimize_init() routine chooses the fastest function with acceptable
    // accuracy for various needs. Among those is the chirping function, where
    // some versions use significant extra memory for cached arrays.
    // Formerly there was an oversimplified check assuming that a system with
    // 64 MB of installed RAM could comfortably run an application which uses
    // 67 MB of RAM above and beyond whatever the OS and BOINC require.
    // Now, optimize_init() does more thorough checks including allowing for
    // hosts with multiple CPUs. It returns a "true" if a chirp function which
    // uses cached arrays was fastest AND there's enough memory to use it.
    CacheChirpCalc  = optimize_init();

    FftNum = 0;
    FftLen = 1;
    notify_user("Generating FFT Coefficients" );

    load_wisdom();
    #if defined( USE_FFTWF )
    WorkData = ( sah_complex * ) MEM.alloc( "plan WorkData", bitfield * sizeof(sah_complex) );
    #endif
    while ( bitfield != 0 )
        {
        if ( bitfield & 1 )
            {
            swi.analysis_fft_lengths[FftNum] = FftLen;
            #if defined( USE_IPP )
                int order = 0;
                for ( int tmp = FftLen; !(tmp & 1); order++ )
                    tmp >>= 1;
                if ( ippsFFTInitAlloc_C_32fc(
                         &FftSpec[FftNum], order, IPP_FFT_NODIV_BY_ANY, ippAlgHintFast ) )
                    {
                    SETIERROR( MALLOC_FAILED, "ippsFFTInitAlloc failed" );
                    }
            #elif defined( USE_FFTWF )
                int templen = FftLen * sizeof(sah_complex);
                sah_complex *scratch;
                scratch  = ( sah_complex * ) MEM.alloc( "plan scratch", templen );

                // TODO: Deallocate these at the end of the function
                analysis_plans[FftNum] = fftwf_plan_dft_1d(
                        FftLen,
                        scratch,
                        WorkData,
                        FFTW_BACKWARD,
                        FFTW_MEASURE  | FFTW_PRESERVE_INPUT );
                MEM.free(scratch);
            #else   // Ooura FFT
                // See docs in fft8g.C for sizing guidelines for BitRevTab and CoeffTab.
    BitRevTab[FftNum] = (int*) calloc_a(3+(int)sqrt((float)swi.analysis_fft_lengths[FftNum]), sizeof(int), MEM_ALIGN);
         if(BitRevTab[FftNum] == NULL)  SETIERROR(MALLOC_FAILED, "BitRevTab[FftNum] == NULL");
	BitRevTab[FftNum][0] = 0;
    #endif
            FftNum++;
            }

        FftLen *= 2;
        bitfield >>= 1;
        }
    if( WorkData) free_a( WorkData );
    WorkData = NULL;
    save_wisdom();

    int maxFFTLen = FftLen/2;
    int MaxTransposeFftLen = getMTFL(maxFFTLen);
    swi.num_fft_lengths = FftNum;

#if defined( USE_IPP )
  if (MaxBufSize) {
      FftBuf = (Ipp8u*) malloc_a (MaxBufSize, MEM_ALIGN);
          if (FftBuf == NULL) SETIERROR (MALLOC_FAILED, "FftBuf == NULL");
	    }
    
#elif defined( USE_FFTWF )
        boinc_install_signal_handlers();
#else
        for ( FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++ )
            {
            CoeffTab[FftNum] = ( float * ) MEM.calloc( "CoeffTab[FftNum]",
                    swi.analysis_fft_lengths[FftNum] / 2, sizeof(float));
            }
#endif

    // Allocate WorkData array sized for the biggest FFTs we'll do
    int combo_len, maxdivs, div_len, fold_len, work_len, trans_len;
    pulse_mem( maxdivs, div_len, fold_len);
    combo_len = PoTInfo.MaxPoTLen + 16 + div_len + 32 + fold_len;
    combo_len *= sizeof(float);
    trans_len = MaxTransposeFftLen * 3 * sizeof(float) + MaxTransposeFftLen * sizeof(sah_complex);
    work_len = UNSTDMAX( UNSTDMAX( combo_len, maxFFTLen * sizeof(sah_complex) ), trans_len);

 WorkData = (sah_complex *)malloc_a(work_len * sizeof(sah_complex),MEM_ALIGN);
   if (WorkData == NULL) {
       SETIERROR(MALLOC_FAILED, "WorkData == NULL");
         }
	 
//    WorkData = ( sah_complex * ) MEM.alloc( "WorkData", work_len );
//  fprintf(stderr, "work_len= %d  MaxTransposeFftLen= %d\n", work_len, MaxTransposeFftLen);

    // Smooth Baseline
    #ifdef DO_SMOOTH
        notify_user("Doing Baseline Smoothing" );
        retval = v_BaseLineSmooth(
                DataIn,
                NumDataPoints,
                swi.analysis_cfg.bsmooth_boxcar_length,
                swi.analysis_cfg.bsmooth_chunk_size );
        if ( retval ) SETIERROR( retval, "from v_BaseLineSmooth" );
        save_wisdom();
    #endif

    // gernerate table of chirp/fft pairs (we may read table from file if testing)
    if ( cfft_file != NULL )
        num_cfft = ReadCFftFile( &ChirpFftPairs, &MinChirpStep );
    else
        num_cfft = GenChirpFftPairs( &ChirpFftPairs, &MinChirpStep );
    if ( num_cfft == MALLOC_FAILED )
        {
        SETIERROR( MALLOC_FAILED, "num_cfft == MALLOC_FAILED" );
        }

    // Get together various values that we'll need to analyse power over time
    ComputePoTInfo( num_cfft, NumDataPoints );

    // used to calculate percent done
    // ProgressUnitSize = GetProgressUnitSize(NumDataPoints, num_cfft, swi);
    ProgressUnitSize = GetProgressUnitSize( NumDataPoints, num_cfft );

    // test and set pulse folding subroutines
    FoldTest(ChirpFftPairs, num_cfft, swi.nsamples, show_benchmark);

    // Simon Zadra - Improved stats output, moved from chirpfft.cpp to here, added
    // BENH: CPUID information & only display to stderr if working on 1st part of WU
    if(state.icfft == 0) {
    CPU_INFO    our_cpu;
/*        fprintf( stderr, "Optimized SETI@Home Enhanced application\n" );
        fprintf( stderr, "%10s: Ben Herndon, Josef Segur, Alex Kan, Simon Zadra\n", "Optimizers" );
        fprintf( stderr, "%10s: Linux 32-bit based on S@H V5.15  'Noo? No - Ni!'\n", "Version");
//            ,_OS_, _fft_simd_, _bits_ // SIMD_SubVersion not working --- removed
//            , gmajor_version, gminor_version );
        fprintf( stderr, "%10s: R-%s|%s\n", "Revision", _release_, _compOps_ );

        fprintf( stderr, "%10s: %s %s '%s'\n"
            , "CPUID"
            , our_cpu.vendor_name()
            , our_cpu.model()
            , our_cpu.core()
            );
		/* 
		fprintf( stderr, "%9s: %s %s %s '%s'\n"
            , "CPUID"
            , our_cpu.vendor_name()
            , our_cpu.generation()
            , our_cpu.model()
            , our_cpu.core()
            );
		*/
/*
        int dies = our_cpu.cpu_cnt();
        int cores = our_cpu.core_cnt();
        int threads = our_cpu.thread_cnt();
        fprintf( stderr, "%10s: %d, cores: %d, threads: %d   cache: L1=%dK,  L2=%dK, L3=%dK\n"
            , "CPUs"
            , dies, cores, threads
            , our_cpu.cacheL1()/1024
            , our_cpu.cacheL2()/1024
            , our_cpu.cacheL3()/1024
            );

        fprintf( stderr, "%10s: %s%s%s%s%s%s%s \n"
            , "Features"
            , our_cpu.mmx() ? "MMX, " : ""
            , our_cpu._3Dnow() ? "3DNow!, " : ""
            , our_cpu._3DnowPlus() ? "3DNow!+, " : ""
            , our_cpu.sse()  ? "SSE, " : ""
            , our_cpu.sse2() ? "SSE2, " : ""
            , our_cpu.sse3() ? "SSE3, " : ""
            , our_cpu.ssse3() ? "SSSE3" : ""
            );

        our_cpu.L1_speed();
        while( our_cpu.cpu_mhz() == 0 )
            ;   // Wait til 1 second passes

        char L3speed[16] = {0};
        if(our_cpu.cacheL3())
            sprintf( L3speed," L3=%d,", our_cpu.L3_speed());

        fprintf( stderr, "%10s: %d MHz  -- read MB/s: L1=%d, L2=%d,%s RAM=%d\n\n", "speed"
            , our_cpu.cpu_mhz()
            , our_cpu.L1_speed(), our_cpu.L2_speed(), L3speed, our_cpu.RAM_speed()  );
*/
        fprintf( stderr, "Work Unit Info\n" );
        fprintf( stderr, "True angle range:  %f\n", swi.angle_range );
    }
    else fprintf( stderr, "Restarted at %.2f percent.\n", progress * 100);
    fflush( stderr );

    // Initialize TrigArray's if needed
    if ( CacheChirpCalc )
        TrigArrayInit( NumDataPoints, MinChirpStep,
            ChirpFftPairs[state.icfft].ChirpRateInd, swi.subband_sample_rate );

    boinc_wu_cpu_time( cputime0 );
    reset_units();

    double  chirp_units = 0;

    //--------------------
    // Loop through chirp/fft pairs - this is the top level analysis loop.
    //
    double  last_ptime = 0;
    int     rollovers = 0;
    double  clock_max = 0;
    for ( icfft = state.icfft; icfft < num_cfft; icfft++ )
        {
        fftlen = ChirpFftPairs[icfft].FftLen;
        chirprate = ChirpFftPairs[icfft].ChirpRate;
        chirprateind = ChirpFftPairs[icfft].ChirpRateInd;
        boinc_fpops_cumulative( analysis_state.FLOP_counter * LOAD_STORE_ADJUSTMENT );

        remaining = 1.0 - ( double ) icfft / num_cfft;

        if ( chirprateind != last_chirp_ind )
            {
            notify_user( "Chirping data" );

            retval = ChirpData( DataIn,  ChirpedData,  chirprateind,
                    chirprate,  NumDataPoints,  swi.subband_sample_rate );

            if ( retval ) SETIERROR( retval, "from v_ChirpData()" );

            progress    += ( double ) ( ProgressUnitSize * CHIRP_PROGRESS_FACTOR );
            chirp_units += ( double ) ( ProgressUnitSize * CHIRP_PROGRESS_FACTOR );
            progress = std::min( progress, 1.0 );
            }

        last_chirp_ind = chirprateind;

        // Process this FFT length.
        // For a given FFT length (at a given chirp), we construct
        // PowerSpectrum[] which is a "waterfall" array of power spectra,
        // each fftlen long on the frequency axis and sample time long
        // on the time axis.
        // As we go along, we check each spectra for spikes.
        state.icfft = icfft;        // update analysis state

        // Find index into FFT length table for the current
        // FFT length. This will be the same index needed
        // for ooura's coeffecient and bit reverse tables.
        for ( FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++ )
            {
            if ( swi.analysis_fft_lengths[FftNum] == fftlen )
                {
                break;
                }
            }

        #ifdef BOINC_APP_GRAPHICS
            if ( !nographics() )
                {
                notify_user("Computing Fast Fourier Transform" );
                sah_graphics->fft_info.chirp_rate = chirprate;
                sah_graphics->fft_info.fft_len = fftlen;
                 }
        #endif

        // If PoT freq bin is non-negative, we are into PoT analysis
        // for this cfft pair and should not re-output an "ogh" line.
        if ( state.PoT_freq_bin == -1 )
            {
            retval = result_group_start();
            if ( retval ) SETIERROR( retval, "from result_group_start" );
            }

        // Number of FFTs for this length
        NumFfts = NumDataPoints / fftlen;

        #ifdef BOINC_APP_GRAPHICS
            GRAPH_BUFFER    *gbp = 0;
            if ( !nographics() )
                {
                gbp = sah_graphics->get_generate_buffer( true );
                if ( gbp )
                    {
                    gbp->rarray.init_data( fftlen, NumFfts );
                    }
                }
        #endif

        // Now we will process the just chirped data...
        //  We will break the entire ChirpedData[] into 'fftlen' chunks and do each one
        //  as a block
        //
        //  1. backward FFT over each block and output to WorkData[]
        //  2. compute power spectrum table over fft'ed output
        //  3. if pot freq bin == -1, then try to find spike in PowerSpectrum[]
        //
        if (fftlen > MaxTransposeFftLen)
            {  // Simpler case, no transpose
            for ( ifft = 0; ifft < NumFfts; ifft++ )
                {
                CurrentSub = fftlen * ifft;
                #ifndef USE_FFTW        // FFTW now uses out of place transforms.
                    memcpy( WorkData, &ChirpedData[CurrentSub], int(fftlen * sizeof(sah_complex)) );
                #endif

                count_flops( 4 * fftlen + 5 * fftlen * log( double(fftlen) ) / log(2.0 ) );
                #if defined( USE_IPP )
                    ippsFFTInv_CToC_32fc(
                        ( Ipp32fc * ) WorkData,
                        ( Ipp32fc * ) WorkData,
                        FftSpec[FftNum],
                        FftBuf );
                #elif defined( USE_FFTWF )
                    fftwf_execute_dft(
                        analysis_plans[FftNum],
                        &ChirpedData[CurrentSub],
                        WorkData );
                #else
                    // replace time with freq - ooura FFT
                    cdft( fftlen * 2, 1, WorkData, BitRevTab[FftNum], CoeffTab[FftNum] );
                #endif

                // replace freq with power
                PwrSpectrumOnly( WorkData, (float *)WorkData, fftlen );

                // any ETIs ?!
                // If PoT freq bin is non-negative, we are into PoT analysis
                // for this cfft pair and need not redo spike finding.
                if ( state.PoT_freq_bin == -1 )
                    {
                    count_flops( fftlen );
                    retval = FindSpikes( (float *)WorkData, fftlen, ifft, swi );
                    progress += SpikeProgressUnits( fftlen ) * ProgressUnitSize / NumFfts;
                    if ( retval ) SETIERROR( retval, "from FindSpikes" );
                    }

                // progress = ((float)icfft)/num_cfft + ((float)ifft)/(NumFfts*num_cfft);
                progress = std::min( progress, 1.0 );
                #ifdef BOINC_APP_GRAPHICS
                    if ( !nographics() )
                        {
                        if ( gbp )
                            {
                            gbp->rarray.add_source_row( (float *)WorkData );
                            }

                        sah_graphics->local_progress = ( (( float ) ifft + 1) / NumFfts );
                        }
                #endif
                remaining = 1.0 - ( double ) ( icfft + 1 ) / num_cfft;
                fraction_done( progress, remaining );

                }
            }
        else       // transpose needed, use strips of 4
            {
             for ( ifft = 0; ifft < NumFfts - 3; ifft += 4 )
                {
                for ( int iC = 0; iC < 4; iC++ )
                    {
                    CurrentSub = fftlen * (ifft + iC);
                    sah_complex *WorkArea = &WorkData[iC * fftlen / 2];  // assume sah_complex 2 floats
                    #ifndef USE_FFTW        // FFTW now uses out of place transforms.
                        memcpy( WorkArea, &ChirpedData[CurrentSub], int(fftlen * sizeof(sah_complex)) );
                    #endif

                    #if defined( USE_IPP )
                        ippsFFTInv_CToC_32fc(
                            ( Ipp32fc * ) WorkArea, ( Ipp32fc * ) WorkArea, FftSpec[FftNum], FftBuf );
                    #elif defined( USE_FFTWF )
                        fftwf_execute_dft( analysis_plans[FftNum], &ChirpedData[CurrentSub], WorkArea );
                    #else // replace time with freq - ooura FFT
                        cdft( fftlen * 2, 1, WorkArea, BitRevTab[FftNum], CoeffTab[FftNum] );
                    #endif

                    // replace freq with power
                    PwrSpectrumOnly( WorkArea, (float *)WorkArea, fftlen );
                    count_flops( 4 * fftlen + 5 * fftlen * log( double(fftlen) ) / log(2.0 ) );

                    // any ETIs ?!
                    // If PoT freq bin is non-negative, we are into PoT analysis
                    // for this cfft pair and need not redo spike finding.
                    if ( state.PoT_freq_bin == -1 )
                        {
                        count_flops( fftlen );
                        retval = FindSpikes( (float *)WorkArea, fftlen, ifft + iC, swi );
                        progress += SpikeProgressUnits( fftlen ) * ProgressUnitSize / NumFfts;
                        if ( retval ) SETIERROR( retval, "from FindSpikes" );
                        }

                    // progress = ((float)icfft)/num_cfft + ((float)ifft)/(NumFfts*num_cfft);
                    progress = std::min( progress, 1.0 );
                    #ifdef BOINC_APP_GRAPHICS
                        if ( !nographics() )
                            {
                            if ( gbp ) gbp->rarray.add_source_row( (float *)WorkData );
                            sah_graphics->local_progress = ( (( float ) ifft + 1) / NumFfts );
                            }
                    #endif
                    remaining = 1.0 - ( double ) ( icfft + 1 ) / num_cfft;
                    fraction_done( progress, remaining );

                    }
                TransposeStrip(fftlen, NumFfts, ifft, (float *)WorkData, PowerSpectrum);
                }
            for (         ; ifft < NumFfts; ifft++ ) // Obligatory tail, should never be used.
//          for ( ifft = 0; ifft < NumFfts; ifft++ ) // test version when above commented out.
                {
                CurrentSub = fftlen * ifft;
                #ifndef USE_FFTW        // FFTW now uses out of place transforms.
                    memcpy( WorkData, &ChirpedData[CurrentSub], int(fftlen * sizeof(sah_complex)) );
                #endif

                #if defined( USE_IPP )
                    ippsFFTInv_CToC_32fc( ( Ipp32fc * ) WorkData,
                        ( Ipp32fc * ) WorkData, FftSpec[FftNum], FftBuf );
                #elif defined( USE_FFTWF )
                    fftwf_execute_dft( analysis_plans[FftNum], &ChirpedData[CurrentSub], WorkData );
                #else // replace time with freq - ooura FFT
                    cdft( fftlen * 2, 1, WorkData, BitRevTab[FftNum], CoeffTab[FftNum] );
                #endif

                // BENH: new version replace freq with power does transpose as well as puts values back
                // into WorkData (for use by findSpikes)
                //
                GetPowerSpectrum( WorkData, PowerSpectrum, fftlen, ifft, NumFfts);
                count_flops( 4 * fftlen + 5 * fftlen * log( double(fftlen) ) / log(2.0 ) );

                // any ETIs ?!
                // If PoT freq bin is non-negative, we are into PoT analysis
                // for this cfft pair and need not redo spike finding.
                if ( state.PoT_freq_bin == -1 )
                    {
                    count_flops( fftlen );
                    retval = FindSpikes( (float *)WorkData, fftlen, ifft, swi );
                    progress += SpikeProgressUnits( fftlen ) * ProgressUnitSize / NumFfts;
                    if ( retval ) SETIERROR( retval, "from FindSpikes" );
                    }

                // progress = ((float)icfft)/num_cfft + ((float)ifft)/(NumFfts*num_cfft);
                progress = std::min( progress, 1.0 );
                #ifdef BOINC_APP_GRAPHICS
                    if ( !nographics() )
                        {
                        if ( gbp ) gbp->rarray.add_source_row( (float *)WorkData );
                        sah_graphics->local_progress = ( (( float ) ifft + 1) / NumFfts );
                        }
                #endif
                remaining = 1.0 - ( double ) ( icfft + 1 ) / num_cfft;
                fraction_done( progress, remaining );
                }
            }                       // loop through chirped data array
        #ifdef BOINC_APP_GRAPHICS
            if ( !nographics() && gbp) sah_graphics->generate_done( gbp );
        #endif
        fraction_done( progress, remaining );

        // Analyze Power over Time. May return quickly if this FFT
        // length and/or this WUs slew rate places the data block
        // outside PoT analysis limits.
        // Counting flops is done inside analyze_pot
        retval = analyze_pot( PowerSpectrum, NumDataPoints, ChirpFftPairs[icfft] );
        if ( retval ) SETIERROR( retval, "from analyze_pot" );

        #ifdef BOINC_APP_GRAPHICS
            // switch the display back to "best of" signals
            if ( !nographics() )
                {
                sah_graphics->gi.copy( best_gauss, true );
                sah_graphics->pi.copy( best_pulse, true );
                sah_graphics->ti.copy( best_triplet, true );
                }
        #endif

        // Force progress to 100% before calling result_group_end() to store
        // 100% in state file so it will survive exit & relaunch
        if ( icfft == (num_cfft - 1) )
            {
            progress = 1;
            remaining = 0;
            fraction_done( progress, remaining );
            }

        retval = checkpoint();
        if ( retval ) SETIERROR( retval, "from checkpoint() in seti_analyse()" );

        }                   // loop over chirp/fftlen paris

    //--------------------------------
    //  Done with all FFTs
    //

    // Return the "best of" signals. This may include duplicates of
    // already reported interesting signals.
    if ( best_spike->score )
        {
        retval = outfile.printf( "%s",
                best_spike->s.print_xml( 0, 0, 1, "best_spike" ).c_str() );
        if ( retval < 0 )
            SETIERROR( WRITE_FAILED,
                "from outfile.printf (best spike) in seti_analyze()" );
        }

    if ( best_gauss->score )
        {
        retval = outfile.printf( "%s",
                best_gauss->g.print_xml( 0, 0, 1, "best_gaussian" ).c_str() );
        if ( retval < 0 )
            SETIERROR( WRITE_FAILED,
                "from outfile.printf (best gaussian) in seti_analyze()" );
        }

    if ( best_pulse->score )
        {
        retval = outfile.printf( "%s",
                best_pulse->p.print_xml( 0, 0, 1, "best_pulse" ).c_str() );
        if ( retval < 0 )
            SETIERROR( WRITE_FAILED,
                "from outfile.printf (best pulse) in seti_analyze()" );
        }

    if ( best_triplet->score )
        {
        retval = outfile.printf( "%s",
                best_triplet->t.print_xml( 0, 0, 1, "best_triplet" ).c_str() );
        if ( retval < 0 )
            SETIERROR( WRITE_FAILED,
                "from outfile.printf (best triplet) in seti_analyze()" );
        }

    notify_user( "Work unit done" );

    final_report();  // add signal and flop counts to stderr.txt

    // Simon Zadra - Moved checkpoint() here (see below)
    retval = checkpoint();  // allow a final checkpoint

    if ( PowerSpectrum ) free_a( PowerSpectrum );

    #if defined( USE_IPP )
        for ( FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++ )
            if ( FftSpec[FftNum] ) ippsFFTFree_C_32fc( FftSpec[FftNum] );

        if ( FftBuf ) free_a( FftBuf );
        FftBuf = NULL;
    #elif defined( USE_FFTWF )
        for ( FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++ )
            if ( analysis_plans[FftNum] )  fftwf_destroy_plan( analysis_plans[FftNum] );

    #else   // ooura
        for ( FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++ )
            {
            if ( BitRevTab[FftNum] ) free_a( BitRevTab[FftNum] );
            if ( CoeffTab[FftNum] )  free_a( CoeffTab[FftNum] );
            }
    #endif
    if ( WorkData )       free_a( WorkData );
    WorkData = NULL;
    if ( ChirpFftPairs )  free_a( ChirpFftPairs );
    if ( CacheChirpCalc ) FreeTrigArray();

    /* Simon Zadra - moved checkpoint() up retval = checkpoint(); // force a final
       checkpoint */
    xml_indent( -2 );
    outfile.printf( "</result>" );
    outfile.close();

    // if (retval) SETIERROR(WRITE_FAILED,"from outfile.fflush in seti_analyze()");
    return retval;
    }

// =============================================================================
// v_BaseLineSmooth
// purpose: We use a sliding boxcar method for baseline smoothing. Input
// data is the time domain. It is transformed (using a separate array) into the frequency
// domain. After baseline smoothing is done in the frequency domain, it is
// transformed back into the time domain.
// =============================================================================
int v_BaseLineSmooth
    (
        sah_complex *DataIn,
        int         NumDataPoints,
        int         BoxCarLength,
        int         NumPointsInChunk
    )
    {
    sah_complex                 *DataInChunk;
    float                       Total, LocalMean, ScaleFactor;
    int                         NumTimeChunks, TimeChunk, Endpoint;
    int                         h, i, j, k;

    static sah_complex          *DataOutChunk = 0;
    static float                *PowerSpectrum = 0;
    static int                  OldNumPointsInChunk = 0;
    #ifdef USE_IPP
        static IppsFFTSpec_C_32fc   *FftSpec = NULL;
    #elif defined( USE_FFTWF )
        static fftwf_plan           backward_transform, forward_transform;
    #endif /* USE_FFTWF */

    #ifndef USE_FFTWF
        static int                  *BitRevTab = NULL;
        static float                *CoeffTab = NULL;
    #endif /* USE_FFTWF */

    NumTimeChunks = NumDataPoints / NumPointsInChunk;

    // If we keep doing transforms that are the same length, don't reinitialize plans
    if ( NumPointsInChunk != OldNumPointsInChunk )
        {
        if ( OldNumPointsInChunk != 0 )
            {
            #ifdef USE_IPP
                if ( FftSpec ) ippsFFTFree_C_32fc( FftSpec );
            #elif defined( USE_FFTWF )
                if ( backward_transform ) MEM.free( backward_transform );
                if ( forward_transform )  MEM.free( forward_transform );
            #else
                if ( BitRevTab ) free_a( BitRevTab );
                if ( CoeffTab )  free_a( CoeffTab );
            #endif
            if ( PowerSpectrum ) free_a( PowerSpectrum );
            if ( DataOutChunk )  free_a( DataOutChunk );
            }

        PowerSpectrum = ( float * ) calloc_a( NumPointsInChunk, sizeof(float), MEM_ALIGN );
        if ( PowerSpectrum == NULL )
            {
            printf( "Could not allocate Power Spectrum array in v_BaseLineSmooth()\n" );
            exit( 1 );
            }

        // Do the transforms in the DataOutChunk, since DataInChunk won't
        // necessarily be aligned correctly (and we won't get SIMD)
        // TODO: automatically make DataInChunk aligned correctly so we
        // don't need the memcpy. This may already be done if
        // NumPointsInChunk*sizeof(sah_complex) is a multiple of MEM_ALIGN
/*        DataOutChunk = ( sah_complex * )
            MEM.alloc("smooth DataOutChunk", NumPointsInChunk * sizeof(sah_complex));
        OldNumPointsInChunk = NumPointsInChunk;
*/
   DataOutChunk = (sah_complex *)malloc_a(NumPointsInChunk * sizeof(sah_complex),MEM_ALIGN);
       OldNumPointsInChunk = NumPointsInChunk;
       

        #ifdef USE_IPP
            int order = 0;
            for ( int tmp = NumPointsInChunk; !(tmp & 1); order++ )
                tmp >>= 1;
            ippsFFTInitAlloc_C_32fc(&FftSpec, order, IPP_FFT_NODIV_BY_ANY,
                ippAlgHintFast );
        #elif defined( USE_FFTWF )
            sah_complex *scratch = ( sah_complex * )
                MEM.alloc("smooth scratch", NumPointsInChunk * sizeof(sah_complex) );
            backward_transform = fftwf_plan_dft_1d(
                    NumPointsInChunk,
                    scratch,
                    DataOutChunk,
                    FFTW_BACKWARD,
                    FFTW_PATIENT | FFTW_PRESERVE_INPUT );
            forward_transform = fftwf_plan_dft_1d(
                    NumPointsInChunk,
                    DataOutChunk,
                    DataOutChunk,
                    FFTW_FORWARD,
                    FFTW_PATIENT );
            MEM.free( scratch );
        #else
            BitRevTab = ( int * )
                MEM.calloc("smooth BitRevTab",
                    3 + ( int ) sqrt( ( float ) NumPointsInChunk / 2 ), sizeof(int));
            CoeffTab = ( float * )
                MEM.calloc("smooth CoeffTab", NumPointsInChunk / 2, sizeof(float) );
            // flag to tell cdft() to init it's work areas
            // already done since we used calloc_a();
            // BitRevTab[0] = 0;
        #endif

        }

    for ( TimeChunk = 0; TimeChunk < NumTimeChunks; TimeChunk++ )
        {
        #ifdef BOINC_APP_GRAPHICS
            if ( !nographics() )
                sah_graphics->local_progress = ( (( float ) TimeChunk) / NumTimeChunks );
        #endif

        DataInChunk = &( DataIn[TimeChunk * NumPointsInChunk] );
        #ifndef USE_FFTWF
            memcpy( DataOutChunk, DataInChunk, int(NumPointsInChunk * sizeof(sah_complex)) );
        #endif

        // transform to freq
        #ifdef USE_IPP
            ippsFFTInv_CToC_32fc(
                ( Ipp32fc * ) DataOutChunk,
                ( Ipp32fc * ) DataOutChunk,
                FftSpec,
                NULL );
        #elif !defined( USE_FFTWF )
            cdft( NumPointsInChunk * 2, 1, DataOutChunk, BitRevTab, CoeffTab );
        #else
            fftwf_execute_dft( backward_transform, DataInChunk, DataOutChunk );
        #endif

        PwrSpectrumOnly( DataOutChunk, PowerSpectrum, NumPointsInChunk );
        count_flops( 3 * NumPointsInChunk );

        // Begin: normalize in freq. domain via sliding boxcar
        Endpoint = NumPointsInChunk / 2;

        i = Endpoint;       // start i at lowest negative freq;

        // this is low point in first boxcar
        j = i + BoxCarLength / 2;   // start j midpoint in first boxcar
        k = i + BoxCarLength;       // start k at end of first boxcar;

        // thus a boxcar is i j k
        f_sum(Total, &PowerSpectrum[i], k-i );  // Get mean for first boxcar
        LocalMean = Total / BoxCarLength;
        ScaleFactor = ( float ) ( sqrt( LocalMean / 2 ) );

        // normalize 1st half of 1st boxcar
        for ( h = i; h < j; h++ )
            {
            DataOutChunk[h][0] /= ScaleFactor;
            DataOutChunk[h][1] /= ScaleFactor;
            }

        for ( ; k != Endpoint; i++, j++, k++ )
            {                       // sliding boxcar
            if ( k == NumPointsInChunk )
                k = 0;              // take care of wrapping

            if ( j == NumPointsInChunk )
                j = 0;

            if ( i == NumPointsInChunk )
                i = 0;

            LocalMean = LocalMean    - PowerSpectrum[i] /
                        BoxCarLength + PowerSpectrum[k] /
                        BoxCarLength;
            ScaleFactor = ( float ) ( sqrt( LocalMean / 2 ) );

            DataOutChunk[j][0] /= ScaleFactor;
            DataOutChunk[j][1] /= ScaleFactor;
            }                       // End sliding boxcar

        // normalize final half of final boxcar
        for ( h = j; h < k; h++ )
            {
            DataOutChunk[h][0] /= ScaleFactor;
            DataOutChunk[h][1] /= ScaleFactor;
            }

        // End: normalize in freq. domain via sliding boxcar
        // transform back to time
        #ifdef USE_IPP
            ippsFFTFwd_CToC_32fc(
                ( Ipp32fc * ) DataOutChunk,
                ( Ipp32fc * ) DataOutChunk,
                FftSpec,
                NULL );
        #elif !defined( USE_FFTWF )
            cdft( NumPointsInChunk * 2, -1, DataOutChunk, BitRevTab, CoeffTab );
        #else
            fftwf_execute( forward_transform );
        #endif
        count_flops(  NumPointsInChunk * (10 + log( double(NumPointsInChunk) ) / log( 2.0 ) ));

        // return powers to normal
        for ( i = 0; i < NumPointsInChunk; i++ )
            {
            DataInChunk[i][0] = DataOutChunk[i][0] / NumPointsInChunk;
            DataInChunk[i][1] = DataOutChunk[i][1] / NumPointsInChunk;
            }
        }

        if ( DataOutChunk )  free_a( DataOutChunk );
        if ( PowerSpectrum ) free_a( PowerSpectrum );
        #ifdef USE_IPP
            if ( FftSpec ) ippsFFTFree_C_32fc( FftSpec );
        #elif defined( USE_FFTWF )
            if ( forward_transform )  fftwf_destroy_plan( forward_transform );
            if ( backward_transform ) fftwf_destroy_plan( backward_transform );
        #else
            if ( CoeffTab )  MEM.free( CoeffTab );
            if ( BitRevTab ) MEM.free( BitRevTab );
        #endif
      return 0;
    }

// =======================================================================================
//    opt_v_GetPowerSpectrum - replace freq with power
//
//  this converts each irrational frequency number in a buffer into its representational power
//  in a 2nd table
//  It places the frequencies in bins throughout the powerspectrum table
//    this is done by official seti source by a separate function 'transpose'
//
//  Separate Transpose function idea implemented by Alex Kan on Optimzied Mac source
//  and included in 5.17 berkeley source by Eric K.
//
//  Transpose inside of powerspectrum by BENH
//   Faster this way, as source data already in cache, spreads out the memory bandwidth overload
//   into blocks
// =======================================================================================
//#define   GetPowerSpectrum_ptt( func_ ) void func_ ( sah_complex *FreqData,   \
//          float *PowerSpectrum, int this_fft_len, int bin_off, int bin_len )
//
GetPowerSpectrum_ptt( std_v_GetPowerSpectrum )
    {
    int i, bin;
    float *workBuf = (float *)FreqData;

    ALIGNED_YES( FreqData );
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0, bin = 0; i < this_fft_len; i++, bin += bin_len)
        {
        float psNum = FreqData[i][0] * FreqData[i][0] + FreqData[i][1] * FreqData[i][1];
        workBuf[i] = psNum; // also store in linear form for use by FindSpikes
        PowerSpectrum[bin_off + bin] = psNum;
        }
    }


//  v_GetPowerSpectrum - replace freq with power
//  Used by baseline_smooth and for chirp/fft pairs for which only Spike searches are performed.
//
PwrSpectrumOnly_ptt( orig_v_GetPowerSpectrum )
    {
    int i;

    ALIGNED_YES( FreqData );    // BENH - tell ICC they're aligned
    ALIGNED_YES( PowerSpectrum );
    for ( i = 0; i < this_fft_len; i++ )
        {
        PowerSpectrum[i] = FreqData[i][0] *  FreqData[i][0] +
            FreqData[i][1] * FreqData[i][1];
        }
    }

// Original Transpose routines from 5.17 source,
// modified to do a strip of exactly 4 lines. The r parameter is the beginning line number.

TransposeStrip_ptt( orig_v_Transpose )
    {
    // Stupidest possible algorithm, assumes in and out can't overlap
    int i, j;
    FORCE_FRAME_POINTER;
    ALIGNED_YES( in );
    ALIGNED_YES( out );
    for (j = 0; j < 4; j++)
        {
        for (i = 0; i < x; i++) out[i * y + r + j] = in[j * x + i];
        }
    }       

template <int x>
inline void v_subTranspose(float *in, float *out, int xline, int yline)
    {
    // Transpose an X by X subsection of a XLINE by YLINE matrix into the 
    // appropriate part of a YLINE by XLINE matrix.  "IN" points to the first
    // (lowest address) element of the input submatrix.  "OUT" points to the
    // first (lowest address) element of the output submatrix.
    int i, j;
    float *p;
    register float tmp[x * x];
    for (j = 0; j < x; j++)
        {
        p = in + j * xline;
        for (i = 0; i < x; i++) tmp[j + i * x] = *p++;
        }
    for (j = 0; j < x; j++)
        {
        p = out + j * yline;
	    for (i = 0; i < x; i++) *p++ = tmp[i + j * x];
        }
    }

TransposeStrip_ptt( orig_v_Transpose2 )
    {
    // Attempts to improve cache hit ratio by transposing 4 elements at a time.
    int i, j;
    ALIGNED_YES( in );
    ALIGNED_YES( out );
    for (j = 0; j < 3; j += 2)
        {
        for (i = 0; i < x - 1; i += 2) v_subTranspose<2>(in + j * x + i, out + y * i + j + r, x, y);
        for (     ; i < x; i++)
            {
            out[i * y + j + r] = in[j * x + i];
            out[i * y + j + 1 + r] = in[(j + 1) * x + i];
            }
        }
    }

TransposeStrip_ptt( orig_v_Transpose4 )
    {
    // Attempts to improve cache hit ratio by transposing 16 elements at a time.
    int i, j;
    ALIGNED_YES( in );
    ALIGNED_YES( out );
    for (i = 0; i < x - 3; i += 4) v_subTranspose<4>(in + i, out + y * i + r, x, y);
    for (     ; i < x; i++)
        {
        for (j = 0; j < 4; j++) out[i * y + r + j] = in[j * x + i];
        }
    }
