// Copyright 2003 Regents of the University of California SETI_BOINC is free
// software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation;
// either version 2, or (at your option) any later version. &
//
// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details. &
//
// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions and
// distribute a linked executable. You must obey the GNU General Public
// License in all respects for all of the code used other than the FFT
// library itself. Any modification required to support these libraries must
// be distributed in source code form. If you modify this file, you may
// extend this exception to your version of the file, but you are not
// obligated to do so. If you do not wish to do so, delete this exception
// statement from your version. &
//
// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING. If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#include "config.h"

#include <cstdio>
#include <cstdlib>
#include CMATH_LIB
#include <map>

#include "diagnostics.h"
#include "util.h"
#include "s_util.h"
#ifdef BOINC_APP_GRAPHICS
    #include "graphics_api.h"
    #ifdef DYNAMIC_GRAPHICS
        #include "graphics_lib.h"
    #endif
#endif

#include "seti_header.h"
#include "seti.h"
#include "util.h"
#include "filesys.h"
#include "s_util.h"
#include "analyze.h"
#include "worker.h"
#include "chirpfft.h"
//#include "sincos.h"

#include "Optimizer/optimize.hpp"

sh_sint8_t  num_gaussian_searches;
sh_sint8_t  num_pulse_searches;
sh_sint8_t  num_triplet_searches;

char        *cfft_file = NULL;

ChirpFftPair_t	*ChirpFftPairs = NULL;

SinCosArray		*TrigStep =	NULL;		// trigonometric array of MinChirpStep
SinCosArray		*CurrentTrig = NULL;	// current chirprate trigonometric array
int				CurrentChirpRateInd;	// current chirprate index (absolute value)

bool			CacheChirpCalc;


// =============================================================================
//    GenChirpFftPairs -
// =============================================================================
int GenChirpFftPairs( ChirpFftPair_t **ChirpFftPairs, double *MinChirpStep )
    {

    long            i, j, max_fft_len;
    double          CRate;
    double          ChirpRes;
    double          NumSamples;
    double          WUDuration;
    double          SlewRate;
    long            MaxGaussFFTLen;

    double          BeamsPerWU;
    long            MinPulseFFTLen;
    long            MaxPulseFFTLen;
    long            NumLimits;
    double          *ChirpSteps;
    ChirpFftPair_t  tmp;

		// Set initial values
	ChirpRes 	= swi.analysis_cfg.chirp_resolution;
    NumSamples	= swi.nsamples;
    WUDuration	= ( double ) ( swi.nsamples / swi.subband_sample_rate );
    SlewRate	= static_cast< double >( swi.angle_range ) / WUDuration;
    MaxGaussFFTLen = swi.nsamples / swi.analysis_cfg.gauss_pot_length;
    BeamsPerWU	= std::max( swi.true_angle_range / swi.beam_width, 1.0 );
    MinPulseFFTLen = (long)ceil(
            swi.nsamples / (swi.analysis_cfg.pulse_max * BeamsPerWU * swi.analysis_cfg.pulse_beams) );
	MaxPulseFFTLen = std::min( ( unsigned long ) swi.analysis_cfg.pulse_fft_max, ( unsigned long ) (
			swi.nsamples / (swi.analysis_cfg.pulse_min * BeamsPerWU  * swi.analysis_cfg.pulse_beams) )
             );
    NumLimits	= ( long ) swi.analysis_cfg.chirps.size();

    std::map<sh_sint8_t, ChirpFftPair_t> ChirpFftMap;

    // An arrays associated with the FFT length array.
//    ChirpSteps = ( double * ) MEM.calloc( "ChirpSteps", swi.num_fft_lengths, sizeof(double) );
 ChirpSteps      = (double *)calloc(swi.num_fft_lengths, sizeof(double));
    // Each FFT length has a (different) associated chirp step.
    CalcChirpSteps( ChirpSteps, MinChirpStep );

    // Lets simplify the logic for doing chirps by using the STL map
    // container for sorting.
    max_fft_len = swi.analysis_fft_lengths[swi.num_fft_lengths - 1];
    for ( i = 0; i < NumLimits; i++ )
        {
        for ( j = 0; j < swi.num_fft_lengths; j++ )
            {
            CRate = 0.0;

            // do we need to generate entries for this fft length?
            if ( swi.analysis_fft_lengths[j] & swi.analysis_cfg.chirps[i].fft_len_flags )
                {

                tmp.FftLen = swi.analysis_fft_lengths[j];

                // Are we going to fit gaussians at this chirpfft?
                tmp.GaussFit = (
                        (tmp.FftLen <= MaxGaussFFTLen)
                    &&  (SlewRate >= swi.analysis_cfg.pot_min_slew)
                    &&  (SlewRate <= swi.analysis_cfg.pot_max_slew)
                    );
                tmp.PulseFind = 0;
                while ( CRate <= swi.analysis_cfg.chirps[i].chirp_limit )
                    {

                    // Pulse find on alternate chirp rates when fitting
                    // Gaussians
                    if ( (SlewRate >= swi.analysis_cfg.pot_min_slew)
                     &&  (SlewRate <= swi.analysis_cfg.pot_max_slew) )
                        {
                        tmp.PulseFind = !( tmp.PulseFind ) && tmp.GaussFit
                    &&  ( tmp.FftLen <= MaxPulseFFTLen ) && ( tmp.FftLen >= MinPulseFFTLen );

                        // else pulse find when our FFT liength is in
                        // range...
                        }
                    else
                        {
                        tmp.PulseFind = ( tmp.FftLen <= MaxPulseFFTLen )
                    &&  ( tmp.FftLen >= MinPulseFFTLen );
                        }

                    // Force all chirp rates to be a multiple of the minimum
                    // step.
                    tmp.ChirpRate =
                        ( tmp.ChirpRateInd = (int32_t) floor( CRate / (*MinChirpStep) + 0.5 ) ) *
                        ( *MinChirpStep );

                    // Generate keys specifying the order of the chirps
                    // chirp_rate,sign,fft_length order...
                    sh_sint8_t  key;
                    key = static_cast< sh_sint8_t >( floor( CRate / (*MinChirpStep) + 0.5 ) ) * 2;
                    key *= max_fft_len * 2;
                    key += tmp.FftLen;

                    // Insert the positive chirp overwriting any existing
                    // with this key value;
                    ChirpFftMap[key] = tmp;

                    // fprintf(stderr,"%f %d
                    // %f\n",CRate,tmp.ChirpRateInd,tmp.ChirpRate);
                    // fflush(stderr); Now the negative chirps for non-zero
                    // chirp;
                    if ( CRate != 0.0 )
                        {
                        tmp.ChirpRate *= -1;
                        tmp.ChirpRateInd *= -1;
                        key += max_fft_len * 2; // flip the bit representing the sign in the

                        // key
                        ChirpFftMap[key] = tmp;
                        }

                    CRate += ChirpSteps[j];
                    }                           // end while loop through chirp rates
                }                               // end if
            }                                   // end for loop through fft lenngths
        }                                       // end for loop through cfft parameter table

    // Copy our ordered CFFT map into a more standard array
//    *ChirpFftPairs = ( ChirpFftPair_t * ) 
//		MEM.calloc( "ChirpFftPairs", ChirpFftMap.size(), sizeof(ChirpFftPair_t) );
 *ChirpFftPairs = (ChirpFftPair_t *)calloc(ChirpFftMap.size(),sizeof(ChirpFftPair_t));
    std::map<sh_sint8_t, ChirpFftPair_t>::const_iterator p;
    i = 0;
    for ( p = ChirpFftMap.begin(); p != ChirpFftMap.end(); p++ )
        {
        num_gaussian_searches += ( p->second.GaussFit * (p->second.FftLen - 1) );
        num_pulse_searches +=
            (sh_sint8_t)(
                p->second.PulseFind *
                (p->second.FftLen - 1) *
                (NumSamples / p->second.FftLen) *
                log( ( double ) NumSamples / p->second.FftLen ) /
                log( 2.0 )
            );
        num_triplet_searches +=
            (sh_sint8_t)(
                p->second.PulseFind *
                (p->second.FftLen - 1) *
                (NumSamples / p->second.FftLen) *
                (NumSamples / p->second.FftLen)
            );
        ( *ChirpFftPairs )[i++] = p->second;
        }

    /* Edit by Simon Zadra - improved stats output in analyzeFuncs.cpp
     * fprintf( stderr,"ar=%f NumCfft=%d NumGauss="SINT8_FMT"
     * NumPulse="SINT8_FMT" NumTriplet="SINT8_FMT"\n", swi.angle_range,
     * static_cast<int>(ChirpFftMap.size()),
     * SINT8_FMT_CAST(num_gaussian_searches),
     * SINT8_FMT_CAST(num_pulse_searches),
     * SINT8_FMT_CAST(num_triplet_searches) ); fflush(stderr); End edit
     */
    if ( ChirpSteps )free_a( ChirpSteps );
    return ( ChirpFftMap.size() );
    }

// =======================================================================================
//    CalcChirpSteps -

// For each FFT length, calculate the corresponding chirp
// step (in Hz/s), given the chirp resolution (in FFT bins).
// Keep track of the minimum chirp step.
// The goal here is to have the drift be less than 1 bin over the
// duration of interest to the analysis. In the case of gaussians
// that duration is 2 beam crossings. In the case of pulses it's
// one beam crossing. In the case of spikes it's one fft length.
// =======================================================================================
void CalcChirpSteps( double *ChirpSteps, double *MinChirpStep )
    {

    double  ChirpRes;
    double  NumSamples;
    double  WUDuration;

    int     i;
    double  SecondsInAnalysis;
    double  SlewRate;
    long    FftLen;
    long    MaxGaussFFTLen;
    double  BeamsPerWU;
    long    MinPulseFFTLen;
    long    MaxPulseFFTLen;
    long    MaxTripletFFTLen;
    double  SecondsInBeam;

    ChirpRes    = swi.analysis_cfg.chirp_resolution;
    NumSamples = swi.nsamples;
    WUDuration = ( double ) ( swi.nsamples / swi.subband_sample_rate );
    SlewRate = static_cast< double >( swi.angle_range ) / WUDuration;
    MaxGaussFFTLen = swi.nsamples / swi.analysis_cfg.gauss_pot_length;
    BeamsPerWU = std::max( swi.true_angle_range / swi.beam_width, 1.0 );
    MinPulseFFTLen = (long)ceil(
            swi.nsamples / (swi.analysis_cfg.pulse_max * BeamsPerWU * swi.analysis_cfg.pulse_beams) );
    MaxPulseFFTLen = std::min(
            ( unsigned long )
                (
                    swi.nsamples /
                        (swi.analysis_cfg.pulse_min * swi.analysis_cfg.pulse_beams * BeamsPerWU)
                ),
            ( unsigned long ) swi.analysis_cfg.pulse_fft_max );
    MaxTripletFFTLen = swi.nsamples / swi.analysis_cfg.triplet_min;
    SecondsInBeam = WUDuration / swi.true_angle_range * swi.beam_width;

    *MinChirpStep = 0.0;
    for ( i = 0; i < swi.num_fft_lengths; i++ )
        {
        FftLen = swi.analysis_fft_lengths[i];

        double  SecondsInFFT;
        SecondsInFFT = WUDuration / ( NumSamples / FftLen );

        // Are we going to do Gaussians?
        if ( (FftLen <= MaxGaussFFTLen)
         &&  (SlewRate >= swi.analysis_cfg.pot_min_slew)
         &&  (SlewRate <= swi.analysis_cfg.pot_max_slew) )
            {
            SecondsInAnalysis = SecondsInBeam * 2;

            // In not, are we going to do pulses or triplets?
            }
        else if ( (FftLen <= MaxPulseFFTLen) && (FftLen <= MaxTripletFFTLen) )
            {

            // Look for pulses over one beam crossing or one beam crossing
            // at the pot_min_slew, whichever is lower.
            SecondsInAnalysis = std::min(
                    SecondsInBeam,
                    swi.beam_width / swi.analysis_cfg.pot_min_slew );
            }
        else
            {
            SecondsInAnalysis = SecondsInFFT;
            }

        ChirpSteps[i] = ChirpRes / ( SecondsInAnalysis * SecondsInFFT );

        // fprintf(stderr,"%d %f\n",i,ChirpSteps[i]); fflush(stderr);
        if ( ChirpSteps[i] < *MinChirpStep || *MinChirpStep == 0.0 )
            {
            *MinChirpStep = ChirpSteps[i];
            }
        }
    }

// =============================================================================
//    VeryClose -
// =============================================================================
bool VeryClose( double a, double b, double CloseEnough )
    {

    double  Diff;

    Diff = fabs( a - b );

    if ( Diff > CloseEnough )
        return ( false );
    else
        return ( true );
    }

// =============================================================================
//    ReadCFftFile -
// =============================================================================
int ReadCFftFile( ChirpFftPair_t **ChirpFftPairs, double *MinChirpStep )
    {

    // ChirpFftSet_t ChirpFftSet;
    ChirpFftPair_t  *ChirpFftPair;
    int             NumChirpFftPairs;
    FILE            *cfftfp;
    double          ChirpRate;
    int             FftLen;

    NumChirpFftPairs = 0;

    cfftfp = boinc_fopen( cfft_file, "r" );

    ChirpFftPair = ( ChirpFftPair_t * ) calloc( 1, sizeof(ChirpFftPair_t) );

    while ( fscanf( cfftfp, "%lf %d", &ChirpRate, &FftLen ) == 2 )
        {
        ChirpFftPair = ( ChirpFftPair_t * ) realloc(
                ( void * ) ChirpFftPair,
                sizeof(ChirpFftPair_t) * (NumChirpFftPairs + 1) );
        if ( ChirpFftPair == NULL ) SETIERROR( MALLOC_FAILED, "ChirpFftPair == NULL" );
        ChirpFftPair[NumChirpFftPairs].ChirpRate = ChirpRate;
        ChirpFftPair[NumChirpFftPairs].FftLen = FftLen;
        NumChirpFftPairs++;
        }
	
    *ChirpFftPairs = ChirpFftPair;
    *MinChirpStep = ( float ) TEST_MIN_CHIRP_STEP;

    fclose( cfftfp );

    return ( NumChirpFftPairs );
    }





// =============================================================================
//	  std_v_ChirpData - Increase the frequency of the input signal over time.
//
// Note that the input angle is calculated in double-precision. The angle
// becomes so large that single-precision arithmetic is insufficient for the
// accuracy that we need. We can only convert to single-precision after
// reducing the range to (-0.5, 0.5).
// Note:	chirp_rate is in Hz	per	second
// =============================================================================
ChirpData_ptt( std_v_ChirpData )
	{
	int	i;

	#ifdef DEBUG
		fprintf( stderr, "icfft	= %6d  crate index = %6d\n", icfft,	chirp_rate_ind );
		fflush(	stderr );
	#endif

	if ( chirp_rate_ind	== 0 )
		{
		// NOTE	int	CAST
		memcpy(	cx_ChirpDataArray, cx_DataArray, ( int ) ul_NumDataPoints *	sizeof(sah_complex)	);
		return 0;
		}


	ALIGNED_YES( cx_ChirpDataArray );
	ALIGNED_YES( cx_DataArray );
	ALIGNED_YES( CurrentTrig );

	// calculate trigonometric array this function returns w/o doing
	// nothing when	sign of	chirp_rate_ind reverses. so	we have	to take
	// care	of it.
	if ( CacheChirpCalc	) CalcTrigArray( 0, ul_NumDataPoints, chirp_rate_ind );

	for	( i	= 0; i < ul_NumDataPoints; i++ )
		{
		float	c, d, real,	imag;

		if ( CacheChirpCalc	)
			{
			c =	CurrentTrig[i].Cos;
			d =	( chirp_rate_ind > 0 ) ? CurrentTrig[i].Sin	: -CurrentTrig[i].Sin;
			}
		else
			{
			double	dd,	cc;
			double	time;

			// since ang is	getting	moded by 2pi, we calculate "ang	mod
			// 2pi"	before the call	to sincos()	inorder	to reduce
			// roundoff	error. (Bug	submitted by Tetsuji "Maverick"	Rai)
			double	ang;
			time = static_cast<	double >( i	) /	sample_rate;
			ang	= 0.5 *	chirp_rate * time *	time;

			ang	-= floor( ang );
			ang	*= M_PI	* 2;
			sincos(	ang, &dd, &cc );
			c =	cc;
			d =	dd;
			}

		// Sometimes chirping is done in place.	We don't want to
		// overwrite data prematurely.
		real = cx_DataArray[i][0] *	c -	cx_DataArray[i][1] * d;
		imag = cx_DataArray[i][0] *	d +	cx_DataArray[i][1] * c;
		cx_ChirpDataArray[i][0]	= real;
		cx_ChirpDataArray[i][1]	= imag;
		}

	count_flops( 12 * ul_NumDataPoints );

	return ( 0 );
	}

// =============================================================================
// InitTrigArray - initialize TrigStep and CurrentTrig
//
// use:	Trigonometric arrays functions These functions makes trigonometric
// arrays very quickly and will	speed up v_ChirpData() by Tetsuji "Maverick" Rai &
//
// The algorithm used to generate the sine / cosine values is based on that
// described in "A Fast, Vectorizable Algorithm for Producing Single-Precision
// Sine-Cosine Pairs" by Marcus H. Mendenhall. The paper is available at:
//     http://arxiv.org/pdf/cs.MS/0406049
//	 
// =============================================================================
// params:  ( int len, double ChirpStep, int InitChirpRateInd, double SampleRate )
TrigArrayInit_ptt( InitTrigArray )
	{
	int		i;
	double	ang, Coef;

	FreeTrigArray();	// Release memory, if we got it earlier
TrigStep    = ( SinCosArray * ) calloc_a(sizeof(SinCosArray), len, MEM_ALIGN );
CurrentTrig = (	SinCosArray * )	calloc_a(sizeof(SinCosArray), len, MEM_ALIGN );

//	TrigStep    = ( SinCosArray * ) calloc(len, sizeof(SinCosArray) );
//	CurrentTrig = (	SinCosArray * )	calloc( len , sizeof(SinCosArray) );

	ALIGNED_YES( TrigStep );
	ALIGNED_YES( CurrentTrig );

	// Make	ChirpStep array
	Coef = ChirpStep / ( SampleRate	* SampleRate );

	for	( i	= 0; i < len; i++ )
		{

		// since ang is	getting	cast to	a float, we	calculate "ang mod 2pi"
		// before the call to sincosf()	inorder	to reduce roundoff error.
		// (Bug	submitted by Tetsuji "Maverick"	Rai) addition: now it's	used
		// as double float,	but	this bug fix is	still preferable
		ang	= 0.5 *	( double ) i * ( double	) i	* Coef;
		ang	-= floor( ang );
		ang	*= 2 * M_PI;
		sincos(	ang, &TrigStep[i].Sin, &TrigStep[i].Cos	);
		}

	// Set initial trigonometric array
	if ( (CurrentChirpRateInd =	abs( InitChirpRateInd )) !=	0 )
		{
		Coef = CurrentChirpRateInd * ChirpStep / ( SampleRate *	SampleRate );

		for	( i	= 0; i < len; i++ )
			{
			ang	= 0.5 *	( double ) i * ( double	) i	* Coef;
			ang	-= floor( ang );
			ang	*= 2 * M_PI;
			sincos(	ang, &CurrentTrig[i].Sin, &CurrentTrig[i].Cos );
			}
		}
	else
		{

		// if it starts	from the beginning,	it's quite simple
		for	( i	= 0; i < len; i++ )
			{
			CurrentTrig[i].Sin = 0.;
			CurrentTrig[i].Cos = 1.0;
			}
		}
	}

// =============================================================================
//	  CalcTrigArray	- calculate	next CurrentTrig

// skip	automatically when abs(new_rate) == CurrentChirpRateInd 
//	(it happens when sign of chirprate reverses)
// Otherwise 
//	sin(x+step) = sin(x)*cos(step)	+ cos(x)*sin(step) 
//  cos(x+step) = cos(x)*cos(step) -  sin(x)*sin(step)
//
// In most cases index increases by	1, and in the later	phase of a WU by
// 4 at	most in	reference WU, but it's 2-3 times faster	than sincos()
// function	on P4 with gcc or icc. When	index increases	by 1, this is
// about 10	times as fast as sincos()
//
// =============================================================================
void CalcTrigArray(	int off, int len, int ChirpRateInd )
	{
	int	TempCRateInd;
	int	i, j;
	int	go_up;

    TempCRateInd    = abs( ChirpRateInd );
    go_up           = !((TempCRateInd - CurrentChirpRateInd) & 0x1);

    for ( i = CurrentChirpRateInd; i < TempCRateInd; i++ )
        {
		// BENH: Process in alternating directions to keep data in cache
		// for re-processing
		if(go_up) for ( j = off; j < off+len; j++ )
            {
            double  SinX, CosX, SinS, CosS;

            SinX = CurrentTrig[j].Sin;
            CosX = CurrentTrig[j].Cos;
            SinS = TrigStep[j].Sin;
            CosS = TrigStep[j].Cos;

            CurrentTrig[j].Sin = SinX * CosS + CosX * SinS;
            CurrentTrig[j].Cos = CosX * CosS - SinX * SinS;
            }
		else for ( j = off+len; j > off; j-- )
            {
            double  SinX, CosX, SinS, CosS;

            SinX = CurrentTrig[j-1].Sin;
            CosX = CurrentTrig[j-1].Cos;
            SinS = TrigStep[j-1].Sin;
            CosS = TrigStep[j-1].Cos;

            CurrentTrig[j-1].Sin = SinX * CosS + CosX * SinS;
            CurrentTrig[j-1].Cos = CosX * CosS - SinX * SinS;
            }
		go_up = !go_up;
        }

    CurrentChirpRateInd = TempCRateInd;
 	}

// =============================================================================
//	  FreeTrigArray	- free TrigStep	and	CurrentTrig
// =============================================================================
TrigArrayFree_ptt( FreeTrigArray )
	{
	if ( TrigStep )	free_a( TrigStep );
	TrigStep = NULL;

	if ( CurrentTrig ) free_a( CurrentTrig );
	CurrentTrig	= NULL;
	}





// *****************************************************************************
// ***************************** FFT functions  ********************************
// *****************************************************************************


#if defined( USE_FFTWF )
void cdft_fpops(uint32 fftLen, double *adds, double *mults) {
	static	oldFftLen = 0;
	double blocks, pwr2;
	struct  {
		double	adds;
		double	mults;
		double	divs;
		} cnt;

		// cft1st
	cnt.adds += 52, cnt.mults += 4;	// 0-15
	if(fftLen >=16) {
		cnt.adds += 60, cnt.mults += 20;	// 16-31
		if(fftLen >=32) {
			blocks = (fftLen-32)/16;
			cnt.adds += blocks*77, cnt.mults += blocks*42;	// 32 - nn
		}
	}

		// cdftmdl
	int	pwr8, pwr;
	for(pwr8 = 16, pwr = 16; (pwr<<3) <= fftLen ; pwr <<= 3 ) {
    	pwr8 = pwr;

		// middleA
		double mdlAdds = 52, mdlMults = 4;

		if ( (pwr8 * 8) < fftLen)	// middleB
			mdlAdds += 60, mdlMults += 20;

		if ( (pwr8 * 16) < fftLen) {	// middleC
			blocks = (fftLen/(pwr8 * 8)) - 2;
			cnt.adds += blocks * 11, cnt.mults += blocks * 10;	// outside loop 'wkr' calcs
			mdlAdds += blocks*66, mdlMults += blocks*32;
		}

		pwr2 = pwr8>>1;
		cnt.adds += pwr2 * mdlAdds, cnt.mults += pwr2 * mdlMults;
	}

		// cftfsub
	pwr2 = pwr8>>1;
	if ((pwr8 << 1) < fftLen)
		cnt.adds += 16*pwr2;
	else if ((pwr8 << 1) == fftLen)
		cnt.adds += 4*pwr2;

	*adds = cnt.adds;
	*mults = cnt.mults;
	}
#endif

