// Copyright 2003 Regents of the University of California SETI_BOINC is free
// software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation;
// either version 2, or (at your option) any later version. &
// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details. &
// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING. If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. &
// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions and
// distribute a linked executable. You must obey the GNU General Public
// License in all respects for all of the code used other than the FFT
// library itself. Any modification required to support these libraries must
// be distributed in source code form. If you modify this file, you may
// extend this exception to your version of the file, but you are not
// obligated to do so. If you do not wish to do so, delete this exception
// statement from your version. &
// gaussfit.C $Id: gaussfit.cpp,v 1.6 2006/10/17 15:59:39 Ben Exp $ &
// The purpose of gaussian fitting is to find signals that rise and fall in
// power over time at a rate consistent with the beam pattern of the
// telescope.
//
#define ODD_MATH true

#include "config.h"

#include <stdio.h>
#include MATH_LIB

#include <algorithm>

// debug stuff
#define DEBUG_POT

// define DUMP_POWER_SPECTRA #define DUMP_GAUSSIAN
#define POT_TO_DUMP     32
#define TOFF_TO_DUMP    15
#define FFT_TO_DUMP     131072
int gul_PoT;
int gul_Fftl;

// end debug stuff
#ifdef BOINC_APP_GRAPHICS
    #include "graphics_api.h"
    #ifdef DYNAMIC_GRAPHICS
        #include "graphics_lib.h"
    #endif
#endif

// include "util.h"
#include "s_util.h"
#include "analyze.h"
#include "seti.h"
#include "worker.h"
#include "analyzeFuncs.h"
#include "analyzeReport.h"
#include "analyzePoT.h"
#include "lcgamm.h"
#include "gaussfit.h"
#include "chirpfft.h"

#ifdef BOINC_APP_GRAPHICS
    #include "sah_gfx.h"
#endif

#include "Optimizer/optimize.hpp"
bool    dump_pot = false;


// =======================================================================================
//   f_GetPeak -

// Peak power is calculated as the weighted
// sum of all powers within ul_HalfSumLength
// of the assumed gaussian peak.
// The weights are given by the gaussian function itself.
// BUG WATCH : for the f_PeakScaleFactor to work,
// ul_HalfSumLength *must* be set to sigma.
// =======================================================================================
f_GetPeak_ptt( std_f_GetPeak )
    {
    int     i;
    float   f_total;

    f_total = 0.0;

	ALIGNED_YES( fp_PoT );
    // Find a weighted sum
    for ( i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++ )
        {
        f_total += ( fp_PoT[i] - f_MeanPower ) * f_weight[abs( i - ul_TOffset )];
        }

    count_flops( 6 * ul_HalfSumLength );
    return ( f_total * f_PeakScaleFactor );
    }

// =============================================================================
//
//  f_GetChiSq  -
//
// We calculate our assumed gaussian powers on the fly as we try to fit
// them to the actual powers at each point along the PoT.
//
// ChiSq in this realm is: 
// sum[0:i]( (observed power - expected power)^2 / expected variance ) 
// The power of a signal is:
// power = (noise + signal)^2 
//       = noise^2 + signal^2 + 2*noise*signal
// With mean power normalization, noise becomes 1, leaving:
// power = signal^2 +or- 2*signal + 1
//
// =============================================================================
f_GetChiSq_ptt( std_f_GetChiSq )
	{
    int     i;
    float   f_ChiSq, f_null_hyp, f_PredictedPower, f_tot_weight;
    double  signal;
    double  noise;

    f_ChiSq	   = 0;
    f_null_hyp = 0;
    f_tot_weight = 0;

    for ( i = 0; i < ul_PowerLen; i++ )
        {
        f_PredictedPower = f_MeanPower + f_PeakPower * f_weight[abs( i - ul_TOffset )];

        f_tot_weight += f_weight[abs( i - ul_TOffset )];

        f_PredictedPower /= f_MeanPower;	//prior = 5pfc

        signal = f_PredictedPower;
        noise  = ( 2.0 * sqrt( std::max( f_PredictedPower, 1.0f ) - 1 ) + 1 );	// 5fpc

        f_ChiSq += ( static_cast< float >( rebin ) 
			*SQUARE ( fp_PoT[i] / f_MeanPower -signal ) / noise );	// 6fpc
        f_null_hyp += ( static_cast< float >( rebin )
			*SQUARE ( fp_PoT[i] / f_MeanPower -1	  ) / noise );	// 6fpc
        }

    count_flops( 20 * ul_PowerLen + 5 );	// 22 actually
    f_ChiSq		/= ul_PowerLen;
    f_null_hyp	/= ul_PowerLen;

	if ( xsq_null ) *xsq_null = f_null_hyp;
    return ( f_ChiSq );
    }



// =============================================================================
//  f_GetTrueMean  -

// TrueMean is the mean power of the data set minus all power out to
// ExcludeLen from our current TOffset.
// =============================================================================
float f_GetTrueMean
    (
        float   fp_PoT[],
        int     ul_PowerLen,
        float   f_TotalPower,
        int     ul_TOffset,
        int     ul_ExcludeLen
    )
    {
    int     first, last, length;
    float   f_ExcludePower;

	ALIGNED_YES( fp_PoT );

    // take care that we do not add to exclude power beyond PoT bounds!
    first =  ul_TOffset - ul_ExcludeLen;
	last  =  ul_TOffset + ul_ExcludeLen + 1;
 	if ( first < 0)
		 first = 0;
	if ( last > swi.analysis_cfg.gauss_pot_length )
		 last = swi.analysis_cfg.gauss_pot_length;
	length = last-first;

	float f_len = 1.0/float(ul_PowerLen - length);	// To avoid IPP double divide

	f_sum(f_ExcludePower, &fp_PoT[first], length );

	count_flops( length + 5 );
    return ( (f_TotalPower - f_ExcludePower) * f_len );
    }



// =============================================================================
//  f_GetPeakScaleFactor  -
//
// The PeakScaleFactor is calculated such that when used in f_GetPeak(),
// the actual peak power can be extracted from a weighted sum. This sum
// (see f_GetPeak()), is calculated as : sum = SUM[x from -sigma to
// +sigma] of (gaussian weights * our data) The gaussian weights are
// e^(-x^2 / sigma^2). Our data is A(e^(-x^2 / sigma^2)), where 'A' is
// the peak power. Through algebraic manipulation, we have: A = sum * (1
// / SUM[x from -sigma to +sigma] of (e^(-x^2 / sigma^2))^2. The factor
// by which we multiply the sum is the PeakScaleFactor. It is completely
// determined by sigma.
//
// =============================================================================
float f_GetPeakScaleFactor( float f_sigma )
    {

    int     i, i_s;
    double   f_sigma_sq;
    double   f_total;

    i_s		    = static_cast< int >( floor( f_sigma + 0.5 ) );
    f_sigma_sq	= f_sigma * f_sigma;
    f_total		= 0.0;

    for ( i = -i_s; i <= i_s; i++ )
        {
        f_total += static_cast< float >( EXP( i, 0, f_sigma_sq ) );
        }

    count_flops( 13 * f_sigma + 3 );
    return ( 1 / f_total );
    }



// =============================================================================
//  ChooseGaussEvent  -
// =============================================================================
int ChooseGaussEvent
    (
        int     ifft,
        float   PeakPower,
        float   TrueMean,
        float   ChiSq,
        float   null_ChiSq,
        int     bin,
        float   sigma,
        float   PoTMaxPower,
        float   fp_PoT[]
    )
    {

    GAUSS_INFO  gi;
    float       scale_factor;
    float       gauss_bins, gauss_dof, null_dof;

    // gaussian info
    gi.bin = bin;
    gi.fft_ind = ifft;
    gi.g.chirp_rate = ChirpFftPairs[analysis_state.icfft].ChirpRate;
    gi.g.fft_len = ChirpFftPairs[analysis_state.icfft].FftLen;
    gi.g.sigma = sigma;
    gi.g.peak_power = PeakPower;
    gi.g.mean_power = TrueMean;
    gi.g.chisqr = ChiSq;
    gi.g.null_chisqr = null_ChiSq;
    gi.g.freq = cnvt_bin_hz( bin, gi.g.fft_len );

    double  t_offset;
    t_offset = ( (( double ) gi.fft_ind + 0.5) / swi.analysis_cfg.gauss_pot_length )
		* PoTInfo.WUDuration;

    gi.g.detection_freq = calc_detection_freq( gi.g.freq, gi.g.chirp_rate, t_offset );
    gi.g.time = swi.time_recorded + t_offset / 86400.0;
    gi.g.max_power = PoTMaxPower;
    time_to_ra_dec( gi.g.time, &gi.g.ra, &gi.g.decl );

    // Scale PoT down to 256 16 bit ints. for (i=0;
    // i<swi.analysis_cfg.gauss_pot_length; i++) gi.pot[i] = fp_PoT[i]; //
    // ???
    //
    scale_factor = static_cast< float >( gi.g.max_power ) / 255.0f;
    if ( gi.g.pot.size() != swi.analysis_cfg.gauss_pot_length )
        {
        gi.g.pot.set_size( swi.analysis_cfg.gauss_pot_length );
        }

    float_to_uchar(
        fp_PoT,
        ( unsigned char * ) (gi.g.pot),
        swi.analysis_cfg.gauss_pot_length,
        scale_factor );

    if ( !swi.analysis_cfg.gauss_null_chi_sq_thresh )
        swi.analysis_cfg.gauss_null_chi_sq_thresh = 1.890;

    // Gauss score used for "best of" and graphics. This score is now set to
    // be based upon the probability that a signal would occur due to noise
    // and the probability that it is shaped like a Gaussian (normalized to 0
    // at thresholds). Thanks to Tetsuji for making me think about this. The
    // Gaussian has 62 degrees of freedom and the null hypothesis has 63
    // degrees of freedom when gauss_pot_length=64;
    //
    gauss_bins = swi.analysis_cfg.gauss_pot_length;
    gauss_dof = gauss_bins - 2;
    null_dof = gauss_bins - 1;
    gi.score = lcgf(0.5 * gauss_dof, std::max( gi.g.chisqr * 0.5 * gauss_bins, 0.5 * gauss_dof + 1 ) )
		- lcgf( 0.5 * gauss_dof, swi.analysis_cfg.gauss_chi_sq_thresh * 0.5 * gauss_bins )
		- lcgf( 0.5 * null_dof, std::max( gi.g.null_chisqr * 0.5 * gauss_bins, 0.5 * null_dof + 1 ) )
		+ lcgf( 0.5 * null_dof, swi.analysis_cfg.gauss_null_chi_sq_thresh * 0.5 * gauss_bins );

    // Only include "real" Gaussians (those meeting the chisqr threshold) in
    // the best Gaussian display.
    //
    if ( gi.score > best_gauss->score && (gi.g.chisqr < swi.analysis_cfg.gauss_chi_sq_thresh) )
        {
        *best_gauss = gi;
        }

    // Update gdata gauss info regardless of whether it is the best thus far
    // or even passes the final threshold. If a gaussian has made it this
    // far, display it.
    //
    #ifdef BOINC_APP_GRAPHICS
    if ( !nographics() ) sah_graphics->gi.copy( &gi );
    #endif

    count_flops( 24);

    // Final thresholding and reporting.
    if ( gi.g.peak_power / gi.g.mean_power >= PoTInfo.GaussPeakPowerThresh )
        {
        if ( (gi.g.chisqr <= PoTInfo.GaussChiSqThresh)
         &&  (gi.g.null_chisqr >= swi.analysis_cfg.gauss_null_chi_sq_thresh) )
            {
            return ( result_gaussian( gi ) );
            }
        }

    return ( 0 );
    }




// =============================================================================
//  GaussFit  -
//
// For setiathome the Sigma and Gaussian PoT length don't change during
// a run of the application, so these frequently used values can be
// precalculated and kept.
//
// =============================================================================
int GaussFit( float *fp_PoT, int ul_FftLength, int ul_PoT )
    {
    int             i, retval;
    BOOLEAN         b_IsAPeak;
    float           f_NormMaxPower;
    float           f_null_hyp;

    int             ul_TOffset;
    int             iSigma;
    float           f_GroupSum, f_GroupMax;
    int             i_f, iPeakLoc;

    float           f_TotalPower, f_MeanPower, f_TrueMean, f_ChiSq, f_PeakPower;

    static float    f_PeakScaleFactor;
    static float    *f_weight;

    iSigma    = static_cast< int >( floor( PoTInfo.GaussSigma + 0.5 ) );

		// Create the f_wieght table if not exist yet.
    if ( !f_weight )
        {
        f_PeakScaleFactor = f_GetPeakScaleFactor ( static_cast< float >( PoTInfo.GaussSigma ) );
		// BENH: double the size of f_weight table
		//  in various functions it is accessed by  f_weight[ abs(i-offset) ]
		//  which creates a nasty dependancy chain to compute the abs and the float load
		//  waits til the integer abs is calculated
		// With double table, negative indexes are valid and abs() is not needed
		//
		// Table length typically around 64 to 128...so not huge
		//
		float *f_weight_base;

		int w_length = PoTInfo.GaussTOffsetStop;

//            f_weight_base = (float*) malloc_a(( w_length * 2)* sizeof(float),MEM_ALIGN );
	    f_weight_base = (float*) malloc_a(( w_length * 2)*sizeof(float),MEM_ALIGN );
		//  f_weight is the middle of the actual table
		f_weight = &f_weight_base[w_length];
		for ( i = 0; i < w_length; i++ )
            {
			float	weight = static_cast< float >( EXP( i, 0, PoTInfo.GaussSigmaSq ) );
            f_weight[0+i] = weight;
            f_weight[0-i] = weight;
            }
        }

	ALIGNED_YES( fp_PoT );


	// BENH: summing func call
    // Find mean over power-of-time array
	f_sum(f_TotalPower, fp_PoT, swi.analysis_cfg.gauss_pot_length );

	f_MeanPower	  = f_TotalPower / swi.analysis_cfg.gauss_pot_length;

    // Normalize power-of-time and check for the existence of at least 1 peak.
    //
	float  f_rMeanPower = 1.0/f_MeanPower;

	// BENH: find max inside loop, then see if max is >powerThresh (change /= to *= recip)
	float  pot_max = PoTInfo.GaussPowerThresh;
	for ( i = swi.analysis_cfg.gauss_pot_length; i; i-- )
        {
        fp_PoT[i-1] *= f_rMeanPower;
        if ( fp_PoT[i-1] > pot_max ) pot_max = fp_PoT[i-1];
        }
    b_IsAPeak = (pot_max > PoTInfo.GaussPowerThresh);

    count_flops( 3 * swi.analysis_cfg.gauss_pot_length + 2);

    if ( !b_IsAPeak )
        {

        // printf("no peak\n");
        return ( 0 );       // no peak - bail on this PoT
        }

    // Recalculate Total Power across normalized array. Given that powers
    // are positive, f_TotalPower will end up being
    //  == swi.analysis_cfg.gauss_pot_length.
    //
    f_TotalPower = 0;
    f_NormMaxPower = 0;

    // Also locate group with the highest sum for use in second bailout check.
    //
	// BENH: break what was one loop into 2, the 'i_f' before 0 loop and the after
	//   avoids a 50/50 conditional jump inside loop
	//	 also changed a max + index store into a reverse logic 'continue' statement
	//   branch almost allways happens (rarely a max) and avoids conditional fwd jump
	//   also calculate  iPeakLoc outside of the loop
	//  profiler showed this loop using lots of cycles on mispredicts
	//
	f_GroupSum = 0;
    f_GroupMax = 0;
    for ( i = 0, i_f = -iSigma;
		  i_f < 0 && i < swi.analysis_cfg.gauss_pot_length;
		  i++, i_f++ )
        {
        f_TotalPower += fp_PoT[i];
        f_GroupSum   += fp_PoT[i] ;
		if ( fp_PoT[i] > f_NormMaxPower ) f_NormMaxPower = fp_PoT[i];
        if ( f_GroupSum <= f_GroupMax ) continue;
        f_GroupMax = f_GroupSum;
        iPeakLoc = i;
        }

     for (    ; i < swi.analysis_cfg.gauss_pot_length; i++, i_f++ )
        {
        f_TotalPower += fp_PoT[i];
        f_GroupSum   += fp_PoT[i] - fp_PoT[i_f];
		if ( fp_PoT[i] > f_NormMaxPower ) f_NormMaxPower = fp_PoT[i];
        if ( f_GroupSum <= f_GroupMax ) continue;
        f_GroupMax = f_GroupSum;
        iPeakLoc = i;
        }
   iPeakLoc -= iSigma / 2;


   // Check at the group peak location whether data may contain Gaussians
    // (but only after the first hurry-up Gaussian has been set for graphics)
    //
    if ( best_gauss->display_power_thresh != 0 )
        {
        iPeakLoc = std::max(
                PoTInfo.GaussTOffsetStart,
                (std::min( PoTInfo.GaussTOffsetStop - 1, iPeakLoc )) );

        f_TrueMean = f_GetTrueMean(
                fp_PoT,
                swi.analysis_cfg.gauss_pot_length,
                f_TotalPower,
                iPeakLoc,
                2 * iSigma );

        f_PeakPower = f_GetPeak(
                fp_PoT,
                iPeakLoc,
                iSigma,
                f_TrueMean,
                f_PeakScaleFactor,
                f_weight );

        count_flops( 5 * swi.analysis_cfg.gauss_pot_length + 5 );

        if ( f_PeakPower / f_TrueMean < best_gauss->display_power_thresh * 0.5f )
            return ( 0 );   // not even a weak peak at max group - bail on this PoT
        }

    // slide dynamic gaussian across the Power Of Time array
    for ( ul_TOffset = PoTInfo.GaussTOffsetStart;
          ul_TOffset < PoTInfo.GaussTOffsetStop;
          ul_TOffset++ )
        {

        // TrueMean is the mean power of the data set minus all power out to
        // 2 sigma from our current TOffset.
        //
        f_TrueMean = f_GetTrueMean(
                fp_PoT,
                swi.analysis_cfg.gauss_pot_length,
                f_TotalPower,
                ul_TOffset,
                2 * iSigma );
		float f_rTrueMean = 1.0/ f_TrueMean;

        f_PeakPower = f_GetPeak(
                fp_PoT,
                ul_TOffset,
                iSigma,
                f_TrueMean,
                f_PeakScaleFactor,
                f_weight );

        // worth looking at ?
        if ( f_PeakPower * f_rTrueMean < best_gauss->display_power_thresh )
            continue;

        // bump up the display threshold to its final value. We could bump
        // it up only to the gaussian just found, but that would cause a lot
        // of time waste computing chisq etc.
        //
        if ( best_gauss->display_power_thresh == 0 )
            {
            best_gauss->display_power_thresh = PoTInfo.GaussPeakPowerThresh / 3;
            }

	    float rebin = swi.nsamples 
					/ ChirpFftPairs[analysis_state.icfft].FftLen 
					/ swi.analysis_cfg.gauss_pot_length;

        // look at it - try to fit
        f_ChiSq = f_GetChiSq(
                fp_PoT,
                swi.analysis_cfg.gauss_pot_length,
                ul_TOffset,
                f_PeakPower,
                f_TrueMean,
                f_weight,
				rebin,
                &f_null_hyp );

        retval = ChooseGaussEvent(
                ul_TOffset,
                f_PeakPower,
                f_TrueMean,
                f_ChiSq,
                f_null_hyp,
                ul_PoT,
                static_cast< float >(PoTInfo.GaussSigma),
                f_NormMaxPower,
                fp_PoT );
        if ( retval ) SETIERROR( retval, "from ChooseGaussEvent" );

        }                   // End of sliding gaussian

    return ( 0 );

    }                       // End of gaussfit()

	
void ex_count_flops(int the_num)
	{
	count_flops( the_num );
	}
